Type: Package
Package: MazamaLocationUtils
Version: 0.4.5
Title: Manage Spatial Metadata for Known Locations
Authors@R: c(
    person("Jonathan", "Callahan", email="jonathan.s.callahan@gmail.com", role=c("aut","cre")),
    person("Eli", "Grosman", email="eligrosman1@gmail.com", role=c("ctb")),
    person("Oliver", "Fogelin", email="oliverfogelin@gmail.com", role=c("ctb"))
    )
Maintainer: Jonathan Callahan <jonathan.s.callahan@gmail.com>
Description: Utility functions for discovering and managing metadata 
    associated with spatially unique "known locations". Applications include
    all fields of environmental monitoring (e.g. air and water quality) where 
    data are collected at stationary sites.
License: GPL-3
URL: https://github.com/MazamaScience/MazamaLocationUtils
BugReports: https://github.com/MazamaScience/MazamaLocationUtils/issues
Depends: R (>= 4.0)
Imports: cluster, dplyr, geodist (>= 0.1.0), httr, jsonlite, leaflet,
        lubridate, magrittr, methods, MazamaCoreUtils (>= 0.6.0),
        MazamaSpatialUtils (>= 0.8.7), readr, rlang, stringr,
        tidygeocoder
Suggests: knitr, markdown, testthat (>= 2.1.0), rmarkdown, roxygen2
Encoding: UTF-8
Language: en-US
Repository: CRAN
VignetteBuilder: knitr
LazyData: true
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-04-19 16:47:23 UTC; jonathancallahan
Author: Jonathan Callahan [aut, cre],
  Eli Grosman [ctb],
  Oliver Fogelin [ctb]
Date/Publication: 2026-04-19 21:30:02 UTC
Built: R 4.5.2; ; 2026-04-20 23:47:11 UTC; unix

# Generated by roxygen2: do not edit by hand

export("%>%")
export(clusterByDistance)
export(coreMetadataNames)
export(getAPIKey)
export(getLocationDataDir)
export(initializeMazamaSpatialUtils)
export(location_createID)
export(location_getCensusBlock)
export(location_getOpenCageInfo)
export(location_getSingleAddress_Photon)
export(location_getSingleAddress_TexasAM)
export(location_getSingleElevation_USGS)
export(location_initialize)
export(setAPIKey)
export(setLocationDataDir)
export(showAPIKeys)
export(table_addClustering)
export(table_addColumn)
export(table_addCoreMetadata)
export(table_addLocation)
export(table_addOpenCageInfo)
export(table_addSingleLocation)
export(table_filterByDistance)
export(table_findAdjacentDistances)
export(table_findAdjacentLocations)
export(table_getDistanceFromTarget)
export(table_getLocationID)
export(table_getNearestDistance)
export(table_getNearestLocation)
export(table_getRecordIndex)
export(table_initialize)
export(table_initializeExisting)
export(table_leaflet)
export(table_leafletAdd)
export(table_load)
export(table_removeColumn)
export(table_removeRecord)
export(table_save)
export(table_updateColumn)
export(table_updateSingleRecord)
export(validateLocationTbl)
export(validateLonLat)
export(validateLonsLats)
export(validateMazamaSpatialUtils)
importFrom(MazamaCoreUtils,getAPIKey)
importFrom(MazamaCoreUtils,setAPIKey)
importFrom(MazamaCoreUtils,showAPIKeys)
importFrom(MazamaCoreUtils,stopIfNull)
importFrom(MazamaCoreUtils,validateLonLat)
importFrom(MazamaCoreUtils,validateLonsLats)
importFrom(MazamaSpatialUtils,getCountryCode)
importFrom(MazamaSpatialUtils,getStateCode)
importFrom(MazamaSpatialUtils,getTimezone)
importFrom(dplyr,all_of)
importFrom(dplyr,bind_rows)
importFrom(dplyr,filter)
importFrom(dplyr,tibble)
importFrom(geodist,geodist)
importFrom(jsonlite,fromJSON)
importFrom(leaflet,addCircleMarkers)
importFrom(leaflet,addProviderTiles)
importFrom(leaflet,fitBounds)
importFrom(leaflet,leaflet)
importFrom(lubridate,now)
importFrom(magrittr,"%>%")
importFrom(methods,as)
importFrom(rlang,.data)
importFrom(stringr,str_detect)
importFrom(stringr,str_sub)
importFrom(stringr,str_subset)
importFrom(utils,capture.output)

[![CRAN\_Status\_Badge](https://www.r-pkg.org/badges/version/MazamaLocationUtils)](https://cran.r-project.org/package=MazamaLocationUtils)
[![Downloads](https://cranlogs.r-pkg.org/badges/MazamaLocationUtils)](https://cran.r-project.org/package=MazamaLocationUtils)
[![DOI](https://zenodo.org/badge/215816820.svg)](https://zenodo.org/badge/latestdoi/215816820)

A dedicated Slack channel has been created for announcements, support and to 
help build a community of practice around this open source package. You may 
request an invitation to join from <jonathan.s.callahan@gmail.com>.

# MazamaLocationUtils

```
Utility functions for discovering and managing metadata associated 
with spatially unique "known locations". Applications include all 
fields of environmental monitoring (e.g. air and water quality) where 
data are collected at stationary sites.
```

## Background

This package is intended for use in data management activities
associated with fixed locations in space. The motivating fields include 
air and water quality monitoring where fixed sensors report at regular time 
intervals.

When working with environmental monitoring time series, one of the first things
you have to do is create unique identifiers for each individual time series. In 
an ideal world, each environmental time series would have both a 
`locationID` and a `deviceID` that uniquely identify the specific instrument 
making measurements and the physical location where measurements are made. A 
unique `timeseriesID` could
be produced as `locationID_deviceID`. Metadata associated with each
`timeseriesID` would contain basic information needed for downstream analysis
including at least:

`timeseriesID, locationID, deviceID, longitude, latitude, ...`

* An extended time series for an occasionally repositioned sensor would group by `deviceID`.
* Multiple sensors placed at a single location could be be grouped by `locationID`.
* Maps would be created using `longitude, latitude`.
* Time series measurements would be accessed from a secondary `data` table with 
`timeseriesID` column names.

Unfortunately, we are rarely supplied with a truly unique and truly spatial 
`locationID`. Instead we often use `deviceID` or an associated non-spatial
identifier as a stand-in for `locationID`.

Complications we have seen include:

* GPS-reported longitude and latitude can have _jitter_ in the fourth or fifth 
decimal place making it challenging to use them to create a unique `locationID`.
* Sensors are sometimes _re-positioned_ in what the scientist considers the "same 
location".
* Data from a single sensor goes through different processing pipelines using
different identifiers and is later brought together as two separate time series.
* The spatial scale of what constitutes a "single location" depends on the 
instrumentation and scientific question being asked.
* Deriving location-based metadata from spatial datasets is computationally 
intensive unless saved and identified with a unique `locationID`.
* Automated searches for spatial metadata occasionally produce incorrect results
because of the non-infinite resolution of spatial datasets and must be corrected
by hand.

## A Solution

A solution to all these problems is possible if we store spatial metadata in
simple tables in a standard directory. These tables will be referred to as 
_collections_. Location lookups can be performed with
geodesic distance calculations where a longitude-latitude pair is assigned to a pre-existing
_known location_ if it is within `distanceThreshold` meters of that location. 
These lookups will be extremely fast.

If no previously _known location_ is found, the relatively slow (seconds)
creation of a new _known location_ metadata record can be performed and then 
added to the growing collection.

For collections of stationary environmental monitors that only number in the 
thousands, this entire _collection_ can be stored as either a 
`.rda` or `.csv` file and will be under a megabyte in size making it fast to 
load. This small size also makes it possible to save multiple _collections_ 
files, each created with different locations and/or different distance thresholds
to address the needs of different scientific studies.

## Immediate Advantages

Working in this manner solves the problems initially mentioned but also 
provides further useful functionality:

* Administrators can correct entries in an individual _collection_.  (_e.g._ 
locations in river bends that even high resolution spatial datasets mis-assign)
* Additional, non-automatable metadata can be added to a _collection_. (_e.g._
commonly used location names within a community of practice)
* Different field campaigns can maintain separate _collections_.
* `.csv` or `.rda` versions of well populated tables can be downloaded from a
URL and used locally, giving scientists and analysts working with known locations 
instant access to location-specific spatial metadata data that otherwise requires 
special software and skills, large datasets and many compute cycles to generate.

----

Development of this R package has been supported with funding from the 
following institutions:

* USFS [AirFire Research Team](https://www.airfire.org)

Questions regarding further development of the package should be directed to 
<jonathan.callahan@dri.edu>.

## ----echo=FALSE---------------------------------------------------------------
knitr::opts_chunk$set(fig.width = 7, fig.height = 5)

## ----load_data----------------------------------------------------------------
wa <- get(data("wa_airfire_meta", package = "MazamaLocationUtils"))
names(wa)

## ----load_data_hidden, eval = TRUE, echo = FALSE------------------------------
library(MazamaLocationUtils)
wa_monitors_500 <- 
  get(data("wa_monitors_500", package = "MazamaLocationUtils")) %>%
  dplyr::mutate(elevation = as.numeric(NA))

## ----create_table, eval = FALSE, echo = TRUE----------------------------------
#  library(MazamaLocationUtils)
#  
#  # Initialize with standard directories
#  initializeMazamaSpatialUtils()
#  setLocationDataDir("./data")
#  
#  wa_monitors_500 <-
#    table_initialize() %>%
#    table_addLocation(wa$longitude, wa$latitude, distanceThreshold = 500)

## ----basic_columns------------------------------------------------------------
dplyr::glimpse(wa_monitors_500, width = 75)

## ----import_colmns------------------------------------------------------------
# Use a subset of the wa metadata
wa_indices <- seq(5,65,5)
wa_sub <- wa[wa_indices,]

# Use a generic name for the location table
locationTbl <- wa_monitors_500

# Find the location IDs associated with our subset
locationID <- table_getLocationID(
  locationTbl, 
  longitude = wa_sub$longitude, 
  latitude = wa_sub$latitude, 
  distanceThreshold = 500
)

# Now add the "AQSID" column for our subset of locations
locationData <- wa_sub$AQSID
locationTbl <- table_updateColumn(
  locationTbl, 
  columnName = "AQSID", 
  locationID = locationID, 
  locationData = locationData
)

# Lets see how we did
locationTbl_indices <- table_getRecordIndex(locationTbl, locationID)
locationTbl[locationTbl_indices, c("city", "AQSID")]

## ----new_locations------------------------------------------------------------
# Create new locations near our known locations
lons <- jitter(wa_sub$longitude) 
lats <- jitter(wa_sub$latitude)

# Any known locations within 50 meters?
table_getNearestLocation(
  wa_monitors_500,
  longitude = lons,
  latitude = lats,
  distanceThreshold = 50
) %>% dplyr::pull(city)

# Any known locations within 250 meters
table_getNearestLocation(
  wa_monitors_500,
  longitude = lons,
  latitude = lats,
  distanceThreshold = 250
) %>% dplyr::pull(city)

# How about 5000 meters?
table_getNearestLocation(
  wa_monitors_500,
  longitude = lons,
  latitude = lats,
  distanceThreshold = 5000
) %>% dplyr::pull(city)


## ----MSU_setup, echo = TRUE, eval = FALSE-------------------------------------
#    library(MazamaSpatialUtils)
#    setSpatialDataDir("~/Data/Spatial")
#  
#    installSpatialData("EEZCountries")
#    installSpatialData("OSMTimezones")
#    installSpatialData("NaturalEarthAdm1")
#    installSpatialData("USCensusCounties")

## ----standard_setup, echo = TRUE, eval = FALSE--------------------------------
#    MazamaSpatialUtils::setSpatialDataDir("~/Data/Spatial")
#  
#    MazamaSpatialUtils::loadSpatialData("EEZCountries.rda")
#    MazamaSpatialUtils::loadSpatialData("OSMTimezones.rda")
#    MazamaSpatialUtils::loadSpatialData("NaturalEarthAdm1.rda")
#    MazamaSpatialUtils::loadSpatialData("USCensusCounties.rda")

## ----easy_setup, echo = TRUE, eval = FALSE------------------------------------
#    library(MazamaLocationUtils)
#    initializeMazamaSpatialData()
#    setLocationDataDir("~/Data/KnownLocations")

#' @export 
#' 
#' @title Add address fields to a known location table
#' 
#' @description The OpenCage reverse geocoding service is used to update an 
#' existing table. Updated columns include:
#' 
#' \itemize{
#' \item{countryCode}
#' \item{stateCode}
#' \item{countyName}
#' \item{timezone}
#' \item{houseNumber}
#' \item{street}
#' \item{city}
#' \item{postalCode}
#' \item{address}
#' }
#' 
#' When `replaceExisting = TRUE`, all existing address fields are discarded
#' in favor of the OpenCage versions. To only fill in missing values in
#' `locationTbl`, use `replaceExisting = FALSE`.
#' 
#' The OpenCage service returns a large number of fields, some of which may be
#' useful. To add all OpenCage fields to a location table, use 
#' `retainOpenCage = TRUE`. This will append 78+ fields of information,
#' each each named with a prefix of `"opencage_"`.
#' 
#' @note The OpenCage service requires an API key which can be obtained from 
#' their web site. This API key must be set as an environment variable with:
#' 
#' \preformatted{
#' Sys.setenv("OPENCAGE_KEY" = "<your api key>")
#' }
#' 
#' Parameters are set for use at the OpenCage "free trial" level which allows
#' for 1 request/sec and a maximum of 2500 requests per day.
#' 
#' Because of the 1 request/sec default, it is recommended that
#' `table_addOpenCageInfo()` only be used in an interactive session when 
#' updating a table with a large number of records.
#' 
#' @param locationTbl Tibble of known locations.
#' @param replaceExisting Logical specifying whether to replace existing data
#' with data obtained from OpenCage.
#' @param retainOpenCage Logical specifying whether to retain all fields obtained
#' from OpenCage, each named with a prefix of `opencage_`.
#' @param verbose Logical controlling the generation of progress messages.
#' 
#' @return Tibble of "known locations" enhanced with information from the 
#' OpenCage reverse geocoding service.
#' 
#' @references \url{https://opencagedata.com}
#' 
#' @examples
#' \donttest{
#' library(MazamaLocationUtils)
#' 
#' # Fail gracefully if any resources are not available
#' try({
#' 
#'   myTbl <- id_monitors_500[1:3,]
#'   myTbl$countryCode[1] <- NA
#'   myTbl$countryCode[2] <- "WRONG"
#'   myTbl$countyName[3] <- "WRONG"
#'   myTbl$timezone <- NA
#' 
#'   dplyr::glimpse(myTbl)
#' 
#'   Sys.setenv("OPENCAGE_KEY" = "<YOUR_KEY>")
#' 
#'   table_addOpenCageInfo(myTbl) %>% 
#'     dplyr::glimpse()
#' 
#'   table_addOpenCageInfo(myTbl, replaceExisting = TRUE) %>% 
#'     dplyr::glimpse()
#' 
#'   table_addOpenCageInfo(myTbl, replaceExisting = TRUE, retainOpenCage = TRUE) %>% 
#'     dplyr::glimpse()
#'   
#' }, silent = FALSE)
#' }

table_addOpenCageInfo <- function(
  locationTbl = NULL,
  replaceExisting = FALSE,
  retainOpenCage = FALSE,
  verbose = FALSE
) {
  
  # ----- Validate parameters --------------------------------------------------
  
  MazamaLocationUtils::validateLocationTbl(locationTbl, locationOnly = FALSE)
  
  if ( Sys.getenv("OPENCAGE_KEY") == "" )  {
    stop("Revese geocoding with OpenCage requires an API key 
Please set one with Sys.setenv(\"OPENCAGE_KEY\" = \"<YOUR_KEY>\").")
  }
  
  # ----- Reverse geocode ------------------------------------------------------
  
  openCageTbl <- tidygeocoder::reverse_geocode(
    locationTbl,
    lat = "latitude", 
    long = "longitude",  
    method = "opencage",
    address = "address",
    return_input = FALSE,
    limit = 1,
    full_results = TRUE,
    mode = "",
    unique_only = FALSE,
    return_coords = TRUE,
    min_time = NULL,
    progress_bar = FALSE,
    quiet = !verbose,
    api_url = NULL,
    timeout = 20,
    flatten = TRUE,
    batch_limit = NULL,
    verbose = verbose, # Very detailed
    no_query = FALSE,
    custom_query = list(),
    api_options = list()
  )
  
  # > dplyr::glimpse(openCageTbl, width = 75)
  # Rows: 5
  # Columns: 78
  # $ lat                                         <dbl> 43.64789, 45.18191, 4
  # $ long                                        <dbl> -116.2694, -113.8903,
  # $ address                                     <chr> "5164 North Alworth S
  # $ confidence                                  <int> 10, 9, 9, 10, 10
  # $ annotations.MGRS                            <chr> "11TNJ5890232972", "1
  # $ annotations.Maidenhead                      <chr> "DN13up75pj", "DN35be
  # $ annotations.callingcode                     <int> 1, 1, 1, 1, 1
  # $ annotations.flag                            <chr> <flag emojis>
  # $ annotations.geohash                         <chr> "9rv261z09hpr429xq8r8
  # $ annotations.qibla                           <dbl> 23.89, 25.88, 23.82, 
  # $ annotations.DMS.lat                         <chr> "43° 38' 50.70048'' N
  # $ annotations.DMS.lng                         <chr> "116° 16' 10.67196'' 
  # $ annotations.FIPS.county                     <chr> "16001", "16059", "16
  # $ annotations.FIPS.state                      <chr> "16", "16", "16", "16
  # $ annotations.Mercator.x                      <dbl> -12943076, -12678177,
  # $ annotations.Mercator.y                      <dbl> 5381538, 5619763, 537
  # $ annotations.OSM.edit_url                    <chr> "https://www.openstre
  # $ annotations.OSM.note_url                    <chr> "https://www.openstre
  # $ annotations.OSM.url                         <chr> "https://www.openstre
  # $ annotations.UN_M49.statistical_groupings    <list> "MEDC", "MEDC", "MEDC
  # $ annotations.UN_M49.regions.AMERICAS         <chr> "019", "019", "019", 
  # $ annotations.UN_M49.regions.NORTHERN_AMERICA <chr> "021", "021", "021",
  # $ annotations.UN_M49.regions.US               <chr> "840", "840", "840", 
  # $ annotations.UN_M49.regions.WORLD            <chr> "001", "001", "001", 
  # $ annotations.currency.alternate_symbols      <list> "US$", "US$", "US$", 
  # $ annotations.currency.decimal_mark           <chr> ".", ".", ".", ".", "
  # $ annotations.currency.disambiguate_symbol    <chr> "US$", "US$", "US$",
  # $ annotations.currency.html_entity            <chr> "$", "$", "$", "$", "
  # $ annotations.currency.iso_code               <chr> "USD", "USD", "USD", 
  # $ annotations.currency.iso_numeric            <chr> "840", "840", "840", 
  # $ annotations.currency.name                   <chr> "United States Dollar
  # $ annotations.currency.smallest_denomination  <int> 1, 1, 1, 1, 1
  # $ annotations.currency.subunit                <chr> "Cent", "Cent", "Cent
  # $ annotations.currency.subunit_to_unit        <int> 100, 100, 100, 100, 1
  # $ annotations.currency.symbol                 <chr> "$", "$", "$", "$", "
  # $ annotations.currency.symbol_first           <int> 1, 1, 1, 1, 1
  # $ annotations.currency.thousands_separator    <chr> ",", ",", ",", ",", "
  # $ annotations.roadinfo.drive_on               <chr> "right", "right", "ri
  # $ annotations.roadinfo.road                   <chr> "North Alworth Street
  # $ annotations.roadinfo.speed_in               <chr> "mph", "mph", "mph", 
  # $ annotations.sun.rise.apparent               <int> 1642173360, 164217312
  # $ annotations.sun.rise.astronomical           <int> 1642167300, 164216682
  # $ annotations.sun.rise.civil                  <int> 1642171500, 164217114
  # $ annotations.sun.rise.nautical               <int> 1642169340, 164216898
  # $ annotations.sun.set.apparent                <int> 1642120380, 164211948
  # $ annotations.sun.set.astronomical            <int> 1642126440, 164212578
  # $ annotations.sun.set.civil                   <int> 1642122240, 164212146
  # $ annotations.sun.set.nautical                <int> 1642124400, 164212362
  # $ annotations.timezone.name                   <chr> "America/Boise", "Ame
  # $ annotations.timezone.now_in_dst             <int> 0, 0, 0, 0, 0
  # $ annotations.timezone.offset_sec             <int> -25200, -25200, -2520
  # $ annotations.timezone.offset_string          <chr> "-0700", "-0700", "-0
  # $ annotations.timezone.short_name             <chr> "MST", "MST", "MST", 
  # $ annotations.what3words.words                <chr> "sake.pardon.topped",
  # $ bounds.northeast.lat                        <dbl> 43.64747, 45.18156, 4
  # $ bounds.northeast.lng                        <dbl> -116.2696, -113.8889,
  # $ bounds.southwest.lat                        <dbl> 43.64737, 45.17625, 4
  # $ bounds.southwest.lng                        <dbl> -116.2697, -113.8955,
  # $ `components.ISO_3166-1_alpha-2`             <chr> "US", "US", "US", "US
  # $ `components.ISO_3166-1_alpha-3`             <chr> "USA", "USA", "USA", 
  # $ components._category                        <chr> "building", "road", "
  # $ components._type                            <chr> "building", "road", "
  # $ components.continent                        <chr> "North America", "Nor
  # $ components.country                          <chr> "United States", "Uni
  # $ components.country_code                     <chr> "us", "us", "us", "us
  # $ components.county                           <chr> "Ada County", "Lemhi 
  # $ components.house_number                     <chr> "5164", NA, NA, "201"
  # $ components.postcode                         <chr> "83714", "83467", "83
  # $ components.road                             <chr> "North Alworth Street
  # $ components.state                            <chr> "Idaho", "Idaho", "Id
  # $ components.state_code                       <chr> "ID", "ID", "ID", "ID
  # $ components.town                             <chr> "Garden City", "Salmo
  # $ geometry.lat                                <dbl> 43.64742, 45.18115, 4
  # $ geometry.lng                                <dbl> -116.2696, -113.8900,
  # $ annotations.roadinfo.road_type              <chr> NA, "residential", "t
  # $ components.road_type                        <chr> NA, "residential", "t
  # $ components.village                          <chr> NA, "West Salmon", NA
  # $ components.city                             <chr> NA, NA, "Meridian", N
  
  # ----- Replace values -------------------------------------------------------
  
  # * countryCode -----
  
  if ( "components.country_code" %in% names(openCageTbl) ) {
    if ( replaceExisting ) {
      locationTbl$countryCode <- toupper(openCageTbl$components.country_code)
    } else {
      mask <- is.na(locationTbl$countryCode)
      locationTbl$countryCode[mask] <- toupper(openCageTbl$components.country_code[mask])
    }
  }
  
  # * stateCode -----
  
  if ( "components.state_code" %in% names(openCageTbl) ) {
    if ( replaceExisting ) {
      locationTbl$stateCode <- toupper(openCageTbl$components.state_code)
    } else {
      mask <- is.na(locationTbl$stateCode)
      locationTbl$stateCode[mask] <- toupper(openCageTbl$components.state_code[mask])
    }
  }
  
  # * countyName -----
  
  if ( "components.county" %in% names(openCageTbl) ) {
    if ( replaceExisting ) {
      locationTbl$countyName <- 
        stringr::str_replace(openCageTbl$components.county, " County", "")
    } else {
      mask <- is.na(locationTbl$countyName)
      locationTbl$countyName[mask] <- 
        stringr::str_replace(openCageTbl$components.county[mask], " County", "")
    }
  }
  
  # * timezone -----
  
  if ( "annotations.timezone.name" %in% names(openCageTbl) ) {
    if ( replaceExisting ) {
      locationTbl$timezone <- openCageTbl$annotations.timezone.name
    } else {
      mask <- is.na(locationTbl$timezone)
      locationTbl$timezone[mask] <- openCageTbl$annotations.timezone.name[mask]
    }
  }
  
  # * houseNumber -----
  
  if ( "components.house_number" %in% names(openCageTbl) ) {
    if ( replaceExisting ) {
      locationTbl$houseNumber <- as.character(openCageTbl$components.house_number)
    } else {
      mask <- is.na(locationTbl$houseNumber)
      locationTbl$houseNumber[mask] <- as.character(openCageTbl$components.house_number[mask])
    }
  }
  
  # * street -----
  
  if ( "components.road" %in% names(openCageTbl) ) {
    if ( replaceExisting ) {
      locationTbl$street <- as.character(openCageTbl$components.road)
    } else {
      mask <- is.na(locationTbl$street)
      locationTbl$street[mask] <- as.character(openCageTbl$components.road[mask])
    }
  }
  
  # * city -----
  
  if ( "components.town" %in% names(openCageTbl) ) {
    if ( replaceExisting ) {
      locationTbl$city <- as.character(openCageTbl$components.town)
    } else {
      mask <- is.na(locationTbl$city)
      locationTbl$city[mask] <- as.character(openCageTbl$components.town[mask])
    }
  }
  
  # NOTE:  Some OpenCage records are missing "town" but have "city" so add this
  # NOTE:  where records are still missing a value
  if ( "components.city" %in% names(openCageTbl) ) {
    mask <- is.na(locationTbl$city)
    locationTbl$city[mask] <- as.character(openCageTbl$components.city[mask])
  }
  
  # * postalCode -----
  
  if ( "components.postcode" %in% names(openCageTbl) ) {
    if ( replaceExisting ) {
      locationTbl$postalCode <- as.character(openCageTbl$components.postcode)
    } else {
      mask <- is.na(locationTbl$postalCode)
      locationTbl$postalCode[mask] <- as.character(openCageTbl$components.postcode[mask])
    }
  }
  
  # * address -----
  
  # NOTE:  'address' is not part of the core metadata but is very useful
  if ( !"address" %in% names(locationTbl) ) 
    locationTbl$address <- as.character(NA)
  
  if ( "address" %in% names(openCageTbl) ) {
    if ( replaceExisting ) {
      locationTbl$address <- as.character(openCageTbl$address)
    } else {
      mask <- is.na(locationTbl$address)
      locationTbl$address[mask] <- as.character(openCageTbl$address[mask])
    }
  }
  
  # ----- Add openCage ---------------------------------------------------------
  
  if ( retainOpenCage ) {
    
    names(openCageTbl) <- paste0("opencage_", names(openCageTbl))
    locationTbl <- dplyr::bind_cols(locationTbl, openCageTbl)
    
  }
  
  # ----- Return ---------------------------------------------------------------
  
  return(locationTbl)
  
}

# ===== DEBUGGING ==============================================================

if ( FALSE ) {
  
  library(MazamaLocationUtils)
  
  locationTbl <- 
    MazamaLocationUtils::id_monitors_500 %>%
    dplyr::slice(1:5)
  
  replaceExisting <- TRUE
  retainOpenCage <- TRUE
  verbose <- FALSE
  
}

#' @title Save a known location table
#' @description Save a tibble of known locations to the preferred directory. If
#' `outputType` is a vector, the known locations table will be saved to the 
#' preferred directory in multiple formats.
#' 
#' @param locationTbl Tibble of known locations.
#' @param collectionName Character identifier for this table.
#' @param backup Logical specifying whether to save a backup version of any
#' existing tables sharing `collectionName`.
#' @param outputType Vecctor of output formats. (Currently only "rda" or "csv" are supported.)
#' @return File path of saved file.
#' @examples
#' library(MazamaLocationUtils)
#' 
#' # Set the directory for saving location tables
#' setLocationDataDir(tempdir())
#' 
#' # Load an example table and check the dimensions
#' locationTbl <- get(data("wa_monitors_500"))
#' dim(locationTbl)
#' 
#' # Save it as "table_save_example"
#' table_save(locationTbl, "table_save_example")
#' 
#' # Add a column and save again 
#' locationTbl %>% 
#'   table_addColumn("my_column") %>% 
#'   table_save("table_save_example")
#'   
#' # Check the locationDataDir
#' list.files(getLocationDataDir(), pattern = "table_save_example")
#' 
#' @details Backup files are saved with "YYYY-mm-ddTHH:MM:SS"
#' @rdname table_save
#' @export 
#' @importFrom MazamaCoreUtils stopIfNull
#' @importFrom lubridate now
table_save <- function(
    locationTbl = NULL,
    collectionName = NULL,
    backup = TRUE,
    outputType = "rda"
) {
  
  # ----- Validate parameters --------------------------------------------------
  
  MazamaLocationUtils::validateLocationTbl(locationTbl, locationOnly = FALSE)
  MazamaCoreUtils::stopIfNull(collectionName)
  
  for ( type in outputType ) {
    if ( !type %in% c("rda", "csv") ) 
      stop("outputType '%s' is not supported.", type)
  }
  
  dataDir <- getLocationDataDir()
  
  # ----- Save data ------------------------------------------------------------
  
  for ( type in outputType ) {
    
    result <- try({
      
      fileName <- paste0(collectionName, ".", type)
      filePath <- file.path(dataDir, fileName)
      
      # Save backups
      if ( backup && file.exists(filePath) ) {
        backupName <- paste0(
          collectionName, ".",
          strftime(lubridate::now(), "%Y-%m-%dT%H:%M:%S"),
          ".rda"
        )
        backupPath <- file.path(dataDir, backupName)
        file.rename(filePath, backupPath)
      }
      
      if ( type == "rda" ) {
        # Assign a name and save
        assign(collectionName, locationTbl)
        save(list = c(collectionName), file = filePath)
      }
      
      if ( type == "csv" ) {
        readr::write_csv(locationTbl, file = filePath)
      }
      
    }, silent = TRUE)
    
    if ( "try-error" %in% class(result) ) {
      warning(sprintf("Could not write %s", filePath))
    }
    
  } # END of type loop
  
  
  # ----- Return ---------------------------------------------------------------
  
  return(invisible(filePath))
  
}

#' @rdname table_getNearestDistance
#' @export
#' @importFrom MazamaCoreUtils stopIfNull
#' @importFrom geodist geodist
#' @importFrom rlang .data
#' 
#' @title Return distances to nearest known locations
#' 
#' @param locationTbl Tibble of known locations.
#' @param longitude Vector of target longitudes in decimal degrees E.
#' @param latitude Vector of target latitudes in decimal degrees N.
#' @param distanceThreshold Distance in meters.
#' @param measure One of "geodesic", "haversine", "vincenty" or "cheap" 
#' specifying desired method of geodesic distance calculation. 
#' 
#' @description Returns distances between target locations and the closest
#' location found in `locationTbl` (if any). Target locations are specified 
#' with `longitude` and `latitude`.
#' 
#' For each target location, only a single distance to the closest known location 
#' is returned. If no known location is found within
#' `distanceThreshold`, the distance associated with that target location
#' will be `NA`. The length and order of resulting distances will match the
#' order of the incoming target locations.
#' 
#' @section Use Case:
#' You may have a set of locations of interest for which you want to assess whether
#' any monitoring locations are nearby. In this case, the locations of interest
#' will provide `longitude` and `latitude` while `locationTbl`
#' will be the known location table associated with the monitoring locations. 
#' 
#' The resulting vector of distances will tell you the distance, for each 
#' target location, to the nearst monitoring location.
#' 
#' @note The measure `"cheap"` may be used to speed things up depending on
#' the spatial scale being considered. Distances calculated with 
#' `measure = "cheap"` will vary by a few meters compared with those 
#' calculated using `measure = "geodesic"`.
#' 
#' See [geodist::geodist()] for details.
#' 
#' @return Vector of closest distances between target locations and known locations.
#' @examples
#' library(MazamaLocationUtils)
#' 
#' locationTbl <- get(data("wa_monitors_500"))
#' 
#' # Wenatchee
#' lon <- -120.325278
#' lat <- 47.423333
#' 
#' # Too small a distanceThreshold will not find a match
#' table_getNearestDistance(locationTbl, lon, lat, distanceThreshold = 50)
#' 
#' # Expanding the distanceThreshold will find one
#' table_getNearestDistance(locationTbl, lon, lat, distanceThreshold = 5000)
#' 
table_getNearestDistance <- function(
  locationTbl = NULL,
  longitude = NULL,
  latitude = NULL,
  distanceThreshold = NULL,
  measure = c("geodesic", "haversine", "vincenty", "cheap")
) {

  # ----- Validate parameters --------------------------------------------------

  MazamaLocationUtils::validateLocationTbl(locationTbl, locationOnly = TRUE)
  MazamaLocationUtils::validateLonsLats(longitude, latitude)
  MazamaCoreUtils::stopIfNull(distanceThreshold)
  measure <- match.arg(measure)

  if ( !is.numeric(distanceThreshold) )
    stop("Parameter 'distanceThreshold' must be a numeric value.")
  
  distanceThreshold <- round(distanceThreshold)

  # ----- Find unique locations ------------------------------------------------
  
  # NOTE:  For the case where the incoming longitude and latitude have many 
  # NOTE:  duplicated locations (a "tidy" dataframe perhaps), we need to speed
  # NOTE:  things up by only calculating distances for unique locations and then
  # NOTE:  merging with all incoming locations.
  
  myTbl <- dplyr::tibble(longitude = longitude, latitude = latitude)
  
  myUniqueTbl <- myTbl %>% dplyr::distinct()
  
  # ----- Calculate distances --------------------------------------------------

  distance <-
    geodist::geodist(
      y = cbind(
        "x" = myUniqueTbl$longitude,
        "y" = myUniqueTbl$latitude
      ),
      x = cbind(
        "x" = locationTbl$longitude,
        "y" = locationTbl$latitude
      ),
      paired = FALSE,
      sequential = FALSE,
      pad = FALSE,
      measure = measure
    )

  # NOTE:  distance matrix is nrow(locationTbl) X nrow(myUniqueTbl)

  # ----- Find unique distances ------------------------------------------------
  
  myUniqueTbl$nearestDistance <- as.numeric(NA)
  
  for ( index in seq_len(nrow(myUniqueTbl)) ) {
    
    if ( any(distance[,index] <= distanceThreshold, na.rm = TRUE) ) {
      myUniqueTbl$nearestDistance[index] <-  min(distance[,index], na.rm = TRUE)
    }
    
  }
  
  # ----- Merge with all locations ---------------------------------------------
  
  nearestDistance <- 
    dplyr::full_join(
      myTbl,
      myUniqueTbl,
      by = c("longitude", "latitude")
    ) %>%
    dplyr::pull(.data$nearestDistance)
  
  # ----- Return ---------------------------------------------------------------

  return(nearestDistance)

}

context("table_getNearestDistance")

test_that("table_getNearestDistance() works", {
  locationTbl <- get(data("wa_monitors_500"))
  
  # Wenatchee
  lon <- -120.325278
  lat <- 47.423333
  
  testClose <- table_getNearestDistance(locationTbl, lon, lat, distanceThreshold = 100)
  testFar <- table_getNearestDistance(locationTbl, lon, lat, distanceThreshold = 10000)
  expect_equal(testClose, as.numeric(NA))
  expect_equal(testFar, 1495, tolerance = 1.0)
})

#' @rdname location_initialize
#' @export 
#' @importFrom MazamaCoreUtils stopIfNull
#' @importFrom MazamaSpatialUtils getCountryCode getStateCode getTimezone
#' @importFrom dplyr tibble
#' @importFrom stringr str_sub
#' 
#' @title  Create known location record with core metadata
#' 
#' @param longitude Single longitude in decimal degrees E.
#' @param latitude Single latitude in decimal degrees N.
#' @param stateDataset Name of spatial dataset to use for determining state
#' @param elevationService Name of the elevation service to use for determining
#' the elevation. Default: NULL skips this step. Accepted values: "usgs".
#' @param addressService Name of the address service to use for determining
#' the street address. Default: NULL skips this step. Accepted values: "photon".
#' @param precision `precision` argument passed on to [location_createID()].
#' @param verbose Logical controlling the generation of progress messages.
#' 
#' @return Tibble with a single new known location.
#' 
#' @description Creates a known location record with the following columns
#' of core metadata:
#' \itemize{
#' \item{locationID}
#' \item{locationName}
#' \item{longitude}
#' \item{latitude}
#' \item{elevation}
#' \item{countryCode}
#' \item{stateCode}
#' \item{countyName}
#' \item{timezone}
#' \item{houseNumber}
#' \item{street}
#' \item{city}
#' \item{postalCode}
#' }
#' 
#' @examples
#' \donttest{
#' library(MazamaLocationUtils)
#' 
#' # Fail gracefully if any resources are not available
#' try({
#' 
#'   # Set up standard directories and spatial data
#'   spatialDataDir <- tempdir() # typically "~/Data/Spatial"
#'   initializeMazamaSpatialUtils(spatialDataDir)
#' 
#'   # Wenatchee
#'   lon <- -120.325278
#'   lat <- 47.423333
#'   
#'   locationRecord <- location_initialize(lon, lat)
#'   str(locationRecord)
#'   
#' }, silent = FALSE)
#' }

location_initialize <- function(
  longitude = NULL,
  latitude = NULL,
  stateDataset = "NaturalEarthAdm1",
  elevationService = NULL,
  addressService = NULL,
  precision = 10,
  verbose = TRUE
) {
  
  validateMazamaSpatialUtils()
  
  # ----- Validate parameters --------------------------------------------------
  
  validateLonLat(longitude, latitude)
  
  MazamaCoreUtils::stopIfNull(stateDataset)
  MazamaCoreUtils::stopIfNull(precision)
  
  if ( !exists(stateDataset) ) {
    stop(paste0(
      "You must load \"stateDataset\" with: \n",
      "  loadSpatialData(\"", stateDataset, "\")\n"
    ))
  }
  
  if ( !is.null(addressService) ) {
    if ( !is.character(addressService) ) {
      stop("Currently, only the \"photon\" address service is supported.")
    } else {
      if ( tolower(addressService) != "photon" )
        stop("Currently, only the \"photon\" address service is supported.")
    }
  }
  
  if ( !is.null(elevationService) ) {
    if ( !is.character(elevationService) ) {
      stop("Currently, only the \"usgs\" address service is supported.")
    } else {
      if ( tolower(elevationService) != "usgs" )
        stop("Currently, only the \"usgs\" address service is supported.")
    }
  }
  
  # ----- locationID, Elevation ------------------------------------------------
  
  locationID <- location_createID(
    longitude = longitude,
    latitude = latitude,
    algorithm = "geohash",
    precision = precision
  )
  
  if ( is.null(elevationService) ) {
    
    elevation <- NA
    
  } else {
    
    if ( tolower(elevationService) != "usgs" ) {
      
        elevation <- NA
      
    } else {
      
      elevation <- location_getSingleElevation_USGS(
        longitude = longitude,
        latitude = latitude,
        verbose = verbose
      )  
      
    }
  }

  
  # ----- Country, State, County, Timezone -------------------------------------
  
  countryCode <- MazamaSpatialUtils::getCountryCode(
    lon = longitude,
    lat = latitude,
    dataset = "EEZCountries",
    useBuffering = FALSE
  )
  
  stateCode <- MazamaSpatialUtils::getStateCode(
    lon = longitude,
    lat = latitude,
    dataset = stateDataset,
    useBuffering = TRUE
  )
  
  countyName <- MazamaSpatialUtils::getUSCounty(
    lon = longitude,
    lat = latitude,
    dataset = "USCensusCounties",
    stateCodes = stateCode,
    useBuffering = TRUE
  )
  
  timezone <- MazamaSpatialUtils::getTimezone(
    lon = longitude,
    lat = latitude,
    dataset = "OSMTimezones",
    useBuffering = TRUE
  )
  
  # ----- LocationName ---------------------------------------------------------
  
  # NOTE:  The default locationName is intended to give folks a more memorable
  # NOTE:  handel than the locationID but is not guaranteed to be unique. It is 
  # NOTE:  expected that users will add their own, more relevant names 
  # NOTE:  appropriate for the community of practice using a particular
  # NOTE:  collectionName of known locations.
  
  locationName <- paste0(
    tolower(countryCode), ".",
    tolower(stateCode), "_",
    stringr::str_sub(locationID, 1, 6)
  )
  
  # ----- Address --------------------------------------------------------------
  
  if ( is.null(addressService) ) {
    
    addressList <- list(
      "houseNumber" = as.character(NA),
      "street" = as.character(NA),
      "city" = as.character(NA),
      "postalCode" = as.character(NA)
    )
    
  } else {
    
    if ( tolower(addressService) != "photon" ) {
      
      addressList <- list(
        "houseNumber" = as.character(NA),
        "street" = as.character(NA),
        "city" = as.character(NA),
        "postalCode" = as.character(NA)
      )
      
    } else {
      
      addressList <- location_getSingleAddress_Photon(
        longitude = longitude,
        latitude = latitude,
        verbose = verbose
      )
      
      # NOTE:  The Photon reverse geocoding service returns NA for countryCode
      # NOTE:  and stateCode when no specific address can be found. In these 
      # NOTE:  cases (remote areas) we want to use the MazamaSpatialUtils results.
      # NOTE:  Otherwise, we trust that Photon reverse geocoding is more accurate
      # NOTE:  than the point-in-polygon seraches done by MazamaSpatialUtils.
      
      if ( !is.na(addressList$countryCode) &&
           addressList$countryCode != countryCode ) {
        # Trust address data over MazamaSpatialUtils result
        if ( verbose ) {
          warning(sprintf(
            "Using address countryCode \"%s\" over MazamaSpatialUtils \"%s\"",
            addressList$countryCode, countryCode
          ))
        }
        countryCode <- addressList$countryCode
      }
      
      if ( !is.na(addressList$stateCode) &&
           addressList$stateCode != stateCode ) {
        
        # Trust address data over MazamaSpatialUtils result
        MazamaStateCode <- stateCode
        
        # Change stateCode and reset possibly incorrect countyName
        stateCode <- addressList$stateCode
        countyName <- as.character(NA)
        
        # Update locationName
        locationName <- paste0(
          tolower(countryCode), ".",
          tolower(stateCode), "_",
          stringr::str_sub(locationID, 1, 6)
        )
        
        if ( verbose ) {
          warning(sprintf(
            "For %s, using address stateCode \"%s\" over MazamaSpatialUtils \"%s\"",
            locationName, addressList$stateCode, MazamaStateCode
          ))
        }
        
      }
      
    } 
    
  } # END is.null(addressService) check
  
  # ----- Return ---------------------------------------------------------------
  
  locationTbl <- dplyr::tibble(
    "locationID" = locationID,
    "locationName" = locationName,
    "longitude" = longitude,
    "latitude" = latitude,
    "elevation" = elevation,
    "countryCode" = toupper(countryCode),
    "stateCode" = toupper(stateCode),
    "countyName" = countyName,
    "timezone" = timezone,
    "houseNumber" = as.character(addressList$houseNumber),
    "street" = as.character(addressList$street),
    "city" = as.character(addressList$city),
    "postalCode" = as.character(addressList$postalCode)
  )
  
  return(locationTbl)
  
}

# ----- API Keys ---------------------------------------------------------------

#' @docType data
#' @keywords environment
#' @name APIKeys
#' @title API keys for data services.
#' @format Character strings.
#' @description This package maintains an internal set of API keys which
#' users can set using `setAPIKey()`. These keys will be remembered for
#' the duration of an R session. This functionality provides an abstraction
#' layer in dependent packages so that data access functions can test for and
#' access specific API keys with generic code.
#' 
#' @details
#' The following functions help with the management of API keys:
#' 
#' `getAPIKey()` -- Returns the API key associated with a web service.
#' If `provider == NULL` a list is returned containing all recognized
#' API keys.
#' 
#' `setAPIKey()` -- Sets the API key associated with a web service.
#' Silently returns previous value of the API key.
#' 
#' `showAPIKeys()` -- Returns a list of all currently set API keys.
#'
NULL


#' @rdname APIKeys
#' @name getAPIKey
#' @export
#' @importFrom MazamaCoreUtils getAPIKey
NULL

#' @rdname APIKeys
#' @name setAPIKey
#' @export
#' @importFrom MazamaCoreUtils setAPIKey
NULL

#' @rdname APIKeys
#' @name showAPIKeys
#' @export
#' @importFrom MazamaCoreUtils showAPIKeys
NULL

#' @title Return IDs of known locations
#' @description Returns a vector of `locationID`s for the known locations
#' that each incoming location will be assigned to within the given. If more
#' than one known location exists within the given `distanceThreshold`, the closest will be
#' assigned.  `NA` will be returned for each incoming that cannot be 
#' assigned to a known location in `locationTbl`.
#' 
#' @note The measure `"cheap"` may be used to speed things up depending on
#' the spatial scale being considered. Distances calculated with 
#' `measure = "cheap"` will vary by a few meters compared with those 
#' calculated using `measure = "geodesic"`.
#' 
#' @param locationTbl Tibble of known locations.
#' @param longitude Vector of longitudes in decimal degrees E.
#' @param latitude Vector of latitudes in decimal degrees N.
#' @param distanceThreshold Distance in meters.
#' @param measure One of "geodesic", "haversine", "vincenty" or "cheap" 
#' specifying desired method of geodesic distance calculation. 
#' See [geodist::geodist()].
#' 
#' @return Vector of known `locationID`s.
#' @examples
#' locationTbl <- get(data("wa_monitors_500"))
#' 
#' # Wenatchee
#' lon <- -120.325278
#' lat <- 47.423333
#' 
#' # Too small a distanceThreshold will not find a match
#' table_getLocationID(locationTbl, lon, lat, distanceThreshold = 50)
#' 
#' # Expanding the distanceThreshold will find one
#' table_getLocationID(locationTbl, lon, lat, distanceThreshold = 5000)
#' 
#' @rdname table_getLocationID
#' @export 
#' @importFrom MazamaCoreUtils stopIfNull
#' @importFrom geodist geodist
#' @importFrom rlang .data
table_getLocationID <- function(
  locationTbl = NULL,
  longitude = NULL,
  latitude = NULL,
  distanceThreshold = NULL,
  measure = c("geodesic", "haversine", "vincenty", "cheap")
) {
  
  # ----- Validate parameters --------------------------------------------------
  
  MazamaLocationUtils::validateLocationTbl(locationTbl, locationOnly = FALSE)
  MazamaLocationUtils::validateLonsLats(longitude, latitude)
  MazamaCoreUtils::stopIfNull(distanceThreshold)
  measure <- match.arg(measure)
  
  if ( !is.numeric(distanceThreshold) )
    stop("Parameter 'distanceThreshold' must be a numeric value.")
  
  distanceThreshold <- round(distanceThreshold)

  # ----- Find unique locations ------------------------------------------------
  
  # NOTE:  For the case where the incoming longitude and latitude have many 
  # NOTE:  duplicated locations (a "tidy" dataframe perhaps), we need to speed
  # NOTE:  things up by only calculating distance for unique locations and then
  # NOTE:  merging with all incoming locations.
  
  myTbl <- dplyr::tibble(longitude = longitude, latitude = latitude)
  
  myUniqueTbl <- myTbl %>% dplyr::distinct()
  
  # ----- Calculate distances --------------------------------------------------

  distance <-
    geodist::geodist(
      y = cbind(
        "x" = myUniqueTbl$longitude,
        "y" = myUniqueTbl$latitude
      ),
      x = cbind(
        "x" = locationTbl$longitude,
        "y" = locationTbl$latitude
      ),
      paired = FALSE,
      sequential = FALSE,
      pad = FALSE,
      measure = measure
    )

  # NOTE:  distance matrix is nrow(locationTbl) X nrow(myUniqueTbl)

  # ----- Find unique locationIDs ----------------------------------------------

  myUniqueTbl$locationID <- as.character(NA)
  
  for ( index in seq_len(nrow(myUniqueTbl)) ) {

    if ( any(distance[,index] <= distanceThreshold, na.rm = TRUE) ) {
      row <- which(distance[,index] == min(distance[,index], na.rm = TRUE))
      myUniqueTbl$locationID[index] <- locationTbl$locationID[row]
    }

  }
  
  # ----- Merge with all locations ---------------------------------------------
  
  locationID <- 
    dplyr::full_join(
      myTbl,
      myUniqueTbl,
      by = c("longitude", "latitude")
    ) %>%
    dplyr::pull(.data$locationID)
  
  # ----- Return ---------------------------------------------------------------
  
  return(locationID)
  
}

# ===== DEBUG ==================================================================

if ( FALSE ) {
  
  library(MazamaLocationUtils)
  
  locationTbl <- MazamaCoreUtils::loadDataFile(
    filename = "airnow_PM2.5_sites.rda",
    dataUrl = NULL,
    dataDir = "~/Data/known_locations"
  )
  
  airnow_data <- MazamaCoreUtils::loadDataFile(
    filename = "airnow_PM2.5_202111.rda",
    dataUrl = "https://data-monitoring1.airfire.org/monitoring-v2/airnow-latency/2021",
    dataDir = NULL
  )
  
  longitude <- airnow_data$longitude
  latitude <- airnow_data$latitude
  
  distanceThreshold <- 100
  measure <- "geodesic"
  
  
  locationID <- 
    table_getLocationID(
      locationTbl = locationTbl,
      longitude = longitude,
      latitude = latitude,
      distanceThreshold = distanceThreshold,
      measure = measure
    )
  
  distance <- 
    table_getNearestDistance(
      locationTbl = locationTbl,
      longitude = longitude,
      latitude = latitude,
      distanceThreshold = distanceThreshold,
      measure = measure
    )

  
}

context("table_getLocationID")

test_that("correct locationID is returned", {
  locationTbl <- get(data("wa_monitors_500"))
  
  locationID <- wa_monitors_500$locationID
  longitude <- wa_monitors_500$longitude
  latitude <- wa_monitors_500$latitude
  
  testID <- table_getLocationID(locationTbl, longitude, latitude, distanceThreshold = 1000)
  expect_equal(locationID, testID)
})

test_that("radius works", {
  locationTbl <- get(data("wa_monitors_500"))
  
  locationID <- wa_monitors_500$locationID
  longitude <- wa_monitors_500$longitude + 0.01 # ~ 780 m east
  latitude <- wa_monitors_500$latitude + 0.01 # ~ 780 m north
  allNA <- rep(as.character(NA), length(locationID))
  
  testID <- table_getLocationID(locationTbl, longitude, latitude, distanceThreshold = 1000)
  expect_equal(allNA, testID)
  
  testID <- table_getLocationID(locationTbl, longitude, latitude, distanceThreshold = 10000)
  expect_equal(locationID, testID)
})

context("table_updateColumn")

test_that("table_updateColumn() works with no data", {
  locationTbl <- get(data("wa_monitors_500"))
  testTbl <- table_updateColumn(locationTbl, "AQSID")
  expect_equal(c(names(locationTbl),"AQSID"), names(testTbl))
})

test_that("table_updateColumn() works with data", {
  
  # Update records in a known location table with information obtained elsewhere
  
  locationTbl <- get(data("wa_monitors_500"))
  airfire <- get(data("wa_airfire_meta"))
  
  # Record indices for airfire
  airfire_indices <- seq(5,65,5)
  airfire_sub <- airfire[airfire_indices,]
  
  # Get matching locations in locationTbl
  # NOTE:  Some locationIDs may be NA
  locationID <- table_getLocationID(locationTbl, airfire_sub$longitude, airfire_sub$latitude, distanceThreshold = 1000)
  locationData <- airfire_sub$AQSID
  
  # Create a new column
  testTbl <- table_updateColumn(locationTbl, "AQSID", locationID, locationData)
  
  testTbl_indices <- table_getRecordIndex(testTbl, locationID)
  
  locationTbl_AQSID <- testTbl$AQSID[testTbl_indices]
  airfire_AQSID <- airfire$AQSID[airfire_indices]
  
  # NOTE:  locationTbl may not have every location in airfire
  mask <- !is.na(locationTbl_AQSID)

  expect_equal(locationTbl_AQSID[mask], airfire_AQSID[mask])
  
})

test_that("table_updateColumn() skips unknown locations", {
  
  locationTbl <- get(data("wa_monitors_500"))
  airfire <- get(data("wa_airfire_meta"))
  
  # Record indices into for airfire
  airfire_indices <- seq(5,65,5) 
  airfire_sub <- airfire[airfire_indices,]
  
  # NOTE:  Include locationIDs not found in locationTbl
  locationID <- c("locationID_NOT_FOUND", table_getLocationID(locationTbl, airfire_sub$longitude, airfire_sub$latitude, distanceThreshold = 1000))
  locationData <- c("locationData_NOT_FOUND", airfire_sub$AQSID)
  
  # update an Existing column
  testTbl <- table_updateColumn(locationTbl, "locationName", locationID, locationData)
  testTbl_indices <- 
    table_getRecordIndex(testTbl, locationID) %>% 
    na.omit() %>% as.numeric()
  
  expect_true({
    all( testTbl$locationName[testTbl_indices] %in% airfire$AQSID )
  })
  
})

#' @rdname table_getDistanceFromTarget
#' @export
#' @importFrom MazamaCoreUtils stopIfNull
#' @importFrom rlang .data
#' 
#' @title Return distances and directions from a target location to known locations
#' 
#' @param locationTbl Tibble of known locations.
#' @param longitude Target longitude in decimal degrees E.
#' @param latitude Target latitude in decimal degrees N.
#' @param measure One of "geodesic", "haversine", "vincenty" or "cheap" 
#' specifying desired method of geodesic distance calculation. 
#' 
#' @return Tibble of distances in meters and cardinal directions from a target location.
#' 
#' @description Returns a tibble with the same number of rows as `locationTbl` 
#' containing the distance and direction from the target location specified by 
#' `longitude` and `latitude` to each known location found in 
#' `locationTbl`. 
#' 
#' @note Only a single target location is allowed.
#' 
#' @examples
#' library(MazamaLocationUtils)
#' 
#' locationTbl <- get(data("wa_monitors_500"))
#' 
#' locationTbl %>%
#'   table_getDistanceFromTarget(
#'     longitude = -117.3647, 
#'     latitude = 47.6725
#'   ) %>% 
#'   dplyr::glimpse()
#' 


table_getDistanceFromTarget <- function(
  locationTbl = NULL,
  longitude = NULL,
  latitude = NULL,
  measure = c("geodesic", "haversine", "vincenty", "cheap")
) {

  # ----- Validate parameters --------------------------------------------------

  MazamaLocationUtils::validateLocationTbl(locationTbl, locationOnly = TRUE)
  MazamaLocationUtils::validateLonLat(longitude, latitude)
  measure <- match.arg(measure)
  
  if ( length(longitude) > 1 || length(latitude) > 1 ) 
    stop("Only a single target location is allowed.")
  
  # ----- Calculate distances --------------------------------------------------
  
  distanceMatrix <-
    geodist::geodist(
      y = cbind(
        "x" = longitude,
        "y" = latitude
      ),
      x = cbind(
        "x" = locationTbl$longitude,
        "y" = locationTbl$latitude
      ),
      paired = FALSE,
      sequential = FALSE,
      pad = FALSE,
      measure = measure
    )
  
  # NOTE:  distanceMatrix is nrow(locationTbl) X 1
  
  returnTbl <- 
    locationTbl %>%
    dplyr::select(c("locationID")) %>%
    dplyr::mutate(
      distanceFromTarget = round(distanceMatrix[,1])
    )

  # ----- Add direction --------------------------------------------------------
  
  # https://stackoverflow.com/questions/30794729/find-angle-from-two-points-formula
  
  x1 <- longitude * pi/180
  x2 <- locationTbl$longitude * pi/180
  y1 <- latitude * pi/180
  y2 <- locationTbl$latitude * pi/180
  
  angle = atan2(y2 - y1, x2 - x1) * 180/pi
  
  # NOTE:  Mathematical angle is zero facing East and increases counter-clockwise
  # NOTE:  Cover everything between -180:180
  
  returnTbl$directionFromTarget <- 
    cut(
      angle,
      breaks = c(-202.5, -157.5, -112.5, -67.5, -22.5, 22.5, 67.5, 112.5, 157.5, 202.5),
      labels = c("W", "SW", "S", "SE", "E", "NE", "N", "NW", "W")
    ) %>%
    as.character()
  
  # ----- Return ---------------------------------------------------------------

  return(returnTbl)

}

#' @title Update a column of metadata in a table
#' 
#' @description Updates records in a location table. Records are identified
#' by `locationID` and the data found in `locationData` is used to 
#' replace any existing value in the `columnName` column.
#' `locationID` and `locationData` must be of the same length.
#' Any `NA` values in `locationID` will be ignored.
#' 
#' If `columnName` is not a named column within `locationTbl`, a new
#' column will be created.
#' 
#' @param locationTbl Tibble of known locations.
#' @param columnName Name of an existing/new column in `locationTbl` whose data
#' will be updated/created.
#' @param locationID Vector of `locationID` strings.
#' @param locationData Vector of data to be inserted at records identified by 
#' `locationID`.
#' @param verbose Logical controlling the generation of progress messages.
#' @return Updated tibble of known locations.
#' @examples
#' library(MazamaLocationUtils)
#' 
#' locationTbl <- get(data("wa_monitors_500"))
#' wa <- get(data("wa_airfire_meta"))
#' 
#' # We will merge some metadata from wa into locationTbl
#' 
#' # Record indices for wa
#' wa_indices <- seq(5,65,5)
#' wa_sub <- wa[wa_indices,]
#' 
#' locationID <- 
#'   table_getLocationID(
#'     locationTbl, 
#'     wa_sub$longitude, 
#'     wa_sub$latitude, 
#'     distanceThreshold = 500
#'   )
#'   
#' locationData <- wa_sub$AQSID
#' 
#' locationTbl <- 
#'   table_updateColumn(locationTbl, "AQSID", locationID, locationData)
#' 
#' # Look at the data we attempted to merge
#' wa$AQSID[wa_indices]
#' 
#' # And two columns from the updated locationTbl
#' locationTbl_indices <- table_getRecordIndex(locationTbl, locationID)
#' locationTbl[locationTbl_indices, c("city", "AQSID")]
#' 
#' @seealso [table_addColumn()]
#' @seealso [table_removeColumn()]
#' @rdname table_updateColumn
#' @export 
#' @importFrom MazamaCoreUtils stopIfNull
#' @importFrom dplyr bind_rows

table_updateColumn <- function(
  locationTbl = NULL,
  columnName = NULL,
  locationID = NULL,
  locationData = NULL,
  verbose = TRUE
) {
  
  # ----- Validate parameters --------------------------------------------------
  
  MazamaLocationUtils::validateLocationTbl(locationTbl, locationOnly = FALSE)
  MazamaCoreUtils::stopIfNull(columnName)
  
  if ( !is.null(locationID) )
    MazamaCoreUtils::stopIfNull(locationData)
  
  if ( !is.null(locationData) )
    MazamaCoreUtils::stopIfNull(locationID)
  
  if ( !is.null(locationID) && !is.null(locationData) ) {
    if ( length(locationID) != length(locationData) ) {
      stop(sprintf(
        "locationID and locationData must have the same length"
      ))
    }
  }
  
  # Remove elements associated with missing locationIDs
  mask <- !is.na(locationID)
  locationID <- locationID[mask]
  locationData <- locationData[mask]
  
  # ----- Update column --------------------------------------------------------
  
  # Add it first if needed
  if ( !columnName %in% names(locationTbl) ) 
    locationTbl <- table_addColumn(locationTbl, columnName)
  
  if ( !is.null(locationData) ) {
    # Get the record indices to be updated
    recordIndex <- table_getRecordIndex(locationTbl, locationID)

    # Remove incoming locationData not associated with any record
    mask <- !is.na(recordIndex)
    recordIndex <- recordIndex[mask]
    locationData <- locationData[mask]
  
    # Updated record data
    locationTbl[[columnName]][recordIndex] <- locationData
  }
  
  # ----- Return ---------------------------------------------------------------
  
  return(locationTbl)
  
}

#'
#' @title Finds adjacent locations in a known locations table.
#' 
#' @description Calculate distances between all locations within a known
#' locations table and return a tibble containing all records that have an 
#' adjacent location separated by less than `distanceThreshold` meters.
#' The return tibble is ordered by separation distance.
#' 
#' It is useful when working with new metadata tables to identify adjacent
#' locations early on so that decisions can be made about the appropriateness
#' of the specified `distanceThreshold`.
#' 
#' @note The measure `"cheap"` may be used to speed things up depending on
#' the spatial scale being considered. Distances calculated with 
#' `measure = "cheap"` will vary by a few meters compared with those 
#' calculated using `measure = "geodesic"`.
#' 
#' @param locationTbl Tibble of known locations.
#' @param distanceThreshold Distance in meters.
#' @param measure One of "haversine" "vincenty", "geodesic", or "cheap" 
#' specifying desired method of geodesic distance calculation.
#' 
#' See [geodist::geodist()] for details.
#' 
#' @return Tibble of known locations separated by less than 
#' `distanceThreshold` meters.
#' 
#' @examples 
#' library(MazamaLocationUtils)
#' 
#' meta <- wa_airfire_meta
#' 
#' # Any locations closer than 2 km?
#' meta %>%
#'   table_findAdjacentLocations(distanceThreshold = 2000) %>%
#'   dplyr::select(AQSID, timezone)
#' 
#' # How about 4 km?
#' meta %>%
#'   table_findAdjacentLocations(distanceThreshold = 4000) %>%
#'   dplyr::select(AQSID, timezone)
#' 
#' @rdname table_findAdjacentLocations
#' @export 
#' @importFrom MazamaCoreUtils stopIfNull
#' @importFrom dplyr tibble filter all_of
#' @importFrom rlang .data
#' 
table_findAdjacentLocations <- function(
  locationTbl = NULL,
  distanceThreshold = NULL,
  measure = c("geodesic", "haversine", "vincenty", "cheap")
) {
  
  # ----- Validate parameters --------------------------------------------------
  
  MazamaLocationUtils::validateLocationTbl(locationTbl, locationOnly = TRUE)
  MazamaCoreUtils::stopIfNull(distanceThreshold)
  measure <- match.arg(measure)
  
  if ( !is.numeric(distanceThreshold) )
    stop("Parameter 'distanceThreshold' must be a numeric value.")
  
  # ----- Subset locationTbl ---------------------------------------------------
  
  adjacentDistanceTbl <- table_findAdjacentDistances(locationTbl, distanceThreshold, measure)
  
  # > head(adjacentDistanceTbl)
  # # A tibble: 6 × 3
  #    row1  row2 distance
  #   <int> <int>    <dbl>
  # 1   117   800    0.334
  # 2   282   652    0.444
  # 3   250   269    0.445
  # 4   117   795    0.445
  # 5    22   910    0.445
  # 6    22   518    0.445
  
  # Base R tricks to return unique set of indices in order of separation distance
  indices <- 
    adjacentDistanceTbl[,1:2] %>%
    as.matrix() %>%
    t() %>%
    as.numeric() %>%
    unique()
  
  adjacentLocationTbl <- locationTbl[indices,]
  
  # ----- Return ---------------------------------------------------------------
  
  return(adjacentLocationTbl)
  
}

#' @rdname location_getSingleAddress_TexasAM
#' @export 
#' 
#' @title Get an address from the Texas A&M reverse geocoding service
#' 
#' @param longitude Single longitude in decimal degrees E.
#' @param latitude Single latitude in decimal degrees N.
#' @param verbose Logical controlling the generation of progress messages.
#' @param apiKey Texas A&M Geocoding requires an API key. The first 2500 requests
#' are free.
#' 
#' @return Numeric elevation value.
#' 
#' @description Texas A&M APIs are used to determine the address associated with
#' the `longitude` and `latitude`.
#' 
#' @examples
#' \dontrun{
#' library(MazamaLocationUtils)
#' 
#' # Fail gracefully if any resources are not available
#' try({
#' 
#'   # Wenatchee
#'   longitude <- -122.47
#'   latitude <- 47.47
#'   apiKey <- YOUR_PERSONAL_API_KEY
#' 
#'   location_getSingleAddress_TexasAM(longitude, latitude, apiKey)
#'   
#' }, silent = FALSE)
#' }
#' 
#' @references \url{https://geoservices.tamu.edu/Services/ReverseGeocoding/WebService/v04_01/HTTP.aspx}
#' 
location_getSingleAddress_TexasAM <- function(
  longitude = NULL,
  latitude = NULL,
  apiKey = NULL,
  verbose = TRUE
) {
  
  # ----- Validate parameters --------------------------------------------------
  
  MazamaCoreUtils::stopIfNull(longitude)
  MazamaCoreUtils::stopIfNull(latitude)
  MazamaCoreUtils::stopIfNull(verbose)
  MazamaCoreUtils::stopIfNull(apiKey)
  
  validateLonLat(longitude, latitude)  
  
  # ----- Get Texas A&M address data -------------------------------------------
  
  # Create url
  url <- httr::parse_url("https://geoservices.tamu.edu/Services/ReverseGeocoding/WebService/v04_01/HTTP/default.aspx")
  
  url$query <- list(
    apikey = apiKey,
    version = "4.01",
    lat = latitude,
    lon = longitude,
    format = "json"
  )
  
  
  result <- try({
    # Get and parse the return
    r <- httr::GET(httr::build_url(url))
  })
  
  if ( "try-error" %in% result )
    stop(geterrmessage())
  
  if ( httr::http_error(r) ) {
    
    addressList <- list(
      houseNumber = as.character(NA),
      street = as.character(NA),
      city = as.character(NA),
      stateCode = as.character(NA),
      postalCode = as.character(NA),
      countryCode = as.character(NA),
      countryName = as.character(NA)
    ) 
    
    if ( verbose ) {
      warning(sprintf(
        "Texas A&M reverse geocoding service failed for URL %s", 
        httr::build_url(url)
      ))
    }
    
  } else {
    
    # Get the json return
    jsonString <- httr::content(r)
    
    # Parse the json
    returnObj <- as.list(jsonlite::fromJSON(jsonString,
                                            simplifyVector = TRUE,
                                            simplifyDataFrame = FALSE,
                                            simplifyMatrix = FALSE,
                                            flatten = FALSE))
    
    # Extract desired information
    locationInfo <- returnObj$StreetAddresses[[1]] 
    
    address <- 
      locationInfo$StreetAddress %>% 
      stringr::str_trim()
    
    # Parse the address string into houseNumber and street
    if ( locationInfo$StreetAddress == "" ) {
      houseNumber <- as.character(NA)
      street <- as.character(NA)
    } else {
      parts <- stringr::str_match(address, "(\\d+) (.+)")
      houseNumber <- parts[,2]
      street <- parts[,3]
    }
    
    if ( locationInfo$City == "" ) {
      locationInfo$City <- as.character(NA)
    }
    
    # Assemble the addressList
    addressList <- list(
      houseNumber = as.character(houseNumber),
      street = as.character(street),
      city = as.character(locationInfo$City),
      stateCode = toupper(locationInfo$State),
      postalCode = as.character(locationInfo$Zip),
      countryCode = "US",
      countryName = "United States of America" # Texas A&M ONLY works in US
    )
    
  }
  
  # Fill in the stateName
  suppressWarnings({
    addressList$stateName <- 
      MazamaSpatialUtils::codeToState(
        stateCodes = addressList$stateCode,
        countryCodes = addressList$countryCode,
        dataset = "NaturalEarthAdm1"
      )
  })
  
  # ----- Return ---------------------------------------------------------------
  
  return(addressList)
  
}

#'
#' @title Add missing core metadata columns to a known location table
#' 
#' @description An existing table will be amended to guarantee that it includes
#' the following core metadata columns.
#' 
#' \itemize{
#' \item{locationID}
#' \item{locationName}
#' \item{longitude}
#' \item{latitude}
#' \item{elevation}
#' \item{countryCode}
#' \item{stateCode}
#' \item{countyName}
#' \item{timezone}
#' \item{houseNumber}
#' \item{street}
#' \item{city}
#' \item{postalCode}
#' }
#' 
#' The `longitude` and `latitude` columns are required to exist in the
#' incoming tibble but all others are optional.
#' 
#' If any of these core metadata columns are found, they will be retained.
#' 
#' The `locationID` will be generated (anew if already found) from existing
#' longitude and latitude data.
#' 
#' Other core metadata columns will be filled with `NA` values of the 
#' proper type.
#' 
#' The result is a tibble with all of the core metadata columns. Theses columns
#' must then be filled in to create a usable "known locations" table.
#' 
#' @note No check is performed for overlapping locations. The returned tibble
#' has the structure of a "known locations" table and is a good starting place
#' for investigation. But further work is required to produce a valid table of
#' "known locations" associated with a specific spatial scale.
#' 
#' @param locationTbl Tibble of known locations. This input tibble need not be a 
#' standardized "known location" with all required columns. They will be added.
#' @param precision `precision` argument passed on to [location_createID()].
#' 
#' @return Tibble with the metadata columns required in a "known locations" table.
#' 
#' @rdname table_addCoreMetadata
#' @export 
#' @importFrom MazamaCoreUtils stopIfNull
#' @importFrom dplyr tibble filter all_of
#' @importFrom rlang .data
#' 
table_addCoreMetadata <- function(
  locationTbl = NULL,
  precision = 10
) {
  
  # ----- Validate parameters --------------------------------------------------
  
  MazamaLocationUtils::validateLocationTbl(locationTbl, locationOnly = TRUE)

  # ----- Create locationTbl ---------------------------------------------------
  
  incomingTblColumns <- names(locationTbl)
  
  # * locationID -----
  
  locationTbl$locationID <- location_createID(
    longitude = locationTbl$longitude,
    latitude = locationTbl$latitude,
    algorithm = "geohash",
    precision = precision
  )
  
  # * elevation -----
  
  if ( !"elevation" %in% incomingTblColumns ) {
    locationTbl$elevation <- as.numeric(NA)
  }
  
  # * countryCode -----
  
  if ( !"countryCode" %in% incomingTblColumns ) {
    locationTbl$countryCode <- as.character(NA)
  }
  
  # * stateCode -----
  
  if ( !"stateCode" %in% incomingTblColumns ) {
    locationTbl$stateCode <- as.character(NA)
  }
  
  # * locationName -----
  
  if ( !"locationName" %in% incomingTblColumns ) {
    locationTbl$locationName <- as.character(NA)
  }
  
  # * countyName -----
  
  if ( !"countyName" %in% incomingTblColumns ) {
    locationTbl$countyName <- as.character(NA)
  }
  
  # * timezone -----
  
  if ( !"timezone" %in% incomingTblColumns ) {
    locationTbl$timezone <- as.character(NA)
  }
  
  # * houseNumber -----
  
  if ( !"houseNumber" %in% incomingTblColumns ) {
    locationTbl$houseNumber <- as.character(NA)
  }
  
  # * street -----
  
  if ( !"street" %in% incomingTblColumns ) {
    locationTbl$street <- as.character(NA)
  }
  
  # * city -----
  
  if ( !"city" %in% incomingTblColumns ) {
    locationTbl$city <- as.character(NA)
  }
  
  # * postalCode -----
  
  if ( !"postalCode" %in% incomingTblColumns ) {
    locationTbl$postalCode <- as.character(NA)
  }
  
  # ----- Reorganize locationTbl -----------------------------------------------
  
  requiredColumns <- c(
    "locationID", "locationName", 
    "longitude", "latitude", "elevation", 
    "countryCode", "stateCode", "countyName", "timezone", 
    "houseNumber", "street", "city", "postalCode"
  )
  
  extraColumns <- setdiff(incomingTblColumns, requiredColumns)
  
  # This is the preferred order
  allColumns <- c(requiredColumns, extraColumns)
  
  # Reorder and convert to tibble
  locationTbl <- 
    dplyr::select(locationTbl, dplyr::all_of(allColumns)) %>%
    dplyr::as_tibble()
  
  # ----- Return ---------------------------------------------------------------
  
  return(locationTbl)
  
}

#' @title Add a single new known location record to a table
#' 
#' @description Incoming `longitude` and `latitude` values are compared 
#' against the incoming `locationTbl` to see if they are already within
#' `distanceThreshold` meters of an existing entry.  A new record is created for
#' if the location is not already found in `locationTbl`.
#' 
#' @param locationTbl Tibble of known locations.
#' @param longitude Single longitude in decimal degrees E.
#' @param latitude Single latitude in decimal degrees N.
#' @param distanceThreshold Distance in meters.
#' @param stateDataset Name of spatial dataset to use for determining state
#' codes, Default: "NaturalEarthAdm1".
#' @param elevationService Name of the elevation service to use for determining
#' the elevation. Default: NULL. Accepted values: "usgs".
#' @param addressService Name of the address service to use for determining
#' the street address. Default: NULL. Accepted values: "photon".
#' @param verbose Logical controlling the generation of progress messages.
#' 
#' @return Updated tibble of known locations.
#' 
#' @examples
#' \donttest{
#' library(MazamaLocationUtils)
#' 
#' # Fail gracefully if any resources are not available
#' try({
#' 
#'   # Set up standard directories and spatial data
#'   spatialDataDir <- tempdir() # typically "~/Data/Spatial"
#'   initializeMazamaSpatialUtils(spatialDataDir)
#' 
#'   locationTbl <- get(data("wa_monitors_500"))
#'   
#'   nrow(locationTbl)
#' 
#'   # Coulee City, WA
#'   lon <- -119.290904
#'   lat <- 47.611942
#' 
#'   locationTbl <- 
#'     locationTbl %>%
#'     table_addSingleLocation(lon, lat, distanceThreshold = 500)
#'     
#'   nrow(locationTbl)
#'   
#' }, silent = FALSE)
#' }
#' 
#' @seealso [table_addLocation()]
#' @seealso [table_removeRecord()]
#' @seealso [table_updateSingleRecord()]
#' @rdname table_addSingleLocation
#' @export 
#' @importFrom MazamaCoreUtils stopIfNull
#' @importFrom dplyr bind_rows
#' 
table_addSingleLocation <- function(
  locationTbl = NULL,
  longitude = NULL,
  latitude = NULL,
  distanceThreshold = NULL,
  stateDataset = "NaturalEarthAdm1",
  elevationService = NULL,
  addressService = NULL,
  verbose = TRUE
) {
  
  validateMazamaSpatialUtils()
  
  # ----- Validate parameters --------------------------------------------------
  
  MazamaLocationUtils::validateLocationTbl(locationTbl, locationOnly = FALSE)
  MazamaCoreUtils::stopIfNull(distanceThreshold)
  MazamaCoreUtils::stopIfNull(stateDataset)
  
  if ( length(longitude) > 1 || length(latitude) > 1 ) {
    stop(paste0(
      "Please use the plural version of the funcion for adding multiple locations.\n",
      "  table_addLocation(...)\n"
    ))
  }

  MazamaLocationUtils::validateLonLat(longitude, latitude)
  
  if ( !exists(stateDataset) ) {
    stop(paste0(
      "You must load \"stateDataset\" with: \n",
      "  loadSpatialData(\"", stateDataset, "\")\n"
    ))
  }
  
  # ----- Check for existing location ------------------------------------------
  
  locationID <- table_getLocationID(locationTbl, longitude, latitude, distanceThreshold)
  
  if ( !is.na(locationID) ) {
    stop(sprintf(
      "The known location %s already exists < %d meters from the requested location.",
      locationID, distanceThreshold
    ))
  }
  
  # ----- Add new record -------------------------------------------------------

  singleRecordTbl <- location_initialize(
    longitude = longitude,
    latitude = latitude,
    stateDataset = stateDataset,
    elevationService = elevationService,
    addressService = addressService,
    verbose = verbose
  )
    
  additionalNames <- setdiff( names(locationTbl), names(singleRecordTbl))
  
  for ( name in additionalNames ) {
    
    if ( verbose ) 
      message(sprintf("Using NA in place of actual metadata for %s", name))
    
    singleRecordTbl[[name]] <- NA
    
  }
  
  locationTbl <- dplyr::bind_rows(locationTbl, singleRecordTbl)
  
  # ----- Return ---------------------------------------------------------------
  
  return(locationTbl)
  
}

#' @export
#'
#' @title Add distance-clustering information to a dataframe
#'
#' @param tbl Tibble with geolocation information.
#' @param clusterDiameter Diameter in meters used to determine the number of
#' clusters (see description).
#' @param lonVar Name of longitude variable in the incoming tibble.
#' @param latVar Name of the latitude variable in the incoming tibble.
#' @param maxClusters Maximum number of clusters to try.
#'
#' @description Distance clustering is used to identify unique deployments of a
#' sensor in an environmental monitoring field study. GPS-reported locations can
#' be jittery and result in a sensor self-reporting from a cluster of nearby locations.
#' Clustering helps resolve this by assigning a single location to the cluster.
#'
#' Standard `kmeans` clustering does not work well when clusters can have 
#' widely differing numbers of members. A much better result is acheived with
#' the Partitioning Around Medoids method available in `cluster::pam()`.
#'
#' The value of `clusterDiameter` is compared with the output of
#' `cluster::pam(...)$clusinfo[,'av_diss']` to determine the number of clusters.
#' 
#' @note In most applications, the [table_addClustering()] function should be 
#' used as it implements two-stage clustering using `clusterbyDistance()`.
#'
#' @return Input tibble with additional columns: `clusterLon, clusterLat, clusterID`.
#'
#' @references [When k-means clustering fails](https://working-with-data.mazamascience.com/2021/07/15/when-k-means-clustering-fails/)
#' 
#' @seealso [table_removeRecord()]
#'
#' @examples
#' library(MazamaLocationUtils)
#'
#' # Fremont, Seattle 47.6504, -122.3509
#' # Magnolia, Seattle 47.6403, -122.3997
#' # Downtown Seattle 47.6055, -122.3370
#' 
#' fremont_x <- jitter(rep(-122.3509, 10), .0005)
#' fremont_y <- jitter(rep(47.6504, 10), .0005)
#' 
#' magnolia_x <- jitter(rep(-122.3997, 8), .0005)
#' magnolia_y <- jitter(rep(47.6403, 8), .0005)
#' 
#' downtown_x <- jitter(rep(-122.3370, 3), .0005)
#' downtown_y <- jitter(rep(47.6055, 3), .0005)
#' 
#' # Apply clustering
#' tbl <-
#'   dplyr::tibble(
#'     longitude = c(fremont_x, magnolia_x, downtown_x),
#'     latitude = c(fremont_y, magnolia_y, downtown_y)
#'   ) %>%
#'   clusterByDistance(
#'     clusterDiameter = 1000
#'   )
#' 
#' plot(tbl$longitude, tbl$latitude, pch = tbl$clusterID)

clusterByDistance <- function(
  tbl,
  clusterDiameter = 1000,
  lonVar = "longitude",
  latVar = "latitude",
  maxClusters = 50
) {

  # ----- Validate parameters --------------------------------------------------

  MazamaCoreUtils::stopIfNull(tbl)
  MazamaCoreUtils::stopIfNull(clusterDiameter)
  MazamaCoreUtils::stopIfNull(lonVar)
  MazamaCoreUtils::stopIfNull(latVar)
  MazamaCoreUtils::stopIfNull(maxClusters)

  # Sanity check -- row count
  if ( nrow(tbl) == 0 ) {
    stop(paste0("Unable to perform clustering: 'tbl' is empty"))
  }

  # Sanity check -- names
  if ( !lonVar %in% names(tbl) ) {
    stop(paste0("Longitudes could not be found.  Did you specify the lonVar argument?"))
  }
  if ( !latVar %in% names(tbl) ) {
    stop(paste0("Latitudes could not be found.  Did you specify the latVar argument?"))
  }

  # If we only have a single row, return immediately
  # NOTE:  Use as.numeric() to remove any names associated with these vectors
  if ( nrow(tbl) == 1 ) {
    tbl$clusterLon <- as.numeric(tbl[[lonVar]][1])
    tbl$clusterLat <- as.numeric(tbl[[latVar]][1])
    tbl$clusterID <- 1
    return(tbl)
  }

  # ----- Clustering -----------------------------------------------------------

  # NOTE:  A monitor will be moved around from time to time, sometimes across the country
  # NOTE:  and sometimes across the street.  We need to assign unique identifiers to each
  # NOTE:  new "deployment" but not when the monitor is moved only a short distance.
  # NOTE:
  # NOTE:  We use clustering to find an appropriate number of unique "deployments".
  # NOTE:  The sensitivity of this algorithm can be adjused with the clusterDiameter argument.
  # NOTE:
  # NOTE:  Standard kmeans clustering does not work well when clusters can have widely
  # NOTE:  differing numbers of members. A much better result is acheived with
  # NOTE:  the Partitioning Around Medoids method available in cluster::pam().
  # NOTE:
  # NOTE:  Try the following example:
  # NOTE:    x <- y <- c(rep(0,3), rep(1,3), rep(10,20), rep(11,20), rep(100,50), rep(101,50))
  # NOTE:    m <- cbind(x,y)
  # NOTE:    layout(matrix(seq(2)))
  # NOTE:    plot(x, y, pch = as.character( stats::kmeans(m,3)$cluster ))
  # NOTE:    plot(x, y, pch = as.character( cluster::pam(m,3)$cluster ))
  # NOTE:    plot(x, y, pch = as.character( cluster::clara(m,3)$cluster ))
  # NOTE:
  # NOTE:  Run the plots a few times and you will see that kmeans clustering sometimes
  # NOTE:  gets it wrong.

  # NOTE:  We need to use cluster::clara when we get above ~2K points.
  # NOTE:  For this reason we need to use clusinfo[,'max_diss'] instead
  # NOTE:  of clusinfo[,'diameter'] as the latter is only provided by
  # NOTE:  cluster::pam and not cluster::clara.
  # NOTE:  (Is there really any difference between 'max_diss' and 'distance'?)

  # Perform clustering
  for ( clusterCount in 1:maxClusters ) {
    
    if ( nrow(tbl) <= clusterCount ) {
      # NOTE:  k = positive integer specifying the number of clusters, less than the number of observations.
      clusterObj <- cluster::pam(tbl[,c(lonVar,latVar)], nrow(tbl) - 1)
    } else if ( nrow(tbl) < 2000 ) {
      clusterObj <- cluster::pam(tbl[,c(lonVar,latVar)], clusterCount)
    } else {
      clusterObj <- cluster::clara(tbl[,c(lonVar,latVar)], clusterCount, samples = 50)
    }
    clusterLats <- clusterObj$medoids[,latVar]
    diameters <- 2 * clusterObj$clusinfo[,'max_diss'] # decimal degrees
    # NOTE:  We don't know whether distance is pure NS, EW or some combination
    # NOTE:  so we can't accurately convert to meters. We approximate by
    # NOTE:  assuming a 50-50 split and using 111,320 meters/degree at the equator.
    radianClusterLats <- clusterLats * pi/180
    meters <- diameters * (1 + cos(radianClusterLats))/2 * 111320
    if ( max(meters) < clusterDiameter ) break
    
  }
  
  # NOTE:  If we only had a very small table with widely separate locations,
  # NOTE:  we can end up never satisfying max(meters) < clusterDiameter. In this
  # NOTE:  case, every location represents a separate cluster
  if ( nrow(tbl) <= clusterCount ) {
    for ( i in 1:seq_len(nrow(tbl)) ) {
      tbl$clusterLon <- as.numeric(tbl[[lonVar]][i])
      tbl$clusterLat <- as.numeric(tbl[[latVar]][i])
      tbl$clusterID <- i
    }
    return(tbl)
  }

  # Create the vector of deployment identifiers
  if ( nrow(tbl) < 2000 ) {
    clusterObj <- cluster::pam(tbl[,c(lonVar,latVar)], clusterCount)
  } else {
    clusterObj <- cluster::clara(tbl[,c(lonVar,latVar)], clusterCount, samples = 50)
  }

  # Add cluster lons and lats to the tibble
  # NOTE:  Use as.numeric() to remove any names associated with these vectors
  tbl$clusterLon <- as.numeric(clusterObj$medoids[,lonVar][clusterObj$clustering])
  tbl$clusterLat <- as.numeric(clusterObj$medoids[,latVar][clusterObj$clustering])
  tbl$clusterID <- as.character(clusterObj$clustering)

  # ----- Return ---------------------------------------------------------------

  return(tbl)

}

# ===== DEBUG ==================================================================

if ( FALSE ) {

  library(MazamaCoreUtils)
  logger.setup()
  logger.setLevel(TRACE)

  library(AirMonitorIngest)

  tbl <-
    wrcc_downloadData(20150701, 20150930, unitID = 'SM16') %>%
    wrcc_parseData() %>%
    wrcc_qualityControl()

  clusterDiameter = 1000
  lonVar = "GPSLon"
  latVar = "GPSLat"
  maxClusters = 50




  newTbl <- clusterByDistance(
    tbl,
    clusterDiameter = clusterDiameter,
    lonVar = lonVar,
    latVar = latVar,
    maxClusters = maxClusters
  )


}

context("clusterByDistance")

test_that("clustering works for small tables", {
  
  # ?cluster::pam says: 
  #   maxClusters = positive integer specifying the number of clusters, less than the number of observations.
  
  # Table with only 2 rows
  tbl <- dplyr::tibble(longitude = c(-127.1, -127.3, -127.5), latitude = c(45.1, 45.3, 45.5))
  
  expect_no_error({
    clusteredTbl <- 
      clusterByDistance(
        tbl,
        clusterDiameter = 1000,
        lonVar = "longitude",
        latVar = "latitude",
        maxClusters = 50
      )
  })
  
})

#' @title Return indexes of known location records
#' @description Returns a vector of `locationTbl` row indexes for the 
#' locations associated with each `locationID`.
#' @param locationTbl Tibble of known locations.
#' @param locationID Vector of `locationID` strings.
#' @param verbose Logical controlling the generation of progress messages.
#' @return Vector of `locationTbl` row indexes.
#' @examples
#' library(MazamaLocationUtils)
#' 
#' locationTbl <- get(data("wa_monitors_500"))
#' 
#' # Wenatchee
#' lon <- -120.325278
#' lat <- 47.423333
#' 
#' # Get the locationID first
#' locationID <- table_getLocationID(locationTbl, lon, lat, distanceThreshold = 5000)
#' 
#' # Now find the row associated with this ID
#' recordIndex <- table_getRecordIndex(locationTbl, locationID)
#' 
#' str(locationTbl[recordIndex,])
#' 
#' @rdname table_getRecordIndex
#' @export 
#' @importFrom MazamaCoreUtils stopIfNull
#' @importFrom dplyr bind_rows
table_getRecordIndex <- function(
  locationTbl = NULL,
  locationID = NULL,
  verbose = TRUE
) {
  
  # ----- Validate parameters --------------------------------------------------
  
  MazamaLocationUtils::validateLocationTbl(locationTbl, locationOnly = FALSE)
  MazamaCoreUtils::stopIfNull(locationID)
  
  # ----- Find recordIndexes ---------------------------------------------------
  
  recordIndex <- rep(as.numeric(NA), length(locationID))
    
  for ( index in seq_along(locationID) ) {
    
    if ( !is.na(locationID[index]) ) {
      if ( any(locationTbl$locationID == locationID[index], na.rm = TRUE) ) {
        recordIndex[index] <- which(locationTbl$locationID == locationID[index])
      }
    }
    
  }
  
  # ----- Return ---------------------------------------------------------------
  
  return(recordIndex)
  
}

context("table_getRecordIndex")

test_that("correct index is returned", {
  locationTbl <- get(data("wa_monitors_500"))
  
  indices <- c(5,10,15)
  locationID <- wa_monitors_500$locationID[indices]

  testIndices <- table_getRecordIndex(locationTbl, locationID)
  expect_equal(indices, testIndices)
})

test_that("NA is returned", {
  locationTbl <- get(data("wa_monitors_500"))
  
  indices <- c(5,10,15)
  locationID <- c(wa_monitors_500$locationID[indices], "Rumplestiltskin")
  
  testIndices <- table_getRecordIndex(locationTbl, locationID)
  expect_equal(c(indices, NA), testIndices)
})

#'
#' @title Add a new column of metadata to a table
#' 
#' @description A new metadata column is added to the `locationTbl`. For
#' matching `locationID` records, the associated `locationData` is
#' inserted. Otherwise, the new column will be initialized with `NA`.
#' 
#' @param locationTbl Tibble of known locations.
#' @param columnName Name to use for the new column.
#' @param locationID Vector of `locationID` strings.
#' @param locationData Vector of data to used at matching records.
#' @param verbose Logical controlling the generation of progress messages.
#' 
#' @return Updated tibble of known locations.
#' 
#' @examples
#' library(MazamaLocationUtils)
#' 
#' # Starting table
#' locationTbl <- get(data("wa_monitors_500")) 
#' names(locationTbl)
#' 
#' # Add an empty column
#' locationTbl <-
#'   locationTbl %>%
#'   table_addColumn("AQSID")
#'   
#' names(locationTbl)
#' 
#' @seealso [table_removeColumn()]
#' @seealso [table_updateColumn()]
#' @rdname table_addColumn
#' @export 
#' @importFrom MazamaCoreUtils stopIfNull
#' @importFrom methods as
#' 
table_addColumn <- function(
  locationTbl = NULL,
  columnName = NULL,
  locationID = NULL,
  locationData = NULL,
  verbose = TRUE
) {
  
  # ----- Validate parameters --------------------------------------------------
  
  MazamaLocationUtils::validateLocationTbl(locationTbl, locationOnly = FALSE)
  MazamaCoreUtils::stopIfNull(columnName)
  
  if ( !is.null(locationID) )
    MazamaCoreUtils::stopIfNull(locationData)
  
  if ( !is.null(locationData) )
    MazamaCoreUtils::stopIfNull(locationID)
  
  if ( !is.null(locationID) && !is.null(locationData) ) {
    if ( length(locationID) != length(locationData) ) {
      stop(sprintf(
        "locationID and locationData must have the same length"
      ))
    }
    dataClass <- class(locationData)
  } else {
    dataClass <- "character"
  }
  
  # ----- Add new column -------------------------------------------------------
  
  locationTbl[[columnName]] <- methods::as(rep(NA, nrow(locationTbl)), dataClass)
  
  # Update if any data are passed in
  if ( !is.null(locationID) ) {
    locationTbl <- 
      table_updateColumn(locationTbl, columnName, locationID, locationData)
  }

  # ----- Return ---------------------------------------------------------------
  
  return(locationTbl)
  
}

context("table_addColumn")

test_that("table_addColumn() works", {
  locationTbl <- get(data("wa_monitors_500"))
  testTbl <- table_addColumn(locationTbl, "siteName")
  expect_equal(c(names(locationTbl),"siteName"), names(testTbl))
})

#'
#' @title Create one or more unique locationIDs
#' @description A unique locationID is created for each incoming
#' `longitude` and `latitude`. 
#' 
#' See [MazamaCoreUtils::createLocationID()] for details.
#' 
#' At `precision = 10`, this results in a maximum error of 0.6 meters which 
#' is more than precise enough for environmental monitoring studies making use
#' of this package.
#' 
#' An excellent way to become familiar with geohash is through the 
#' [GeoHash Explorer](https://geohash.softeng.co).
#' 
#' @note The `"digest"` algorithm is deprecated but provided for backwards 
#' compatibility with databases that were built using locationIDs generated
#' with this algorithm.
#' 
#' @param longitude Vector of longitudes in decimal degrees E.
#' @param latitude Vector of latitudes in decimal degrees N.
#' @param algorithm Algorithm to use -- either `"geohash"` or `"digest"`.
#' @param precision `precision` argument used when encoding with `"geohash"`.
#' 
#' @return Vector of character locationIDs.
#' 
#' @examples
#' library(MazamaLocationUtils)
#' 
#' # Wenatchee
#' lon <- -120.325278
#' lat <- 47.423333
#' locationID <- location_createID(lon, lat)
#' print(locationID)
#' 
#' location_createID(lon, lat, algorithm = "geohash")
#' location_createID(lon, lat, algorithm = "geohash", precision = 7)
#' 
#' @references \url{https://en.wikipedia.org/wiki/Decimal_degrees}
#' @references \url{https://www.johndcook.com/blog/2017/01/10/probability-of-secure-hash-collisions/}
#' @rdname location_createID
#' @export 
#' 
location_createID <- function(
  longitude = NULL,
  latitude = NULL,
  algorithm = c("geohash", "digest"),
  precision = 10
) {
  
  algorithm <- match.arg(algorithm)
  
  returnVal <- 
    MazamaCoreUtils::createLocationID(
      longitude = longitude, 
      latitude = latitude, 
      algorithm = algorithm,
      precision = precision,
      invalidID = as.character(NA)
    )
  
  return(returnVal)
  
}

context("table_initializeExisting")

test_that("existing locationID is ignored", {
  
  skip_on_cran()
  skip_on_travis()
  
  initializeMazamaSpatialUtils()
  data("wa_airfire_meta")
  
  meta <-
    dplyr::as_tibble(wa_airfire_meta[1:3,]) %>%
    dplyr::mutate(
      locationID = location_createID(
        .data$longitude, 
        .data$latitude,
        algorithm = "geohash",
        precision = 10
      )
    )
  
  locationTbl <- table_initializeExisting(meta, distanceThreshold = 500)
  
  expect_equal(nrow(locationTbl), 3)
  
})

#' @title Validate proper setup of MazamaSpatialUtils
#' @description The \pkg{MazamaSpatialUtils} package mus be properly installed
#' and initialized before using functions from the \pkg{MazamaLocationUtils} 
#' package. This function tests for this.
#' @return Invisibly returns `TRUE` if no error message has been generated.
#' @rdname validateMazamaSpatialUtils
#' @export 
#' 
validateMazamaSpatialUtils <- function() {
  
  if ( !exists("EEZCountries") ||
       !exists("OSMTimezones") ||
       !exists("NaturalEarthAdm1") ||
       !exists("USCensusCounties") ) {
    
    stop(paste0(
      "\n\nYou must have the MazamaSpatialUtils package ",
      "as well as core datasets installed.\n\n",
      "Install core datasets with:\n\n",
      "  MazamaSpatialUtils::setSpatialDataDir(\"YOUR_DATA_DIR\")\n",
      "  MazamaSpatialUtils::installSpatialData()\n\n",
      "Once installed, initialize spatial data with:\n\n",
      "  MazamaSpatialUtils::setSpatialDataDir(\"YOUR_DATA_DIR\")\n",
      "  MazamaSpatialUtils::loadSpatialData(\"EEZCountries.rda\")\n",
      "  MazamaSpatialUtils::loadSpatialData(\"OSMTimezones.rda\")\n",
      "  MazamaSpatialUtils::loadSpatialData(\"NaturalEarthAdm1.rda\")\n",
      "  MazamaSpatialUtils::loadSpatialData(\"USCensusCounties.rda\")\n"
    ))
    
  }
  
  return(invisible(TRUE))
  
}


#' @title Validate longitude and latitude vectors
#' @description
#' See [MazamaCoreUtils::validateLonsLats()] for details.
#'
#' @name validateLonsLats
#' @rdname validateLonsLats
#' @keywords internal
#' @export
#' @importFrom MazamaCoreUtils validateLonsLats
#' @usage validateLonsLats(longitude = NULL, latitude = NULL, na.rm = FALSE)
NULL


#' @title Validate longitude and latitude values
#' @description
#' See [MazamaCoreUtils::validateLonLat()] for details.
#'
#' @name validateLonLat
#' @rdname validateLonLat
#' @keywords internal
#' @export
#' @importFrom MazamaCoreUtils validateLonLat
#' @usage validateLonLat(longitude = NULL, latitude = NULL)
NULL


#' @title Validate a location table
#' @description Ensures that the incoming table has numeric `longitude` and
#' `latitude` columns.
#' @param locationTbl Tibble of known locations.
#' @param locationOnly Logical specifying whether to check for all standard
#' columns.
#' @return Invisibly returns `TRUE` if no error message has been generated.
#' @rdname validateLocationTbl
#' @export 
#' 
validateLocationTbl <- function(
  locationTbl = NULL,
  locationOnly = TRUE
) {
  
  MazamaCoreUtils::stopIfNull(locationTbl)
  
  if ( !is.logical(locationOnly) ) locationOnly <- TRUE
  
  if ( !"data.frame" %in% class(locationTbl) )
    stop("Parameter 'locationTbl' is not of class \"data.frame\".")
  
  if ( !"longitude" %in% names(locationTbl) )
    stop("Parameter 'locationTbl' does not have a 'longitude' column.")
  
  if ( !"latitude" %in% names(locationTbl) )
    stop("Parameter 'locationTbl' does not have a 'latitude' column.")
  
  if ( !is.numeric(locationTbl$longitude) )
    stop("'locationTbl$longitude' is not numeric. Please ensure that decimal longitudes are used.")
  
  if ( !is.numeric(locationTbl$latitude) )
    stop("'locationTbl$latitude' is not numeric. Please ensure that decimal latitudes are used.")
  
  if ( !locationOnly ) {
    
    missingNames <- 
      setdiff(MazamaLocationUtils::coreMetadataNames, names(locationTbl))
    
    if ( length(missingNames) > 0 ) {
      missingNamesString <- paste0(missingNames, collapse = ", ")
      stop(sprintf("'locationTbl' is missing '%s'", missingNamesString))
    }
 
    if ( any(duplicated(locationTbl$locationID)) ) 
      stop("'locationTbl$locationID' is not unique.")
    
  }
  
  return(invisible(TRUE))
  
}

#' @rdname table_filterByDistance
#' @export
#' @importFrom MazamaCoreUtils stopIfNull
#' 
#' @title Return known locations near a target location
#' 
#' @param locationTbl Tibble of known locations.
#' @param longitude Target longitude in decimal degrees E.
#' @param latitude Target latitude in decimal degrees N.
#' @param distanceThreshold Distance in meters.
#' @param measure One of "haversine" "vincenty", "geodesic", or "cheap" 
#' specifying desired method of geodesic distance calculation. 
#' 
#' @return Tibble of known locations.
#' 
#' @description Returns a tibble of the known locations from `locationTbl` 
#' that are within `distanceThreshold` meters of the target location 
#' specified by `longitude` and `latitude`. 
#' 
#' @note Only a single target location is allowed.
#' 
#' @examples
#' library(MazamaLocationUtils)
#' 
#' locationTbl <- get(data("wa_monitors_500"))
#' 
#' # Too small a distanceThreshold will not find a match
#' locationTbl %>%
#'   table_filterByDistance(
#'     longitude = -117.3647, 
#'     latitude = 47.6725, 
#'     distanceThreshold = 10
#'   ) %>% 
#'   dplyr::glimpse()
#' 
#' # Expanding the distanceThreshold will find several
#' locationTbl %>%
#'   table_filterByDistance(
#'     longitude = -117.3647, 
#'     latitude = 47.6725, 
#'     distanceThreshold = 10000
#'   ) %>%
#'   dplyr::glimpse()

table_filterByDistance <- function(
  locationTbl = NULL,
  longitude = NULL,
  latitude = NULL,
  distanceThreshold = NULL,
  measure = c("geodesic", "haversine", "vincenty", "cheap")
) {

  # ----- Validate parameters --------------------------------------------------

  MazamaLocationUtils::validateLocationTbl(locationTbl, locationOnly = TRUE)
  MazamaLocationUtils::validateLonLat(longitude, latitude)
  MazamaCoreUtils::stopIfNull(distanceThreshold)
  measure <- match.arg(measure)
  
  distanceThreshold <- round(distanceThreshold)

  # ----- Calculate distances --------------------------------------------------
  
  distanceMatrix <-
    geodist::geodist(
      y = cbind(
        "x" = longitude,
        "y" = latitude
      ),
      x = cbind(
        "x" = locationTbl$longitude,
        "y" = locationTbl$latitude
      ),
      paired = FALSE,
      sequential = FALSE,
      pad = FALSE,
      measure = measure
    )
  
  # NOTE:  distanceMatrix is nrow(locationTbl) X 1
  
  distance <- distanceMatrix[,1]
  names(distance) <- locationTbl$locationID
  
  distance <- distance[distance <= distanceThreshold]
  
  if ( length(distance) == 0 ) {
    ids <- "DON'T FIND ME"
  } else {
    ids <-
      distance %>%
      sort() %>%
      names()
  }
  
  # ----- Subset ---------------------------------------------------------------

  subsetTbl <- 
    locationTbl %>%
    dplyr::filter(.data$locationID %in% ids)
  
  # ----- Return ---------------------------------------------------------------

  return(subsetTbl)

}

#' @title Converts an existing table into a known location table
#' 
#' @description An existing table may have much of the data that is needed
#' for a known location table. This function accepts an incoming table and
#' searches for required columns:
#' 
#' \itemize{
#' \item{locationID}
#' \item{locationName}
#' \item{longitude}
#' \item{latitude}
#' \item{elevation}
#' \item{countryCode}
#' \item{stateCode}
#' \item{countyName}
#' \item{timezone}
#' \item{houseNumber}
#' \item{street}
#' \item{city}
#' \item{postalCode}
#' }
#' 
#' The `longitude` and `latitude` columns are required but all others
#' are optional.
#' 
#' If any of these optional columns are found, they will be used and the often 
#' slow and sometimes slightly inaccurate steps to generate that information
#' will be skipped for locations that have non-missing data. Any additional 
#' columns of information that are 
#' not part of the required core metadata will be retained.
#' 
#' This method skips the assignment of columns like `elevation` and all
#' address related fields that require web service requests.
#' 
#' Compared to initializing a brand new table and populating it one record at a
#' time, this is a much faster way of creating a known location table from a
#' pre-existing table of metadata.
#' 
#' @note The measure `"cheap"` may be used to speed things up depending on
#' the spatial scale being considered. Distances calculated with 
#' `measure = "cheap"` will vary by a few meters compared with those 
#' calculated using `measure = "geodesic"`.
#' 
#' @param locationTbl Tibble of known locations. This input tibble need not be a 
#' standardized "known location" table with all required columns. Missing 
#' columns will be added.
#' @param stateDataset Name of spatial dataset to use for determining state
#' codes, Default: 'NaturalEarthAdm1'
#' @param countryCodes Vector of country codes used to optimize spatial
#' searching. (See ?MazamaSpatialUtils::getStateCode())
#' @param distanceThreshold Distance in meters. 
#' @param measure One of "haversine" "vincenty", "geodesic", or "cheap" 
#' specifying desired method of geodesic distance calculation. See 
#' `?geodist::geodist`.
#' @param precision `precision` argument passed on to [location_createID()].
#' @param verbose Logical controlling the generation of progress messages.
#' 
#' @return Known location tibble with the specified metadata columns. Any 
#' locations whose circles (as defined by `distanceThreshold`) overlap will
#' generate warning messages. 
#' 
#' It is incumbent upon the user to address overlapping locations by one of:
#' 
#' \enumerate{
#' \item{reduce the distanceThreshold until no overlaps occur}
#' \item{assign one of the overlapping locations to the other location}
#' }
#' 
#' @rdname table_initializeExisting
#' @export 
#' @importFrom MazamaCoreUtils stopIfNull
#' @importFrom dplyr tibble filter all_of
#' @importFrom rlang .data
#' 
table_initializeExisting <- function(
  locationTbl = NULL,
  stateDataset = "NaturalEarthAdm1",
  countryCodes = NULL,
  distanceThreshold = NULL,
  measure = c("geodesic", "haversine", "vincenty", "cheap"),
  precision = 10,
  verbose = TRUE
) {
  
  validateMazamaSpatialUtils()
  
  # ----- Validate parameters --------------------------------------------------
  
  MazamaLocationUtils::validateLocationTbl(locationTbl, locationOnly = TRUE)
  MazamaCoreUtils::stopIfNull(distanceThreshold)
  measure <- match.arg(measure)
  
  if ( !exists(stateDataset) ) {
    stop(paste0(
      "You must load \"stateDataset\" with: \n",
      "  loadSpatialData(\"", stateDataset, "\")\n"
    ))
  }
  
  if ( !is.numeric(distanceThreshold) )
    stop("Parameter 'distanceThreshold' must be a numeric value.")
  
  # ----- Create locationTbl ---------------------------------------------------
  
  locationTbl <- table_addCoreMetadata(locationTbl)
  
  # * locationID -----
  
  # locationID should have been added by table_add
  if ( anyNA(locationTbl$locationID) ) {
    locationTbl$locationID <- location_createID(
      longitude = locationTbl$longitude,
      latitude = locationTbl$latitude,
      algorithm = "geohash",
      precision = precision
    )
  }
  
  # * elevation -----
  
  # Slow web service so skip for now
  
  # * countryCode -----
  
  tbl_1 <- dplyr::filter(locationTbl, !is.na(.data$countryCode))
  tbl_2 <- dplyr::filter(locationTbl, is.na(.data$countryCode))
  
  if ( nrow(tbl_2) > 0 ) {
    
    if ( verbose ) 
      message(sprintf("Creating countryCodes for %d locations ...", nrow(tbl_2)))
    
    tbl_2$countryCode <- MazamaSpatialUtils::getCountryCode(
      lon = tbl_2$longitude,
      lat = tbl_2$latitude,
      dataset = "EEZCountries",
      countryCodes = countryCodes,
      useBuffering = FALSE
    )
    
    locationTbl <- dplyr::bind_rows(tbl_1, tbl_2)
    
  }
  
  # * stateCode -----
  
  tbl_1 <- dplyr::filter(locationTbl, !is.na(.data$stateCode))
  tbl_2 <- dplyr::filter(locationTbl, is.na(.data$stateCode))
  
  if ( nrow(tbl_2) > 0 ) {
    
    if ( verbose ) 
      message(sprintf("Creating stateCodes for %d locations ...", nrow(tbl_2)))
    
    tbl_2$stateCode <- MazamaSpatialUtils::getStateCode(
      lon = tbl_2$longitude,
      lat = tbl_2$latitude,
      dataset = stateDataset,
      countryCodes = countryCodes,
      useBuffering = TRUE
    )
    
    locationTbl <- dplyr::bind_rows(tbl_1, tbl_2)
  
  }
  
  # * locationName -----
  
  # NOTE:  The default locationName is intended to give folks a more memorable
  # NOTE:  handel than the locationID but is not guaranteed to be unique. It is 
  # NOTE:  expected that users will add their own, more relevant names 
  # NOTE:  appropriate for the community of practice using a particular
  # NOTE:  collectionName of known locations.
  
  tbl_1 <- dplyr::filter(locationTbl, !is.na(.data$locationName))
  tbl_2 <- dplyr::filter(locationTbl, is.na(.data$locationName))
  
  if ( nrow(tbl_2) > 0 ) {
    
    if ( verbose ) 
      message(sprintf("Creating locationNames for %d locations ...", nrow(tbl_2)))
    
    tbl_2$locationName <- paste0(
      tolower(tbl_2$countryCode), ".",
      tolower(tbl_2$stateCode), "_",
      stringr::str_sub(tbl_2$locationID, 1, 6)
    )
    
    locationTbl <- dplyr::bind_rows(tbl_1, tbl_2)
    
  }
  
  # * countyName -----
  
  # Slow so skip for now
  
  # tbl_1 <- dplyr::filter(locationTbl, !is.na(.data$countyName))
  # tbl_2 <- dplyr::filter(locationTbl, is.na(.data$countyName))
  # 
  # if ( nrow(tbl_2) > 0 ) {
  #   
  #   if ( verbose ) 
  #     message(sprintf("Creating counties for %d locations ...", nrow(tbl_2)))
  #   
  #   tbl_2$countyName <- MazamaSpatialUtils::getUSCounty(
  #     lon = tbl_2$longitude,
  #     lat = tbl_2$latitude,
  #     dataset = "USCensusCounties",
  #     useBuffering = TRUE
  #   )
  #   
  #   locationTbl <- dplyr::bind_rows(tbl_1, tbl_2)
  #   
  # }
  
  # * timezone -----

  tbl_1 <- dplyr::filter(locationTbl, !is.na(.data$timezone))
  tbl_2 <- dplyr::filter(locationTbl, is.na(.data$timezone))
  
  if ( nrow(tbl_2) > 0 ) {
    
    if ( verbose ) 
      message(sprintf("Creating timezones for %d locations ...", nrow(tbl_2)))
    
    tbl_2$timezone <- MazamaSpatialUtils::getTimezone(
      lon = tbl_2$longitude,
      lat = tbl_2$latitude,
      dataset = "OSMTimezones",
      useBuffering = TRUE
    )
    
    locationTbl <- dplyr::bind_rows(tbl_1, tbl_2)
    
  }
  
  # * houseNumber -----
  
  # Slow web service so skip for now
  
  # * street -----
  
  # Slow web service so skip for now
  
  # * city -----
  
  # Slow web service so skip for now

  # * postalCode -----
  
  # Slow web service so skip for now

  # ----- Check for adjacent locations -----------------------------------------
  
  overlappingTbl <- table_findAdjacentDistances(locationTbl, distanceThreshold, measure)

  if ( nrow(overlappingTbl) > 0 ) {
    
    overlappingCount <- nrow(overlappingTbl)
    
    # Format the first line of the warning message
    firstLine <- sprintf(
      "%d locations have neighbors that are < %d m apart\n",
      round(overlappingCount),
      distanceThreshold
    )
    
    # Create a vector of lines, on for each overlappingTbl location pair
    overlappingLines <- vector("character", length = overlappingCount)
    for ( i in seq_len(nrow(overlappingTbl)) ) {
      
      overlappingLines[i] <- sprintf(
        "Distance: %6.1f -- rows %s %s",
        round(overlappingTbl[i, 3], 1),
        overlappingTbl[i, 1],
        overlappingTbl[i, 2]
      )
      
    }
    
    instructions <- sprintf("
The presence of locations closer than twice the specified distanceThreshold invalidate the 
uniqueness of a 'known locations' table and should be rectified. There are several 
basic options:

  1) Reduce the distanceThreshold to less than the half the minimum distance.
  2) Manually remove one location from each pair.
  3) Manually merge nearby locations to share the same longitude, latitude and
     locationID
     
Please review the returned locationTbl for the identified rows with:

locationTbl %%>%%
  table_findAdjacentLocations(distanceThreshold = %d, measure = \"%s\") %%>%%
  table_leaflet()

  ", round(distanceThreshold), measure)
    
    lines <- c(firstLine, overlappingLines, instructions)
    
    # Paste the lines together
    warning(paste(lines, collapse = "\n"))
    
  } # END overlapping
  
  # ----- Return ---------------------------------------------------------------
  
  return(locationTbl)
  
}

#' @export
#'
#' @title Add clustering information to a dataframe
#'
#' @param tbl Tibble with geolocation information (*e.g.*.
#' @param clusterDiameter Diameter in meters used to determine the number of
#' clusters (see description).
#' @param lonVar Name of longitude variable in the incoming tibble.
#' @param latVar Name of the latitude variable in the incoming tibble.
#' @param maxClusters Maximum number of clusters to try.
#'
#' @description Clustering is used to identify unique deployments of a
#' sensor in an environmental monitoring field study.
#'
#' Sensors will be moved around from time to time, sometimes across the country
#' and sometimes across the street.  We would like to assign unique identifiers
#' to each new "deployment" but not when the sensor is moved a short distance.
#'
#' We use clustering to find an appropriate number of unique "deployments".
#' The sensitivity of this algorithm can be adjused with the clusterDiameter argument.
#'
#' Standard `kmeans` clustering does not work well when clusters can have widely
#' differing numbers of members. A much better result is acheived with
#' the Partitioning Around Medoids method available in `cluster::pam()`.
#'
#' The value of `clusterRadius` is compared with the output of
#' `cluster::pam(...)$clusinfo[,'av_diss']` to determine the number of clusters.
#' 
#' @note The `table_addClustering()` function implements two-stage clustering
#' using [clusterByDistance()]. If the first attempt at clustering produces
#' clustered locations that are still too close to eachother, another round
#' of clustering is performed using the results of the previous attempt. This
#' two-stage approach seems to work well in. practice. 
#'
#' @return Input tibble with additional columns: `clusterLon, clusterLat`.
#'
#' @references [When k-means clustering fails](https://working-with-data.mazamascience.com/2021/07/15/when-k-means-clustering-fails/)
#' 
#' @seealso [clusterByDistance()]
#' 
#' @examples
#' library(MazamaLocationUtils)
#'
#' # Fremont, Seattle 47.6504, -122.3509
#' # Magnolia, Seattle 47.6403, -122.3997
#' # Downtown Seattle 47.6055, -122.3370
#' 
#' fremont_x <- jitter(rep(-122.3509, 10), .0005)
#' fremont_y <- jitter(rep(47.6504, 10), .0005)
#' 
#' magnolia_x <- jitter(rep(-122.3997, 8), .0005)
#' magnolia_y <- jitter(rep(47.6403, 8), .0005)
#' 
#' downtown_x <- jitter(rep(-122.3370, 3), .0005)
#' downtown_y <- jitter(rep(47.6055, 3), .0005)
#' 
#' # Apply clustering
#' tbl <-
#'   dplyr::tibble(
#'     longitude = c(fremont_x, magnolia_x, downtown_x),
#'     latitude = c(fremont_y, magnolia_y, downtown_y)
#'   ) %>%
#'   table_addClustering(
#'     clusterDiameter = 1000
#'   )
#' 
#' plot(tbl$longitude, tbl$latitude, pch = tbl$clusterID)


table_addClustering <- function(
  tbl,
  clusterDiameter = 1000,
  lonVar = "longitude",
  latVar = "latitude",
  maxClusters = 50
) {

  # ----- Validate parameters --------------------------------------------------

  MazamaCoreUtils::stopIfNull(tbl)
  MazamaCoreUtils::stopIfNull(clusterDiameter)
  MazamaCoreUtils::stopIfNull(lonVar)
  MazamaCoreUtils::stopIfNull(latVar)
  MazamaCoreUtils::stopIfNull(maxClusters)
  
  # Sanity check -- row count
  if ( nrow(tbl) == 0 ) {
    stop(paste0("Unable to perform clustering: 'tbl' is empty"))
  }

  # Sanity check -- names
  if ( !lonVar %in% names(tbl) ) {
    stop(paste0("Longitudes could not be found.  Did you specify the lonVar argument?"))
  }
  if ( !latVar %in% names(tbl) ) {
    stop(paste0("Latitudes could not be found.  Did you specify the latVar argument?"))
  }

  # If we only have a single row, return immediately
  # NOTE:  Use as.numeric() to remove any names associated with these vectors
  if ( nrow(tbl) == 1 ) {
    tbl$clusterLon <- as.numeric(tbl[[lonVar]][1])
    tbl$clusterLat <- as.numeric(tbl[[latVar]][1])
    tbl$clusterID <- 1
    return(tbl)
  }

  # ----- Cluster by distance --------------------------------------------------

  tbl <-
    clusterByDistance(
      tbl,
      clusterDiameter = clusterDiameter,
      lonVar = lonVar,
      latVar = latVar,
      maxClusters = maxClusters
    )

  # ----- Verify clustering ----------------------------------------------------

  # Only use one location record per clusterID
  distinctTbl <-
    tbl %>%
    dplyr::distinct(.data$clusterID, .keep_all = TRUE) %>%
    dplyr::mutate(
      longitude = .data$clusterLon,
      latitude = .data$clusterLat
    ) %>%
    dplyr::select(c("longitude", "latitude"))

  # Check if any of the clustered locations are too close
  adjacentDistances <-
    distinctTbl %>%
    MazamaLocationUtils::table_findAdjacentDistances(
      distanceThreshold = clusterDiameter
    )

  # ----- Fix clustering -------------------------------------------------------

  if ( nrow(adjacentDistances) > 0 ) {

    # NOTE:  Clearly, clustering has failed. At this point we just cluster the
    # NOTE:  distinct locations. Having just a few points should radically
    # NOTE:  improve the reliability of cluster::pam().

    # Cluster the distinct locations
    distinctTbl <-
      clusterByDistance(
        distinctTbl,
        clusterDiameter = clusterDiameter,
        lonVar = "longitude",                # the renamed clusterLon
        latVar = "latitude",                 # the renamed clusterLat
        maxClusters = maxClusters
        #maxClusters = floor(nrow(distinctTbl) / 2) - 1
      )

    clusterCount <- max(as.numeric(distinctTbl$clusterID), na.rm = TRUE)

    # Now use this clusterCount for full clustering -- see clusterByDistance()

    if ( nrow(tbl) < 2000 ) {
      clusterObj <- cluster::pam(tbl[,c(lonVar,latVar)], clusterCount)
    } else {
      clusterObj <- cluster::clara(tbl[,c(lonVar,latVar)], clusterCount, samples = 50)
    }

    tbl$clusterLon <- as.numeric(clusterObj$medoids[,lonVar][clusterObj$clustering])
    tbl$clusterLat <- as.numeric(clusterObj$medoids[,latVar][clusterObj$clustering])
    tbl$clusterID <- as.character(clusterObj$clustering)

  }

  # ----- Return ---------------------------------------------------------------

  return(tbl)

}

# ===== DEBUG ==================================================================

if ( FALSE ) {


}

#' @title Create an empty known location table
#' 
#' @description Creates an empty known location tibble with the following 
#' columns of core metadata:
#' 
#' \itemize{
#' \item{locationID}
#' \item{locationName}
#' \item{longitude}
#' \item{latitude}
#' \item{elevation}
#' \item{countryCode}
#' \item{stateCode}
#' \item{countyName}
#' \item{timezone}
#' \item{houseNumber}
#' \item{street}
#' \item{city}
#' \item{postalCode}
#' }
#' 
#' @return Empty known location tibble with the specified metadata columns.
#' 
#' @examples 
#' library(MazamaLocationUtils)
#' 
#' # Create an empty Tbl
#' emptyTbl <- table_initialize()
#' dplyr::glimpse(emptyTbl)
#' 
#' @rdname table_initialize
#' @export 
#' @importFrom MazamaCoreUtils stopIfNull
#' @importFrom dplyr tibble filter
#' @importFrom rlang .data
table_initialize <- function() {
  
  # ----- Validate parameters --------------------------------------------------
  

  # ----- Create empty tibble --------------------------------------------------
  
  # Build up a tibble with a single record full of NAs
  locationTbl <- dplyr::tibble(
    "locationID" = as.character(NA),
    "locationName" = as.character(NA),
    "longitude" = as.numeric(NA),
    "latitude" = as.numeric(NA),
    "elevation" = as.numeric(NA),
    "countryCode" = as.character(NA),
    "stateCode" = as.character(NA),
    "countyName" = as.character(NA),
    "timezone" = as.character(NA),
    "houseNumber" = as.character(NA),
    "street" = as.character(NA),
    "city" = as.character(NA),
    "postalCode" = as.character(NA)
  )
  
  # Now search for an ID we won't find to end up with an empty tibble with 
  # the correct column names.
  locationTbl <-
    locationTbl %>%
    dplyr::filter(.data$locationID == "Rumplestiltskin")

  # ----- Return ---------------------------------------------------------------
  
  return(locationTbl)
  
}

context("table_initialize")

test_that("table_initialize() has proper columns", {
  skip_on_cran()
  skip_on_travis()
  
  initializeMazamaSpatialUtils()
  locationTbl <- table_initialize()
  expect_equal(names(locationTbl), MazamaLocationUtils::coreMetadataNames)
})

context("table_getNearestLocation")

test_that("table_getNearestLocation() works", {
  
  locationTbl <- get(data("wa_monitors_500"))
  
  emptyRecord <- table_initialize()    # zero rows
  emptyRecord[1,1] <- as.character(NA) # one row, all NA
  wenatcheeRecord <- locationTbl %>% dplyr::filter(city == "Wenatchee")
  
  # Wenatchee
  lon <- -120.325278
  lat <- 47.423333
  
  testClose <- table_getNearestLocation(locationTbl, lon, lat, distanceThreshold = 100)
  testFar <- table_getNearestLocation(locationTbl, lon, lat, distanceThreshold = 10000)
  expect_equal(testClose, emptyRecord)
  expect_equal(testFar, wenatcheeRecord)
  
})

#' @rdname location_getSingleAddress_Photon
#' @export 
#' @importFrom utils capture.output
#' @importFrom jsonlite fromJSON
#' @importFrom stringr str_detect str_sub str_subset
#' 
#' @title Get address data from the Photon API to OpenStreetMap
#' 
#' @param longitude Single longitude in decimal degrees E.
#' @param latitude Single latitude in decimal degrees N.
#' @param baseUrl Base URL for data queries.
#' @param verbose Logical controlling the generation of progress messages.
#' 
#' @return List of address components.
#' 
#' @description The Photon API is used get address data associated with
#' the `longitude` and `latitude`. The following list of data
#' is returned:
#' 
#' \itemize{
#' \item{`houseNumber`}
#' \item{`street`}
#' \item{`city`}
#' \item{`stateCode`}
#' \item{`stateName`}
#' \item{`postalCode`}
#' \item{`countryCode`}
#' \item{`countryName`}
#' }
#' 
#' The function makes an effort to convert both `state` and `country` 
#' `Name` into `Code` with codes defaulting to `NA`. Both 
#' `Name` and `Code` are returned so that improvements can be made in 
#' the conversion algorithm.
#' 
#' @examples
#' \donttest{
#' library(MazamaLocationUtils)
#' 
#' # Fail gracefully if any resources are not available
#' try({
#' 
#'   # Set up standard directories and spatial data
#'   spatialDataDir <- tempdir() # typically "~/Data/Spatial"
#'   initializeMazamaSpatialUtils(spatialDataDir)
#' 
#'   # Wenatchee
#'   lon <- -120.325278
#'   lat <- 47.423333
#' 
#'   addressList <- location_getSingleAddress_Photon(lon, lat)
#'   str(addressList)
#'   
#' }, silent = FALSE)
#' }
#' 
#' @references \url{https://photon.komoot.io}
#' 
location_getSingleAddress_Photon <- function(
  longitude = NULL,
  latitude = NULL,
  baseUrl = "https://photon.komoot.io/reverse",
  verbose = TRUE
) {

  validateMazamaSpatialUtils()
  
  # ----- Validate parameters --------------------------------------------------
  
  validateLonLat(longitude, latitude)  
  
  # ----- Get Photon address data ----------------------------------------------

  .params <- list(
    lon = longitude,
    lat = latitude
  )
    
  suppressWarnings({
    r <- httr::GET(baseUrl, query = .params)
  })

  if ( httr::http_error(r) ) {
    stop("An error has occured while querying the Photon API.")
  }
    
  fileString <- httr::content(r, 'text', encoding = 'UTF-8')
    
  if ( stringr::str_detect(fileString, "404 Not found") ) {
    stop("The Photon API returned no data for the provided longitude and latitude.")
  }
  
  photonJSON <- jsonlite::fromJSON(fileString)
  photonList <- photonJSON$features$properties
  
  # ----- Validate Photon Address Data -----------------------------------------

  # Replace "... Not Found" with NA
  photonList <- lapply(photonList, function(x) {
    if ( typeof(x) == "character" ) {
      if ( stringr::str_detect(x, "Not Found") ) {
        return(as.character(NA))
      } else {
        return(x)
      }
    } else {
      return(x)
    }
  })
  
  # NOTE:  A careful check of the addresses returned by Photon shows them to be
  # NOTE:  incorrect by up to 6 blocks. So we will throw those away. Also, we saw
  # NOTE:  a postalCode in Colleville specified as "SUPER ONE FOODS". So we have to 
  # NOTE:  validate that as well.
  
  photonList$housenumber <- as.character(NA)
  photonList$street <- as.character(NA)
  
  if ( photonList$country %in% c("United States") ) {
    postcode <- 
      stringr::str_sub(photonList$postcode, 1, 5) %>%
      stringr::str_subset("^[0-9]{5}$")
    if ( length(postcode) > 0 ) {
      photonList$postcode <- postcode
    } else {
      photonList$postcode <- as.character(NA)
    }
  }
  
  # NOTE: Sometimes state contains the state code and sometimes contains the state name
  if ( nchar(photonList$state) > 2 ) {
    photonList$stateName <- photonList$state
    photonList$stateCode <- as.character(NA)
  } else {
    photonList$stateName <- as.character(NA)
    photonList$stateCode <- photonList$state
  }
  
  # ----- Create addressList ---------------------------------------------------
  
  addressList <- list(
    houseNumber = as.character(photonList$housenumber),
    street = as.character(photonList$street),
    city = as.character(photonList$city),
    stateName = photonList$stateName,
    stateCode = toupper(photonList$stateCode),
    postalCode = as.character(photonList$postcode),
    countryName = photonList$country,
    countryCode = toupper(photonList$countrycode)
  )
  
  # countryCode 
  if ( is.na(addressList$countryCode) ) {
    suppressWarnings({
      addressList$countryCode <- 
        MazamaSpatialUtils::countryToCode(addressList$countryName)
    })
  }

  # stateCode
  if ( is.na(addressList$stateCode) ) {
    suppressWarnings({
      addressList$stateCode <- 
        MazamaSpatialUtils::stateToCode(
          stateNames = addressList$stateName,
          countryCodes = addressList$countryCode,
          dataset = "NaturalEarthAdm1"
        )
    }) 
  }
  
  # stateName
  if ( is.na(addressList$stateName) ) {
    suppressWarnings({
      addressList$stateName <- 
        MazamaSpatialUtils::US_stateCodeToName(
          stateCode = addressList$stateCode
        )
    }) 
  }

  # ----- Return ---------------------------------------------------------------
  
  return(addressList)
  
}

#' @rdname location_getOpenCageInfo
#' @export 
#'
#' @title Get location information from OpenCage
#' 
#' @description The OpenCage reverse geocoding service is used to obtain all 
#' available information for a specific location.
#' 
#' The data from OpenCage should be considered to be the gold standard for address
#' information could and should be used to override information we get elsewhere.
#' 
#' @note The OpenCage service requires an API key which can be obtained from 
#' their web site. This API key must be set as an environment variable with:
#' 
#' \preformatted{
#' Sys.setenv("OPENCAGE_KEY" = "YOUR_PERSONAL_API_KEY")
#' }
#' 
#' The OpenCage "free trial" level allows for 1 request/sec and a maximum of 
#' 2500 requests per day.
#' 
#' @param longitude Single longitude in decimal degrees E.
#' @param latitude Single latitude in decimal degrees N.
#' @param verbose Logical controlling the generation of progress messages.
#' 
#' @return Single-row tibble with OpenCage information.
#' 
#' @references \url{https://opencagedata.com}
#' 
#' @examples 
#' \donttest{
#' library(MazamaLocationUtils)
#' 
#' # Fail gracefully if any resources are not available
#' try({
#' 
#'   # Wenatchee
#'   lon <- -120.325278
#'   lat <- 47.423333
#' 
#'   Sys.setenv("OPENCAGE_KEY" = "YOUR_PERSONAL_API_KEY")
#' 
#'   openCageTbl <- location_getOpenCageInfo(lon, lat)
#'   dplyr::glimpse(openCageTbl)
#'   
#' }, silent = FALSE)
#' }
#' 
location_getOpenCageInfo <- function(
  longitude = NULL,
  latitude = NULL,
  verbose = FALSE
) {
  
  # ----- Validate parameters --------------------------------------------------
  
  validateLonLat(longitude, latitude)  
  
  if ( Sys.getenv("OPENCAGE_KEY") == "" )  {
    stop("Revese geocoding with OpenCage requires an API key 
Please set one with Sys.setenv(\"OPENCAGE_KEY\" = \"<YOUR_KEY>\").")
  }
  
  # ----- Reverse geocode ------------------------------------------------------
  
  openCageTbl <- tidygeocoder::reverse_geo(
    latitude, 
    longitude,  
    method = "opencage",
    address = "address",
    limit = 1,
    full_results = TRUE,
    mode = "",
    unique_only = FALSE,
    return_coords = TRUE,
    min_time = NULL,
    progress_bar = verbose,
    quiet = !verbose,
    api_url = NULL,
    timeout = 20,
    flatten = TRUE,
    batch_limit = NULL,
    verbose = verbose, # Very detailed
    no_query = FALSE,
    custom_query = list(),
    api_options = list()
  )
  
  # > dplyr::glimpse(openCageTbl, width = 75)
  # Rows: 5
  # Columns: 78
  # $ lat                                         <dbl> 43.64789, 45.18191, 4
  # $ long                                        <dbl> -116.2694, -113.8903,
  # $ address                                     <chr> "5164 North Alworth S
  # $ confidence                                  <int> 10, 9, 9, 10, 10
  # $ annotations.MGRS                            <chr> "11TNJ5890232972", "1
  # $ annotations.Maidenhead                      <chr> "DN13up75pj", "DN35be
  # $ annotations.callingcode                     <int> 1, 1, 1, 1, 1
  # $ annotations.flag                            <chr> <flag emojis>
  # $ annotations.geohash                         <chr> "9rv261z09hpr429xq8r8
  # $ annotations.qibla                           <dbl> 23.89, 25.88, 23.82, 
  # $ annotations.DMS.lat                         <chr> "43° 38' 50.70048'' N
  # $ annotations.DMS.lng                         <chr> "116° 16' 10.67196'' 
  # $ annotations.FIPS.county                     <chr> "16001", "16059", "16
  # $ annotations.FIPS.state                      <chr> "16", "16", "16", "16
  # $ annotations.Mercator.x                      <dbl> -12943076, -12678177,
  # $ annotations.Mercator.y                      <dbl> 5381538, 5619763, 537
  # $ annotations.OSM.edit_url                    <chr> "https://www.openstre
  # $ annotations.OSM.note_url                    <chr> "https://www.openstre
  # $ annotations.OSM.url                         <chr> "https://www.openstre
  # $ annotations.UN_M49.statistical_groupings    <list> "MEDC", "MEDC", "MEDC
  # $ annotations.UN_M49.regions.AMERICAS         <chr> "019", "019", "019", 
  # $ annotations.UN_M49.regions.NORTHERN_AMERICA <chr> "021", "021", "021",
  # $ annotations.UN_M49.regions.US               <chr> "840", "840", "840", 
  # $ annotations.UN_M49.regions.WORLD            <chr> "001", "001", "001", 
  # $ annotations.currency.alternate_symbols      <list> "US$", "US$", "US$", 
  # $ annotations.currency.decimal_mark           <chr> ".", ".", ".", ".", "
  # $ annotations.currency.disambiguate_symbol    <chr> "US$", "US$", "US$",
  # $ annotations.currency.html_entity            <chr> "$", "$", "$", "$", "
  # $ annotations.currency.iso_code               <chr> "USD", "USD", "USD", 
  # $ annotations.currency.iso_numeric            <chr> "840", "840", "840", 
  # $ annotations.currency.name                   <chr> "United States Dollar
  # $ annotations.currency.smallest_denomination  <int> 1, 1, 1, 1, 1
  # $ annotations.currency.subunit                <chr> "Cent", "Cent", "Cent
  # $ annotations.currency.subunit_to_unit        <int> 100, 100, 100, 100, 1
  # $ annotations.currency.symbol                 <chr> "$", "$", "$", "$", "
  # $ annotations.currency.symbol_first           <int> 1, 1, 1, 1, 1
  # $ annotations.currency.thousands_separator    <chr> ",", ",", ",", ",", "
  # $ annotations.roadinfo.drive_on               <chr> "right", "right", "ri
  # $ annotations.roadinfo.road                   <chr> "North Alworth Street
  # $ annotations.roadinfo.speed_in               <chr> "mph", "mph", "mph", 
  # $ annotations.sun.rise.apparent               <int> 1642173360, 164217312
  # $ annotations.sun.rise.astronomical           <int> 1642167300, 164216682
  # $ annotations.sun.rise.civil                  <int> 1642171500, 164217114
  # $ annotations.sun.rise.nautical               <int> 1642169340, 164216898
  # $ annotations.sun.set.apparent                <int> 1642120380, 164211948
  # $ annotations.sun.set.astronomical            <int> 1642126440, 164212578
  # $ annotations.sun.set.civil                   <int> 1642122240, 164212146
  # $ annotations.sun.set.nautical                <int> 1642124400, 164212362
  # $ annotations.timezone.name                   <chr> "America/Boise", "Ame
  # $ annotations.timezone.now_in_dst             <int> 0, 0, 0, 0, 0
  # $ annotations.timezone.offset_sec             <int> -25200, -25200, -2520
  # $ annotations.timezone.offset_string          <chr> "-0700", "-0700", "-0
  # $ annotations.timezone.short_name             <chr> "MST", "MST", "MST", 
  # $ annotations.what3words.words                <chr> "sake.pardon.topped",
  # $ bounds.northeast.lat                        <dbl> 43.64747, 45.18156, 4
  # $ bounds.northeast.lng                        <dbl> -116.2696, -113.8889,
  # $ bounds.southwest.lat                        <dbl> 43.64737, 45.17625, 4
  # $ bounds.southwest.lng                        <dbl> -116.2697, -113.8955,
  # $ `components.ISO_3166-1_alpha-2`             <chr> "US", "US", "US", "US
  # $ `components.ISO_3166-1_alpha-3`             <chr> "USA", "USA", "USA", 
  # $ components._category                        <chr> "building", "road", "
  # $ components._type                            <chr> "building", "road", "
  # $ components.continent                        <chr> "North America", "Nor
  # $ components.country                          <chr> "United States", "Uni
  # $ components.country_code                     <chr> "us", "us", "us", "us
  # $ components.county                           <chr> "Ada County", "Lemhi 
  # $ components.house_number                     <chr> "5164", NA, NA, "201"
  # $ components.postcode                         <chr> "83714", "83467", "83
  # $ components.road                             <chr> "North Alworth Street
  # $ components.state                            <chr> "Idaho", "Idaho", "Id
  # $ components.state_code                       <chr> "ID", "ID", "ID", "ID
  # $ components.town                             <chr> "Garden City", "Salmo
  # $ geometry.lat                                <dbl> 43.64742, 45.18115, 4
  # $ geometry.lng                                <dbl> -116.2696, -113.8900,
  # $ annotations.roadinfo.road_type              <chr> NA, "residential", "t
  # $ components.road_type                        <chr> NA, "residential", "t
  # $ components.village                          <chr> NA, "West Salmon", NA
  # $ components.city                             <chr> NA, NA, "Meridian", N
  

  # ----- Fixes ----------------------------------------------------------------
  
  # NOTE:  Fix any column types (typically <int> or <dbl> needs to become <character>)
  
  if ( "components.road_reference" %in% names(openCageTbl) ) {
    openCageTbl$components.road_reference <-
      as.character(openCageTbl$components.road_reference)
  }
  
  if ( "components.building" %in% names(openCageTbl) ) {
    cat(sprintf("location with 'building'= %f, %f\n", longitude, latitude))
    openCageTbl$components.building <-
      as.character(openCageTbl$components.building)
  }
  
  # ----- Return ---------------------------------------------------------------
  
  return(openCageTbl)
  
}

# ===== DEBUGGING ==============================================================

if ( FALSE ) {
  
  library(MazamaLocationUtils)
  
  locationTbl <- 
    MazamaLocationUtils::id_monitors_500
  
  i <- 22
  longitude <- locationTbl$longitude[i]
  latitude <- locationTbl$latitude[i]
  verbose <- TRUE
  
  
  
}

#' @title Update a single known location record in a table
#' @description Information in the `locationList` is used to replace
#' existing information found in `locationTbl`. This function can be used
#' for small tweaks to an existing `locationTbl`. Wholesale replacement of
#' records should be performed with `table_removeRecord()` followed by
#' `table_addLocation()`. 
#' @param locationTbl Tibble of known locations.
#' @param locationList List containing `locationID` and one or more named
#' columns whose values are to be replaced.
#' @param verbose Logical controlling the generation of progress messages.
#' @return Updated tibble of known locations.
#' @examples
#' library(MazamaLocationUtils)
#' 
#' locationTbl <- get(data("wa_monitors_500"))
#' 
#' # Wenatchee
#' wenatcheeRecord <- 
#'   locationTbl %>% 
#'   dplyr::filter(city == "Wenatchee")
#' 
#' str(wenatcheeRecord)
#' 
#' wenatcheeID <- wenatcheeRecord$locationID
#' 
#' locationTbl <- table_updateSingleRecord(
#'   locationTbl,
#'   locationList = list(
#'     locationID = wenatcheeID,
#'     locationName = "Wenatchee-Fifth St"
#'   )
#' )
#' 
#' # Look at the new record
#' locationTbl %>% 
#'   dplyr::filter(city == "Wenatchee") %>%
#'   str()
#' 
#' @seealso [table_addLocation()]
#' @seealso [table_addSingleLocation()]
#' @seealso [table_removeRecord()]
#' @rdname table_updateSingleRecord
#' @export 
#' @importFrom rlang .data
table_updateSingleRecord <- function(
  locationTbl = NULL,
  locationList = NULL,
  verbose = TRUE
) {
  
  # ----- Validate parameters --------------------------------------------------
  
  MazamaLocationUtils::validateLocationTbl(locationTbl, locationOnly = FALSE)
  MazamaCoreUtils::stopIfNull(locationList)

  invalidNames <- setdiff(names(locationList), names(locationTbl))
  if ( length(invalidNames) > 0 ) {
    invalidString <- paste0(invalidNames, collapse = ", ")
    stop(sprintf(
      "Invalid names found in locationList: %s", invalidString
    ))
  }
  
  if ( !locationList$locationID %in% locationTbl$locationID ) 
    stop(sprintf(
      "locationID %s is not found in locationTbl", locationList$locationID
    ))
  
  # ----- Update locationTbl ---------------------------------------------------
  
  row <- which(locationTbl$locationID == locationList$locationID)
  
  if ( length(row) > 0 ) {
    
    for ( columnName in names(locationList) ) {
      locationTbl[row, columnName] <- locationList[[columnName]]
    }
    
  } else {
    
    if ( verbose ) {
      warning(sprintf(
        "No location found for %s", locationList$locationID
      ))
    }
    
  }

  # ----- Return ---------------------------------------------------------------
  
  return(locationTbl)
  
}

context("table_updateSingleRecord")

test_that("table_updateSingleRecord() works", {
  locationTbl <- get(data("wa_monitors_500"))
  
  wenatcheeRecord <- locationTbl %>% dplyr::filter(city == "Wenatchee")
  wenatcheeID <- wenatcheeRecord$locationID
  wenatcheeOldName <- wenatcheeRecord$locationName
  
  testTbl <- table_updateSingleRecord(
    locationTbl,
    locationList = list(
      locationID = wenatcheeID,
      locationName = "Wenatchee-Fifth St"
    )
  )

  wenatcheeNewName <- 
    testTbl %>% 
    dplyr::filter(city == "Wenatchee") %>%
    dplyr::pull(locationName)
  
  expect_equal(wenatcheeOldName, "us.wa_c26mvd")
  expect_equal(wenatcheeNewName, "Wenatchee-Fifth St")
})

#' @title Remove a column of metadata in a table
#' @description Remove the column matching `columnName`. This function 
#' can be used in pipelines.
#' @param locationTbl Tibble of known locations.
#' @param columnName Name of the colun to be removed.
#' @param verbose Logical controlling the generation of progress messages.
#' @return Updated tibble of known locations.
#' @examples
#' library(MazamaLocationUtils)
#' 
#' # Starting table
#' locationTbl <- get(data("wa_monitors_500")) 
#' names(locationTbl)
#' 
#' # Add a new column
#' locationTbl <-
#'   locationTbl %>%
#'   table_addColumn("AQSID")
#'   
#' names(locationTbl)
#'
#' # Now remove it
#' locationTbl <-
#'   locationTbl %>%
#'   table_removeColumn("AQSID")
#' 
#' names(locationTbl)
#' 
#' try({
#'   # Cannot remove "core" metadata
#'   locationTbl <-
#'     locationTbl %>%
#'     table_removeColumn("longitude")
#' }, silent = FALSE)
#'
#' @seealso [table_addColumn()]
#' @seealso [table_removeColumn()]
#' @rdname table_removeColumn
#' @export 
#' @importFrom MazamaCoreUtils stopIfNull
table_removeColumn <- function(
  locationTbl = NULL,
  columnName = NULL,
  verbose = TRUE
) {
  
  # ----- Validate parameters --------------------------------------------------
  
  MazamaLocationUtils::validateLocationTbl(locationTbl, locationOnly = FALSE)
  MazamaCoreUtils::stopIfNull(columnName)
  
  if ( !columnName %in% names(locationTbl) ) 
    stop(sprintf(
      "columnName %s is not found in locationTbl", columnName
    ))
  
  if ( columnName %in% MazamaLocationUtils::coreMetadataNames )
    stop(sprintf(
      "columnName '%s' is part of the core metdata and cannot be removed", columnName
    ))
  
  
  # ----- Remove column --------------------------------------------------------
  
  locationTbl[[columnName]] <- NULL
  
  # ----- Return ---------------------------------------------------------------
  
  return(locationTbl)
  
}

#' @keywords internal
#' @name MazamaLocationUtils
#' @title Manage Spatial Metadata for Known Locations
#' @description A suite of utility functions for discovering and managing 
#' metadata associated with sets of spatially unique "known locations".
#' 
#' This package is intended to be used in support of data management activities
#' associated with fixed locations in space. The motivating fields include both
#' air and water quality monitoring where fixed sensors report at regular time 
#' intervals.
#'
#' @details
#' When working with environmental monitoring time series, one of the first things
#' you have to do is create unique identifiers for each individual time series. In 
#' an ideal world, each environmental time series would have both a 
#' `locationID` and a `deviceID` that uniquely identify the specific instrument 
#' making measurements and the physical location where measurements are made. A 
#' unique `timeseriesID` could
#' be produced as `locationID_deviceID`. Metadata associated with each
#' `timeseriesID` would contain basic information needed for downstream analysis
#' including at least:
#' 
#' \preformatted{
#'   timeseriesID, locationID, deviceID, longitude, latitude, ...
#' }
#' 
#' \itemize{
#' \item{An extended time series for an occasionally re-positioned sensor would group by `deviceID`.}
#' \item{Multiple sensors placed at a single location could be be grouped by `locationID`.}
#' \item{Maps would be created using `longitude, latitude`.}
#' \item{Time series would be accessed from a secondary `data` table with `timeseriesID`.}
#' }
#' 
#' Unfortunately, we are rarely supplied with a truly unique and truly spatial 
#' `locationID`. Instead we often use `deviceID` or an associated non-spatial
#' identifier as a stand-in for `locationID`.
#' 
#' Complications we have seen include:
#'   
#' \itemize{
#' \item{GPS-reported longitude and latitude can have _jitter_ in the fourth or fifth 
#' decimal place making it challenging to use them to create a unique `locationID`.}
#' \item{Sensors are sometimes _re-positioned_ in what the scientist considers the "same 
#' location".}
#' \item{Data for a single sensor goes through different processing pipelines using
#' different identifiers and is later brought together as two separate time series.}
#' \item{The spatial scale of what constitutes a "single location" depends on the 
#' instrumentation and scientific question being asked.}
#' \item{Deriving location-based metadata from spatial datasets is computationally 
#' intensive unless saved and identified with a unique `locationID`.}
#' \item{Automated searches for spatial metadata occasionally produce incorrect results
#' because of the non-infinite resolution of spatial datasets.}
#' }
#' 
#' This package attempts to address all of these issues by maintaining a table
#' of known locations for which CPU intensive spatial data calculations have
#' already been performed. While requests to add new locations to the table may
#' take some time, searches for spatial metadata associated with existing
#' locations are simple lookups.
#' 
#' Working in this manner will solve the problems initially mentioned but also 
#' provides further useful functionality.
#' 
#' \itemize{
#' \item{Administrators can correct entries in the `collectionName` table.  
#' (_e.g._ locations in river bends that even high resolution spatial datasets mis-assign)}
#' \item{Additional, non-automatable metadata can be added to `collectionName`. 
#' (_e.g._ commonly used location names within a community of practice)}
#' \item{Different field campaigns can have separate `collectionName` tables.}
#' \item{`.csv` or `.rda` versions of well populated tables can be downloaded from a
#' URL and used locally, giving scientists working with known locations instant
#' access to spatial data that otherwise requires special skills, large datasets 
#' and lots of compute cycles.}
#' }
"_PACKAGE"

# ----- Internal Data -------------------------------------------------

#' coreMetadataNames
#'
#' @export
#' @docType data
#' @name coreMetadataNames
#' @title Names of standard spatial metadata columns
#' @format A character vector
#' @description Character string identifiers of the minimum set of fields 
#' required for a table to be considered a valid "known locations" table.
#' 
#' \preformatted{
#' coreMetadataNames <- c(
#'   "locationID",           # from MazamaLocationUtils::location_createID()
#'   "locationName",         # from MazamaLocationUtils::location_initialize()
#'   "longitude",            # user supplied
#'   "latitude",             # user supplied
#'   "elevation",            # from MazamaLocationUtils::getSingleElevation_USGS()
#'   "countryCode",          # from MazamaSpatialUtils::getCountryCode()
#'   "stateCode",            # from MazamaSpatialUtils::getStateCode()
#'   "countyName",           # from MazamaSpatialUtils::getUSCounty()
#'   "timezone",             # from MazamaSpatialUtils::getTimezone()
#'   "houseNumber",
#'   "street",
#'   "city",
#'   "postalCode"
#' )
#' }

coreMetadataNames <- c(
  "locationID",           # from MazamaLocationUtils::location_createID()
  "locationName",         # from MazamaLocationUtils::location_initialize()
  "longitude",            # user supplied
  "latitude",             # user supplied
  "elevation",            # from MazamaLocationUtils::getSingleElevation_USGS()
  "countryCode",          # from MazamaSpatialUtils::getCountryCode()
  "stateCode",            # from MazamaSpatialUtils::getStateCode()
  "countyName",           # from MazamaSpatialUtils::getUSCounty()
  "timezone",             # from MazamaSpatialUtils::getTimezone()
  "houseNumber",          # from MazamaLocationUtils::getSingleAddress_Photon()
  "street",               # from MazamaLocationUtils::getSingleAddress_Photon()
  "city",                 # from MazamaLocationUtils::getSingleAddress_Photon()
  "postalCode"            # from MazamaLocationUtils::getSingleAddress_Photon()
)

# ----- Internal Package State -------------------------------------------------

MazamaLocationUtilsEnv <- new.env(parent = emptyenv())
MazamaLocationUtilsEnv$dataDir <- NULL

# ----- Data Directory Configuration -------------------------------------------

#' @docType data
#' @keywords environment
#' @name LocationDataDir
#' @title Directory for location data
#' @format Absolute path string.
#' @description This package maintains an internal directory path which 
#' users can set using `setLocationDataDir()`. All package functions use 
#' this directory whenever known location tables are accessed.
#' 
#' The default setting when the package is loaded is `getwd()`.
#' @seealso [getLocationDataDir()]
#' @seealso [setLocationDataDir()]
NULL

#' @keywords environment
#' @export
#' @title Get location data directory
#' @description Returns the directory path where known location data tables are located.
#' @return Absolute path string.
#' @seealso [LocationDataDir()]
#' @seealso [setLocationDataDir()]

getLocationDataDir <- function() {
  if (is.null(MazamaLocationUtilsEnv$dataDir)) {
    stop(paste0('No data directory found. Please set a data directory with ',
                'setLocationDataDir("YOUR_DATA_DIR")'),
         call.=FALSE)
  } else {
    return(MazamaLocationUtilsEnv$dataDir)
  }
}

#' @keywords environment
#' @export
#' @title Set location data directory
#' @param dataDir Directory where location tables are stored.
#' @description Sets the data directory where known location data tables are located.
#' If the directory does not exist, it will be created.
#' @return Silently returns previous value of the data directory.
#' @seealso [LocationDataDir()]
#' @seealso [getLocationDataDir()]

setLocationDataDir <- function(dataDir) {
  old <- MazamaLocationUtilsEnv$dataDir
  dataDir <- path.expand(dataDir)
  tryCatch({
    if (!file.exists(dataDir)) dir.create(dataDir)
    MazamaLocationUtilsEnv$dataDir <- dataDir
  }, warning = function(warn) {
    warning("Invalid path name.")
  }, error   = function(err) {
    stop(paste0("Error in setLocationDataDir(",dataDir,")."))
  })
  return(invisible(old))
}

#' @rdname location_getSingleElevation_USGS
#' @export 
#' 
#' @title Get elevation data from a USGS web service
#' 
#' @param longitude Single longitude in decimal degrees E.
#' @param latitude Single latitude in decimal degrees N.
#' @param verbose Logical controlling the generation of progress messages.
#' 
#' @return Numeric elevation value.
#' 
#' @description USGS APIs are used to determine the elevation in meters 
#' associated with the `longitude` and `latitude`.
#' 
#' _Note: The conversion factor for meters to feet is 3.28084._
#' 
#' @examples
#' \donttest{
#' library(MazamaLocationUtils)
#' 
#' # Fail gracefully if any resources are not available
#' try({
#' 
#'   # Wenatchee
#'   longitude <- -120.325278
#'   latitude <- 47.423333
#' 
#'   location_getSingleElevation_USGS(longitude, latitude)
#'   
#' }, silent = FALSE)
#' }
#' 
#' @references \url{https://apps.nationalmap.gov/epqs/}
#' 
location_getSingleElevation_USGS <- function(
  longitude = NULL,
  latitude = NULL,
  verbose = TRUE
) {
  
  # ----- Validate parameters --------------------------------------------------
  
  validateLonLat(longitude, latitude)  

  # ----- Get USGS elevation data ----------------------------------------------
  
  # https://epqs.nationalmap.gov/v1/json?x=-122.5&y=47.5&wkid=4326&units=Meters&includeDate=false
  
  # {
  #   "location": {
  #     "x": -122.2345,
  #     "y": 47.1234,
  #     "spatialReference": {
  #       "wkid": 4326,
  #       "latestWkid": 4326
  #     }
  #   },
  #   "locationId": 0,
  #   "value": "42.277534485",
  #   "rasterId": 56925,
  #   "resolution": 1
  # }
  
  url <- sprintf(
    "https://epqs.nationalmap.gov/v1/json?x=%f&y=%f&wkid=4326&units=Meters&includeDate=false",
    longitude, latitude
  )
  
  r <- httr::GET(url)
  
  # Default to NA unless we get a result
  elevation <- as.numeric(NA)
  
  if ( httr::http_error(r) ) {
    
    if ( verbose ) {
      warning(sprintf(
        "USGS elevation service failed for URL %s", 
        url
      ))
    }
    
  } else {
    
    result <- try({
      
      returnObj <- httr::content(r)
      elevation <- as.numeric(returnObj$value)
      
    }, silent = TRUE)
    
    if ( "try-error" %in% class(result) ) {
      if ( verbose ) {
        warning(sprintf(
          "USGS elevation service returned an error for this request:\n  %s",
          url
        ))
      }
    }
    
  }
  
  # This page says the vertical accuracy has a RMSE of .53 meters
  # https://www.usgs.gov/faqs/what-vertical-accuracy-3d-elevation-program-3dep-dems
  return(round(elevation, 1))
  
}

# ===== DEBUGGING ==============================================================

if ( FALSE ) {
  
  library(MazamaCoreUtils)
  
  longitude <- -112.234
  latitude <- 57.234
  verbose <- FALSE
  
  
  location_getSingleElevation_USGS(longitude, latitude, verbose)
  
  
  
}

#' @title Add new known location records to a table
#' 
#' @description Incoming `longitude` and `latitude` values are compared 
#' against the incoming `locationTbl` to see if they are already within
#' `distanceThreshold` meters of an existing entry. A new record is created for
#' each location that is not already found in `locationTbl`.
#' 
#' @note This function is a vectorized version of `table_addSingleLocation()`.
#' 
#' @param locationTbl Tibble of known locations.
#' @param longitude Vector of longitudes in decimal degrees E.
#' @param latitude Vector of latitudes in decimal degrees N.
#' @param distanceThreshold Distance in meters.
#' @param stateDataset Name of spatial dataset to use for determining state
#' codes, Default: 'NaturalEarthAdm1'
#' @param elevationService Name of the elevation service to use for determining
#' the elevation. Default: NULL skips this step. Accepted values: "usgs".
#' @param addressService Name of the address service to use for determining
#' the street address. Default: NULL skips this step. Accepted values: "photon".
#' @param verbose Logical controlling the generation of progress messages.
#' 
#' @return Updated tibble of known locations.
#' 
#' @examples
#' \donttest{
#' library(MazamaLocationUtils)
#' 
#' # Fail gracefully if any resources are not available
#' try({
#' 
#'   # Set up standard directories and spatial data
#'   spatialDataDir <- tempdir() # typically "~/Data/Spatial"
#'   initializeMazamaSpatialUtils(spatialDataDir)
#' 
#'   locationTbl <- get(data("wa_monitors_500"))
#' 
#'   # Coulee City, WA
#'   lon <- -119.290904
#'   lat <- 47.611942
#' 
#'   locationTbl <- 
#'     locationTbl %>%
#'     table_addLocation(lon, lat, distanceThreshold = 500)
#'   
#'   dplyr::glimpse(locationTbl)
#'   
#' }, silent = FALSE)
#' }
#' @seealso [table_addSingleLocation()]
#' @seealso [table_removeRecord()]
#' @seealso [table_updateSingleRecord()]
#' @rdname table_addLocation
#' @export 
#' @importFrom MazamaCoreUtils stopIfNull
#' @importFrom dplyr bind_rows
#' 
table_addLocation <- function(
  locationTbl = NULL,
  longitude = NULL,
  latitude = NULL,
  distanceThreshold = NULL,
  stateDataset = "NaturalEarthAdm1",
  elevationService = NULL,
  addressService = NULL,
  verbose = TRUE
) {
  
  validateMazamaSpatialUtils()
  
  # ----- Validate parameters --------------------------------------------------
  
  MazamaLocationUtils::validateLocationTbl(locationTbl, locationOnly = FALSE)
  MazamaCoreUtils::stopIfNull(longitude)
  MazamaCoreUtils::stopIfNull(latitude)
  MazamaCoreUtils::stopIfNull(distanceThreshold)
  MazamaCoreUtils::stopIfNull(stateDataset)
  
  if ( length(longitude) != length(latitude) )
    stop("longitude and latitude must have the same length")
  
  if ( length(distanceThreshold) != 1 )
    stop("distanceThreshold must be of length 1")
  
  incomingCount <- length(longitude)
  
  # Remove location pairs if either is missing
  naMask <- is.na(longitude) | is.na(latitude)
  longitude <- longitude[!naMask]
  latitude <- latitude[!naMask]
  
  if ( verbose && any(naMask, na.rm = TRUE) ) {
    message(sprintf(
      "%d of the %d requested locations have NA values",
      length(which(naMask)), incomingCount
    ))
  }
  
  # Validate remaining locations
  MazamaLocationUtils::validateLonsLats(longitude, latitude)
  
  if ( !exists(stateDataset) ) {
    stop(paste0(
      "You must load \"stateDataset\" with: \n",
      "  loadSpatialData(\"", stateDataset, "\")\n"
    ))
  }
  
  if ( !is.null(addressService) ) {
    if ( !is.character(addressService) ) {
      stop("Currently, only the \"photon\" address service is supported.")
    } else {
      if ( tolower(addressService) != "photon" )
        stop("Currently, only the \"photon\" address service is supported.")
    }
  }
  
  if ( !is.null(elevationService) ) {
    if ( !is.character(elevationService) ) {
      stop("Currently, only the \"usgs\" address service is supported.")
    } else {
      if ( tolower(elevationService) != "usgs" )
        stop("Currently, only the \"usgs\" address service is supported.")
    }
  }
  
  # ----- Reduce to only new locations -----------------------------------------

  foundLocationID <- table_getLocationID(locationTbl, longitude, latitude, distanceThreshold)
  existingCount <- sum(!is.na(foundLocationID))
  
  if ( verbose && existingCount > 0 ) {
    message(sprintf(
      "%d of the %d requested locations are already found in the locationTbl",
      existingCount, incomingCount
    ))
  }
  
  longitude <- longitude[is.na(foundLocationID)]
  latitude <- latitude[is.na(foundLocationID)]
  
  # ----- Loop over new locations ----------------------------------------------

  # NOTE:  Yes, this is a loop!
  # NOTE:  But the task of generating spatial data takes many seconds for each
  # NOTE:  iteration so there is nothing to be gained by making the code less
  # NOTE:  readable with functional programming style.
  
  for ( i in seq_along(longitude) ) {
    
    result <- try({
      
      if ( verbose ) {
        message(sprintf(
          "Working on %.7f, %.7f ...",
          longitude[i], latitude[i]
        ))
      }
      
      locationTbl <- table_addSingleLocation(
        locationTbl = locationTbl,
        longitude = longitude[i],
        latitude = latitude[i],
        distanceThreshold = distanceThreshold,
        stateDataset = stateDataset,
        elevationService = elevationService,
        addressService = addressService,
        verbose = verbose
      )
      
    }, silent = TRUE)
    
    if ( "try-error" %in% result ) {
      # Warn but don't stop
      warning(sprintf(
        "Skipping with error: ", geterrmessage()
      ))
    }
    
    # Be _somewhat_ careful with memory
    if ( 1 %% 10 == 0 ) {
      gc()
    }
    
  }

  # ----- Return ---------------------------------------------------------------
  
  return(locationTbl)
  
}

# ===== DEBUGGING ==============================================================

if ( FALSE ) {
  
  
  locationTbl = table_initialize()
  longitude = -113.7868
  latitude = 42.60782
  distanceThreshold = 500
  stateDataset = "NaturalEarthAdm1"
  elevationService = "usgs"
  addressService = "photon"
  verbose = TRUE
  
  
}

context("table_addLocation")

test_that("table_addLocation() works", {
  skip_on_cran()
  skip_on_travis()
  
  initializeMazamaSpatialUtils()
  locationTbl <- get(data("wa_monitors_500"))
  # Add a new record
  testTbl <- table_addLocation(locationTbl, -111.111, 44.444, 500, verbose = FALSE)
  expect_equal(nrow(locationTbl) + 1, nrow(testTbl))
  
  # Don't add it a second time
  testTbl <- table_addLocation(testTbl, -111.111, 44.444, 500, verbose = FALSE)
  expect_equal(nrow(locationTbl) + 1, nrow(testTbl))
})

#' @title Idaho monitor locations dataset
#' @format A tibble with 30 rows and 13 columns of data.
#' @description The `id_monitor_500` dataset provides a set of known
#' locations associated with Idaho state air quality monitors.
#' This dataset was generated on 2023-10-24 by running:
#' 
#' \preformatted{
#' library(AirMonitor)
#' library(MazamaLocationUtils)
#' 
#' initializeMazamaSpatialUtils()
#' setLocationDataDir("./data")
#' 
#' monitor <- monitor_loadLatest() \%>\% monitor_filter(stateCode == "ID")
#' lons <- monitor$meta$longitude
#' lats <- monitor$meta$latitude
#' 
#' table_initialize() \%>\%
#'   table_addLocation(
#'     lons, lats,
#'     distanceThreshold = 500,
#'     elevationService = "usgs",
#'     addressService = "photon"
#'   ) \%>\%
#'   table_save("id_monitors_500")
#' }
#' 
#' @seealso [or_monitors_500()]
#' @seealso [wa_monitors_500()]
"id_monitors_500"


#' @title Oregon monitor locations dataset
#' @format A tibble with 64 rows and 13 columns of data.
#' @description The `or_monitor_500` dataset provides a set of known
#' locations associated with Oregon state air quality monitors.
#' This dataset was generated on 2023-10-24 by running:
#' 
#' \preformatted{
#' library(AirMonitor)
#' library(MazamaLocationUtils)
#' 
#' initializeMazamaSpatialUtils()
#' setLocationDataDir("./data")
#'
#' monitor <- monitor_loadLatest() \%>\% monitor_filter(stateCode == "OR")
#' lons <- monitor$meta$longitude
#' lats <- monitor$meta$latitude
#' 
#' table_initialize() \%>\%
#'   table_addLocation(
#'     lons, lats, 
#'       distanceThreshold = 500,
#'       elevationService = "usgs",
#'       addressService = "photon"
#'     ) \%>\%
#'   table_save("or_monitors_500")
#' }
#' 
#' @seealso [id_monitors_500()]
#' @seealso [wa_monitors_500()]
"or_monitors_500"


#' @title Wshington monitor locations dataset
#' @format A tibble with 78 rows and 13 columns of data.
#' @description The `wa_monitor_500` dataset provides a set of known
#' locations associated with Washington state air quality monitors.
#' This dataset was generated on 2023-10-24 by running:
#' 
#' \preformatted{
#' library(AirMonitor)
#' library(MazamaLocationUtils)
#' 
#' initializeMazamaSpatialUtils()
#' setLocationDataDir("./data")
#' 
#' monitor <- monitor_loadLatest() \%>\% monitor_filter(stateCode == "WA")
#' lons <- monitor$meta$longitude
#' lats <- monitor$meta$latitude
#' 
#' table_initialize() \%>\%
#'   table_addLocation(
#'     lons, lats,
#'     distanceThreshold = 500,
#'     elevationService = "usgs",
#'     addressService = "photon"
#'   ) \%>\%
#'   table_save("wa_monitors_500")
#' }
#' 
#' @seealso [id_monitors_500()]
#' @seealso [or_monitors_500()]
"wa_monitors_500"


#' @title Washington monitor metadata dataset
#' @format A tibble with 92 rows and 29 columns of data.
#' @description The `wa_pwfsl_meta` dataset provides a set of Washington
#' state air quality monitor metadata used by the USFS AirFire group.
#' This dataset was generated on 2023-10-24 by running:
#' 
#' \preformatted{
#' library(AirMonitor)
#' 
#' wa_airfire_meta <-
#'   airnow_loadLatest() \%>\%
#'   monitor_filter(stateCode == "WA") \%>\%
#'   monitor_getMeta() \%>\%
#'   # On 2023-10-24, this metdata still uses zip instead of postalCode
#'   dplyr::rename(postalCode = zip) \%>\%
#'   # Remove internal fields
#'   dplyr::select(-dplyr::starts_with("airnow_"))
#' 
#' save(wa_airfire_meta, file = "data/wa_airfire_meta.rda")
#' }
"wa_airfire_meta"

#'
#' @title Find distances between adjacent locations in a known locations table
#' 
#' @description Calculate distances between all locations within a known
#' locations table and return a tibble with the row indices and separation 
#' distances of those records separated by less than `distanceThreshold` 
#' meters. Records are returned in order of distance.
#' 
#' It is useful when working with new metadata tables to identify adjacent
#' locations early on so that decisions can be made about the appropriateness
#' of the specified `distanceThreshold`.
#' 
#' @note The measure `"cheap"` may be used to speed things up depending on
#' the spatial scale being considered. Distances calculated with 
#' `measure = "cheap"` will vary by a few meters compared with those 
#' calculated using `measure = "geodesic"`.
#' 
#' @param locationTbl Tibble of known locations.
#' @param distanceThreshold Distance in meters.
#' @param measure One of "haversine" "vincenty", "geodesic", or "cheap" 
#' specifying desired method of geodesic distance calculation. 
#' 
#' See [geodist::geodist()] for details.
#' 
#' @return Tibble of row indices and distances for those locations separated by
#' less than `distanceThreshold` meters.
#' 
#' @examples 
#' library(MazamaLocationUtils)
#' 
#' meta <- wa_airfire_meta
#' 
#' # Any locations closer than 2 km?
#' table_findAdjacentDistances(meta, distanceThreshold = 2000)
#' 
#' # How about 4 km?
#' table_findAdjacentDistances(meta, distanceThreshold = 4000)
#' 
#' 
#' @rdname table_findAdjacentDistances
#' @export 
#' @importFrom MazamaCoreUtils stopIfNull
#' @importFrom dplyr tibble filter all_of
#' @importFrom rlang .data
#' 
table_findAdjacentDistances <- function(
  locationTbl = NULL,
  distanceThreshold = NULL,
  measure = c("geodesic", "haversine", "vincenty", "cheap")
) {
  
  # ----- Validate parameters --------------------------------------------------
  
  MazamaLocationUtils::validateLocationTbl(locationTbl, locationOnly = TRUE)
  MazamaCoreUtils::stopIfNull(distanceThreshold)
  measure <- match.arg(measure)
  
  if ( !is.numeric(distanceThreshold) )
    stop("Parameter 'distanceThreshold' must be a numeric value.")
  
  # ----- Calculate distances --------------------------------------------------

  # NOTE:  Suppress annoying message:
  #
  # Maximum distance is > 100km. The 'cheap' measure is inaccurate over such
  # large distances, you'd likely be better using a different 'measure'.
  
  suppressMessages({

    distance <-
      geodist::geodist(
        x = cbind(
          "x" = locationTbl$longitude,
          "y" = locationTbl$latitude
        ),
        paired = FALSE,
        sequential = FALSE,
        pad = FALSE,
        measure = measure
      )
    
  })
  
  # NOTE:  distance matrix is nrow(locationTbl) X length(longitude)
  
  # ----- Check for locations that are too close -------------------------------
  
  # Get distances that are less than the given distanceThreshold
  # NOTE: the distance between a location and itself is always zero
  distanceMask <- (distance != 0) & (distance < distanceThreshold )
  
  # Select the locations that are "too close".
  adjacentIndices <- which(distanceMask, arr.ind = TRUE)
  
  if ( nrow(adjacentIndices) == 0 ) {
    
    # Return an empty tibble
    adjacentDistanceTable <- 
      dplyr::tibble(
        row1 = 1,
        row2 = 1,
        distance = as.numeric(NA)
      ) %>% dplyr::filter(
        .data$row1 == -999
      )
    
  } else {
    
    # NOTE:  If location a and b are too close, two entries will be returned:
    # NOTE:        row  col
    # NOTE:   [#,]  a    b
    # NOTE:    ...
    # NOTE:   [#,]  b    a
    #
    # NOTE:  While often the case, there is no guarantee that complementary
    # NOTE:  rows will be sequential. The next couple of lines
    # NOTE:  find the rows that have the same indices and reduce the table to
    # NOTE:  only unique pairs.
    
    sortedAdjacentIndices <- t(apply(adjacentIndices, 1, sort))
    
    # NOTE:  We have to be careful when there is only one pair so that we don't
    # NOTE:  unintentionally get simplified to a vector instead of a matrix.
    
    if ( nrow(sortedAdjacentIndices) == 2 ) {
      adjacentMatrix <- matrix(sortedAdjacentIndices[1,], nrow = 1, byrow = TRUE)
    } else {
      adjacentMatrix <- sortedAdjacentIndices[!duplicated(sortedAdjacentIndices),]
    }
    
    adjacentDistanceTable <- dplyr::tibble(
      row1 = adjacentMatrix[,1],
      row2 = adjacentMatrix[,2],
      distance = as.numeric(NA)
    )
    
    for ( i in seq_len(nrow(adjacentMatrix)) ) {
      adjacentDistanceTable$distance[i] <- 
        distance[adjacentDistanceTable$row1[i], adjacentDistanceTable$row2[i]]
    }
    
    adjacentDistanceTable <- 
      adjacentDistanceTable %>% 
      dplyr::arrange(.data$distance)
    
  }
  
  # ----- Return ---------------------------------------------------------------
  
  return(adjacentDistanceTable)
  
}

context("table_findAdjacentDistances")

test_that("bad/missing input generates an error", {
  
  # tbl not a dataframe
  expect_error(table_findAdjacentDistances(locationTbl = c("vector", "instead"), distanceThreshold = 10))
  # distanceThreshold not numeric
  expect_error(table_findAdjacentDistances(locationTbl = wa_airfire_meta, distanceThreshold = "bad radius"))
  
  # tbl missing
  expect_error(table_findAdjacentDistances(NULL, distanceThreshold = 10))
  # radius missing
  expect_error(table_findAdjacentDistances(locationTbl = wa_airfire_meta, distanceThreshold = NULL))
  
})

test_that("table missing latitude or longitude generates an error", {
  
  # No latitude
  testTbl <- wa_airfire_meta %>% dplyr::select(-latitude)
  expect_error(table_findAdjacentDistances(locationTbl = testTbl, distanceThreshold = 10))
  
  # No longitude
  testTbl <- wa_airfire_meta %>% dplyr::select(-longitude)
  expect_error(table_findAdjacentDistances(locationTbl = testTbl, distanceThreshold = 10))
  
})

test_that("passing distanceThreshold = 1000 returns empty tibble", {
  
  result <- table_findAdjacentDistances(wa_airfire_meta, distanceThreshold = 1000)
  
  # Is it the right type?
  expect_equal(class(result), c("tbl_df", "tbl", "data.frame"))
  
  # Is it empty?
  expect_equal(nrow(result), 0)
  
})

test_that("passing distanceThreshold = 4000 returns expected output", {
  
  result <- table_findAdjacentDistances(wa_airfire_meta, distanceThreshold = 4000)
  
  # Is it the right type?
  expect_equal(class(result), c("tbl_df", "tbl", "data.frame"))
  
  # Is it the expected tibble?
  correctTbl <- dplyr::tibble(row1 = c(46, 30, 39, 49), row2 = c(74, 46, 70, 74), distance = c(2412, 3329, 3599, 3628))
  expect_equal(result$row1, correctTbl$row1)
  expect_equal(result$row2, correctTbl$row2)
  expect_equal(round(result$distance, 0), correctTbl$distance)
  
})

#' @export
#'
#' @title Initialize MazamaSpatialUtils package
#'
#' @param spatialDataDir Directory where \pkg{MazamaSpatialUtils} datasets are found.
#'
#' @description Convenience function that wraps:
#'
#' \preformatted{
#'   MazamaSpatialUtils::setSpatialDataDir("~/Data/Spatial")
#'   MazamaSpatialUtils::loadSpatialData("EEZCountries.rda")
#'.  MazamaSpatialUtils::loadSpatialData("OSMTimezones.rda")
#'   MazamaSpatialUtils::loadSpatialData("NaturalEarthAdm1.rda")
#'   MazamaSpatialUtils::loadSpatialData("USCensusCounties.rda")
#' }
#' 
#' If spatial data has not yet been installed, an error is returned with an 
#' extended message detailing how to install the appropriate data.
#'
#' @examples
#' \donttest{
#' library(MazamaLocationUtils)
#' 
#' # Fail gracefully if any resources are not available
#' try({
#' 
#'   # Set up directory for spatial data
#'   spatialDataDir <- tempdir() # typically "~/Data/Spatial"
#'   MazamaSpatialUtils::setSpatialDataDir(spatialDataDir)
#' 
#'   exists("NaturalEarthAdm1")
#'   initializeMazamaSpatialUtils(spatialDataDir)
#'   exists("NaturalEarthAdm1")
#'   class(NaturalEarthAdm1)
#'   
#' }, silent = FALSE)
#' }

initializeMazamaSpatialUtils <- function(
  spatialDataDir = "~/Data/Spatial"
) {

  # ----- Validate Parameters --------------------------------------------------


  # ----- Load spatial data ----------------------------------------------------

  # Is everything already initialized?
  result <- try({
    validateMazamaSpatialUtils() # swallow warning messages
  }, silent = TRUE)
  
  if ( "try-error" %in% class(result) ) {
    
    # Not initialized, so try to initialize
    result <- try({
      MazamaSpatialUtils::setSpatialDataDir(spatialDataDir)
      # Install if not found
      MazamaSpatialUtils::installSpatialData("EEZCountries")
      MazamaSpatialUtils::installSpatialData("OSMTimezones")
      MazamaSpatialUtils::installSpatialData("NaturalEarthAdm1")
      MazamaSpatialUtils::installSpatialData("USCensusCounties")
      # Now load
      MazamaSpatialUtils::loadSpatialData("EEZCountries.rda")
      MazamaSpatialUtils::loadSpatialData("OSMTimezones.rda")
      MazamaSpatialUtils::loadSpatialData("NaturalEarthAdm1.rda")
      MazamaSpatialUtils::loadSpatialData("USCensusCounties.rda")
    }, silent = TRUE)
    
    if ( "try-error" %in% class(result) ) {
      # Installation failed so spit out warning messages
      validateMazamaSpatialUtils()
    }
    
  }
  
  # ----- Return ---------------------------------------------------------------
  
  return(invisible(NULL))
  
}

#' @title Load a known location table
#' @description Load a tibble of known locations from the preferred directory.
#' 
#' The known location table must be named either `<collectionName>.rda`
#' or `<collectionName>.csv`. If both are found, only 
#' `<collectionName>.rda` will be loaded to ensure that columns will have
#' the proper type assigned.
#' 
#' @param collectionName Character identifier for this table.
#' @return Tibble of known locations.
#' @examples
#' library(MazamaLocationUtils)
#' 
#' # Set the directory for saving location tables
#' setLocationDataDir(tempdir())
#' 
#' # Load an example table and check the dimensions
#' locationTbl <- get(data("wa_monitors_500"))
#' dim(locationTbl)
#' 
#' # Save it as "table_load_example"
#' table_save(locationTbl, "table_load_example")
#' 
#' # Load it and check the dimensions
#' my_table <- table_load("table_load_example")
#' dim(my_table)
#' 
#' # Check the locationDataDir
#' list.files(getLocationDataDir(), pattern = "table_load_example")
#' @seealso 
#'  [setLocationDataDir()]
#' @rdname table_load
#' @export 
#' @importFrom MazamaCoreUtils stopIfNull
table_load <- function(
  collectionName = NULL
) {
  
  # ----- Validate parameters --------------------------------------------------
  
  MazamaCoreUtils::stopIfNull(collectionName)
  
  dataDir <- getLocationDataDir()
  
  # ----- Load data ------------------------------------------------------------
  
  result <- try({
    
    # First try the .rda version
    fileName <- paste0(collectionName, ".rda")
    filePath <- file.path(dataDir, fileName)
    if ( file.exists(filePath) ) {
      locationTbl <- get(load(filePath))
    } else {
      # Then try the .csv version
      fileName <- paste0(collectionName, ".csv")
      filePath <- file.path(dataDir, fileName)
      if ( file.exists(filePath) ) {
        locationTbl <- readr::read_csv(
          filePath,
          progress = FALSE,
          show_col_types = FALSE
        )
      } else {
        stop(sprintf("No '%s' collection table found in ", collectionName, dataDir))
      }
    }
    
  }, silent = TRUE)
  
  if ( "try-error" %in% class(result) ) {
    stop(sprintf("Could not load %s", filePath))
  }
  
  # ----- Return ---------------------------------------------------------------
  
  return(locationTbl)
  
}

#' @title Get census block data from the FCC API
#' @description The FCC Block API is used get census block, county, and state FIPS associated with
#' the `longitude` and `latitude`. The following list of data
#' is returned:
#' \itemize{
#' \item{`stateCode`}
#' \item{`countyName`}
#' \item{`censusBlock`}
#' }
#' The data from this function should be considered to be the gold standard for state and county.
#' i.e. this information could and should be used to override information we get elsewhere.
#' @param longitude Single longitude in decimal degrees E.
#' @param latitude Single latitude in decimal degrees N.
#' @param censusYear Year the census was taken.
#' @param verbose Logical controlling the generation of progress messages.
#' @return List of census block/county/state data.
#' @examples 
#' \donttest{
#' library(MazamaLocationUtils)
#' 
#' # Fail gracefully if any resources are not available
#' try({
#' 
#'   # Wenatchee
#'   lon <- -120.325278
#'   lat <- 47.423333
#' 
#'   censusList <- location_getCensusBlock(lon, lat)
#'   str(censusList)
#'   
#' }, silent = FALSE)
#' }
#' @references \url{https://geo.fcc.gov/api/census/}
#' @rdname location_getCensusBlock
#' @export 
#' @importFrom utils capture.output
#' @importFrom stringr str_detect str_sub str_subset
#' 
location_getCensusBlock <- function(
  longitude = NULL,
  latitude = NULL,
  censusYear = 2010,
  verbose = TRUE
) {
  
  # ----- Validate parameters --------------------------------------------------
  
  validateLonLat(longitude, latitude)  
  
  # ----- Get FCC API data ----------------------------------------------
  
  # https://geo.fcc.gov/api/census/block/find?latitude=47.423333&longitude=-120.325278&censusYear=2010&showall=false&format=json
  
  # Create url
  
  url <- httr::parse_url("https://geo.fcc.gov/api/census/block/find")
  
  url$query <- list(
    latitude = latitude, 
    longitude = longitude, 
    censusYear = censusYear,
    showall = "False",
    format = "json"
  )
  
  # Get and parse the return
  
  r <- httr::GET(httr::build_url(url))
  if ( httr::http_error(r) ) {
    
    censusBlock <- list()
    
    if ( verbose ) {
      warning(sprintf(
        "FCC Census Block service failed for URL %s", 
        httr::build_url(url)
      ))
    }
    
  } else {
    
    returnObj <- httr::content(r)
    censusList <- list(
      stateCode = toupper(returnObj$State$code),
      countyName = as.character(returnObj$County$name),
      censusBlock = as.character(returnObj$Block$FIPS)
    )
    
    if (is.null(censusList$censusBlock) ) {
      
      if ( verbose ) {
        warning(sprintf(
          "FCC Census Block service returned NULL values."
        ))
      }
      
    }
    
  }
  
  # ----- Return ---------------------------------------------------------------
  
  return(censusList)
  
}

#' @title Remove location records from a table
#' @description Incoming `locationID` values are compared 
#' against the incoming `locationTbl` and any matches are removed.
#' @param locationTbl Tibble of known locations.
#' @param locationID Vector of `locationID` strings.
#' @param verbose Logical controlling the generation of progress messages.
#' @return Updated tibble of known locations.
#' @examples
#' library(MazamaLocationUtils)
#' 
#' locationTbl <- get(data("wa_monitors_500"))
#' dim(locationTbl)
#' 
#' # Wenatchee
#' lon <- -120.325278
#' lat <- 47.423333
#' 
#' # Get the locationID first
#' locationID <- table_getLocationID(locationTbl, lon, lat, distanceThreshold = 500)
#' 
#' # Remove it
#' locationTbl <- table_removeRecord(locationTbl, locationID)
#' dim(locationTbl)
#' 
#' # Test
#' table_getLocationID(locationTbl, lon, lat, distanceThreshold = 500)
#' 
#' @seealso [table_addLocation()]
#' @seealso [table_addSingleLocation()]
#' @seealso [table_updateSingleRecord()]
#' @rdname table_removeRecord
#' @export 
#' @importFrom rlang .data
table_removeRecord <- function(
  locationTbl = NULL,
  locationID = NULL,
  verbose = TRUE
) {
  
  # ----- Validate parameters --------------------------------------------------
  
  MazamaLocationUtils::validateLocationTbl(locationTbl, locationOnly = FALSE)
  MazamaCoreUtils::stopIfNull(locationID)

  # ----- Subset locationTbl ---------------------------------------------------
  
  locationTbl <-
    locationTbl %>%
    dplyr::filter(!.data$locationID %in% !!locationID)

  # ----- Return ---------------------------------------------------------------
  
  return(locationTbl)
  
}

context("table_removeRecord")

test_that("table_removeRecord() works", {
  
  locationTbl <- get(data("wa_monitors_500"))
  
  # First three from wa_monitors_500
  locationID <- c("c2913q48uk", "c28f8z9xq8", "c23hfxrdne")
  
  testTbl <- table_removeRecord(locationTbl, locationID, verbose = FALSE)
  expect_equal(nrow(locationTbl) - 3, nrow(testTbl))
  
})

#' @rdname table_getNearestLocation
#' @export
#' @importFrom MazamaCoreUtils stopIfNull
#' 
#' @title Return known locations
#' 
#' @param locationTbl Tibble of known locations.
#' @param longitude Vector of longitudes in decimal degrees E.
#' @param latitude Vector of latitudes in decimal degrees N.
#' @param distanceThreshold Distance in meters.
#' 
#' @return Tibble of known locations.
#' 
#' @description Returns a tibble of the known locations from `locationTbl` 
#' that are closest to the vector of target locations specified by `longitude`
#' and `latitude`. Only a single known location is returned for each
#' incoming target location. If no known location is found for a particular
#' incoming location, that record in the tibble will contain all `NA`.
#' 
#' @examples
#' library(MazamaLocationUtils)
#' 
#' locationTbl <- get(data("wa_monitors_500"))
#' 
#' # Wenatchee
#' lon <- -120.325278
#' lat <- 47.423333
#' 
#' # Too small a distanceThreshold will not find a match
#' table_getNearestLocation(locationTbl, lon, lat, distanceThreshold = 50) %>% str()
#' 
#' # Expanding the distanceThreshold will find one
#' table_getNearestLocation(locationTbl, lon, lat, distanceThreshold = 5000) %>% str()

table_getNearestLocation <- function(
  locationTbl = NULL,
  longitude = NULL,
  latitude = NULL,
  distanceThreshold = NULL
) {

  # ----- Validate parameters --------------------------------------------------

  MazamaLocationUtils::validateLocationTbl(locationTbl, locationOnly = TRUE)
  MazamaLocationUtils::validateLonsLats(longitude, latitude)
  MazamaCoreUtils::stopIfNull(distanceThreshold)

  distanceThreshold <- round(distanceThreshold)

  # ----- Subset ---------------------------------------------------------------

  targetLocationIDsTbl <- dplyr::tibble(
    locationID = table_getLocationID(locationTbl, longitude, latitude, distanceThreshold)
  )
  
  # NOTE:  dplyr::left_join() guarantees *at least* one record per X depending on
  # NOTE:  how many matches are found in Y. Because validateLocationTbl() ensures
  # NOTE:  that locationID is never duplicated in a locationTbl, this ends up
  # NOTE:  handling the situation of multiple target locations that refer to the
  # NOTE:  same record in locationTbl.
  
  subsetTbl <- dplyr::left_join(targetLocationIDsTbl, locationTbl, by = "locationID")

  # ----- Return ---------------------------------------------------------------

  return(subsetTbl)

}

#' @title Leaflet interactive map for known locations
#'
#' @param locationTbl Tibble of known locations.
#' @param maptype Optional name of leaflet ProviderTiles to use, e.g. `terrain`.
#' @param extraVars Character vector of addition `locationTbl` column names
#' to be shown in leaflet popups.  
#' @param jitter Amount to use to slightly adjust locations so that multiple
#' monitors at the same location can be seen. Use zero or `NA` to see
#' precise locations.
#' @param ... Additional arguments passed to `leaflet::addCircleMarkers()`.
#'
#' @description This function creates interactive maps that will be displayed in
#' RStudio's 'Viewer' tab. The default setting of `jitter` will move locations
#' randomly within an ~50 meter radius so that overlapping locations can be 
#' identified. Set `jitter = 0` to see precise locations.
#'
#' @details The `maptype` argument is mapped onto leaflet "ProviderTile"
#'   names. Current mappings include:
#' \itemize{
#' \item{"roadmap" => "OpenStreetMap"}
#' \item{"satellite" => "Esri.WorldImagery"}
#' \item{"terrain" => "Esri.WorldTopoMap"}
#' \item{"toner" => "Stamen.Toner"}
#' }
#'
#' If a character string not listed above is provided, it will be used as the
#' underlying map tile if available. See
#' \url{https://leaflet-extras.github.io/leaflet-providers/} for a list of
#' "provider tiles" to use as the background map.
#' 
#' @return A leaflet "plot" object which, if not assigned, is rendered in
#' Rstudio's 'Viewer' tab.
#'
#' @rdname table_leaflet
#' @export 
#' @importFrom MazamaCoreUtils stopIfNull
#' @importFrom rlang .data
#' @importFrom leaflet leaflet fitBounds addProviderTiles addCircleMarkers
#' 
#' @examples
#' \donttest{
#' if (interactive()) {
#'   # A table with all core metadata
#'   table_leaflet(wa_monitors_500)
#'   
#'   # A table missing some core metadata
#'   table_leaflet(
#'     wa_airfire_meta
#'   )
#' 
#'   # Customizing the map
#'   table_leaflet(
#'     wa_airfire_meta,
#'     extraVars = c("AQSID"),
#'     radius = 6,
#'     color = "black",
#'     weight = 2,
#'     fillColor = "red",
#'     fillOpacity = 0.3
#'   )
#' }
#' }

table_leaflet <- function(
  locationTbl = NULL,
  maptype = c("terrain", "roadmap", "satellite", "toner"),
  extraVars = NULL,
  jitter = 5e-4,
  ...
) {
  
  # ----- Validate parameters --------------------------------------------------
  
  MazamaLocationUtils::validateLocationTbl(locationTbl, locationOnly = TRUE)
  maptype <- match.arg(maptype)
  
  if ( !is.null(extraVars) ) {
    unrecognizedVars <- setdiff(extraVars, names(locationTbl))
    if ( length(unrecognizedVars) > 0 ) {
      stop("Variables in 'extraVars' not found in 'locationTbl'")
    }
  }
  
  # Filter out missing location data
  locationTbl <-
    locationTbl %>%
    dplyr::filter(!is.na(.data$latitude)) %>%
    dplyr::filter(!is.na(.data$longitude))
  
  hasCoreMetadata <- all(coreMetadataNames %in% names(locationTbl))
  
  if ( is.null(jitter) || is.na(jitter) || jitter == 0 ) {
    jitter <- 1e-9
  }

  # * argsList -----
  
  argsList <- list(...)
  
  argsList$lng <- jitter(locationTbl$longitude, amount = jitter)
  argsList$lat <- jitter(locationTbl$latitude, amount = jitter)
  
  # ----- Create base map ------------------------------------------------------
  
  # * providerTiles -----
  
  # Convert maptype to a character string that addProviderTiles can read
  if ( missing(maptype) || maptype == 'terrain') {
    providerTiles <- "Esri.WorldTopoMap"
  } else if ( maptype == "roadmap" ) {
    providerTiles <- "OpenStreetMap"
  } else if ( maptype == "toner" ) {
    providerTiles <- "Stamen.Toner"
  } else if (maptype == "satellite" ) {
    providerTiles <- "Esri.WorldImagery"
  } else {
    providerTiles <- maptype
  }
  
  # * zoom -----
  
  # Extract view information
  lonRange <- range(locationTbl$longitude, na.rm = TRUE)
  latRange <- range(locationTbl$latitude, na.rm = TRUE)
  
  # * base map -----
  
  argsList$map <- 
    leaflet::leaflet(locationTbl) %>%
    leaflet::fitBounds(lonRange[1], latRange[1], lonRange[2], latRange[2]) %>%
    leaflet::addProviderTiles(providerTiles)
  
  # ----- Add circle markers ---------------------------------------------------
  
  # * weight -----
  
  if ( !"weight" %in% argsList ) 
    argsList$weight <- 1
  
  # * popup text -----
  
  # Initialize empty popupText
  popupText <- vector("character", nrow(locationTbl))
  
  # Create coreText
  
  if ( hasCoreMetadata ) {
    
    # Use guaranteed fields
    coreText <- paste0(
      "<b>", locationTbl$locationName, "</b><br>",
      "locationID = ", locationTbl$locationID, "<br>",
      "longitude = ", locationTbl$longitude, ", ", "latitude = ", locationTbl$latitude, "<br>",
      "timezone = ", locationTbl$timezone, "<br>",
      "ISO = ", locationTbl$countryCode, ".", locationTbl$stateCode, "<br>",
      "county = ", locationTbl$countyName, "<br>"
    )
    
  } else {
    
    # Use reasonable best guesses to create 3 lines of core metadata
    
    tblNames <- names(locationTbl)
    
    # Bold location identifier at the top (in preference order)
    if ( "locationName" %in% tblNames ) {
      coreText_1 <-  paste0("<b>", locationTbl$locationName, "</b><br>")
    } else if ( "siteName" %in% tblNames ) {
      coreText_1 <-  paste0("<b>", locationTbl$siteName, "</b><br>")
    } else if ( "AQSID" %in% tblNames ) {
      coreText_1 <- paste0("<b>", locationTbl$AQSID, "</b><br>")
    } else if ( "aqsID" %in% tblNames ) {
      coreText_1 <- paste0("<b>", locationTbl$aqsID, "</b><br>")
    } else if ( "aqsid" %in% tblNames ) {
      coreText_1 <- paste0("<b>", locationTbl$aqsid, "</b><br>")
    } else {
      coreText_1 <- ""
    }
    
    # US EPA AQSID if not already found
    if ( "AQSID" %in% tblNames ) {
      coreText_2 <-  paste0("AQSID = ", locationTbl$AQSID, "<br>")
    } else if ( "aqsID" %in% tblNames ) {
      coreText_2 <-  paste0("AQSID = ", locationTbl$aqsID, "<br>")
    } else if ( "aqsid" %in% tblNames ) {
      coreText_2 <-  paste0("AQSID = ", locationTbl$aqsid, "<br>")
    } else {
      coreText_2 <- ""
    }

    # Location
    coreText_3 <- paste0(
      "longitude = ", locationTbl$longitude, ", ", "latitude = ", locationTbl$latitude, "<br>"
    )
    
    # Paste them together
    coreText <- paste0(coreText_1, coreText_2, coreText_3)
    
  }
  
  # Add extra vars
  for ( i in seq_along(popupText) ) {
    
    extraText <- vector("character", length(extraVars))
    for ( j in seq_along(extraVars) ) {
      var <- extraVars[j]
      extraText[j] <- paste0(var, " = ", locationTbl[i, var], "<br>")
    }
    extraText <- paste0(extraText, collapse = "")
    
    popupText[i] <- paste0(coreText[i], "<hr>", extraText)
    
  }
  
  locationTbl$popupText <- popupText
  
  argsList$popup <- popupText
  
  # * add markers -----
  
  map <- do.call(leaflet::addCircleMarkers, argsList)
  
  # ----- Return ---------------------------------------------------------------
  
  return(map)
  
}

---
title: "Introduction to MazamaLocationUtils"
author: "Jonathan Callahan"
date: "2023-10-24"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Introduction to MazamaLocationUtils}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, echo=FALSE}
knitr::opts_chunk$set(fig.width = 7, fig.height = 5)
```

# Background

This package is intended for use in data management activities
associated with fixed locations in space. The motivating fields include 
air and water quality monitoring where fixed sensors report at regular time 
intervals.

When working with environmental monitoring time series, one of the first things
you have to do is create unique identifiers for each individual time series. In 
an ideal world, each environmental time series would have both a 
`locationID` and a `deviceID` that uniquely identify the specific instrument 
making measurements and the physical location where measurements are made. A 
unique `timeseriesID` could
be produced as `locationID_deviceID`. Metadata associated with each
`timeseriesID` would contain basic information needed for downstream analysis
including at least:

`timeseriesID, locationID, deviceID, longitude, latitude, ...`

* An extended time series for a mobile sensor would group by `deviceID`.
* Multiple sensors placed at a single location could be be grouped by `locationID`.
* Maps would be created using `longitude, latitude`.
* Time series measurements would be accessed from a secondary `data` table with 
`timeseriesID` column names.

Unfortunately, we are rarely supplied with a truly unique and truly spatial 
`locationID`. Instead we often use `deviceID` or an associated non-spatial
identifier as a stand-in for `locationID`.

Complications we have seen include:

* GPS-reported longitude and latitude can have _jitter_ in the fourth or fifth 
decimal place making it challenging to use them to create a unique `locationID`.
* Sensors are sometimes _repositioned_ in what the scientist considers the "same 
location".
* Data from a single sensor goes through different processing pipelines using
different identifiers and is later brought together as two separate timeseries.
* The spatial scale of what constitutes a "single location" depends on the 
instrumentation and scientific question being asked.
* Deriving location-based metadata from spatial datasets is computationally 
intensive unless saved and identified with a unique `locationID`.
* Automated searches for spatial metadata occasionally produce incorrect results
because of the non-infinite resolution of spatial datasets and must be corrected
by hand.

# Functionality

A solution to all these problems is possible if we store spatial metadata in
simple tables in a standard directory. These tables will be referred to as 
_collections_. Location lookups can be performed with
geodesic distance calculations where a longitude-latitude pair is assigned to a pre-existing
_known location_ if it is within `distanceThreshold` meters of that location. 
These lookups will be extremely fast.

If no previously _known location_ is found, the relatively slow (seconds)
creation of a new _known location_ metadata record can be performed and then 
added to the growing collection.

For collections of stationary environmental monitors that only number in the 
thousands, this entire _collection_ can be stored as either a 
`.rda` or `.csv` file and will be under a megabyte in size making it fast to 
load. This small size also makes it possible to store multiple _known locations_ 
files, each created with different locations and different distance thresholds
to address the needs of different scientific studies.

# Example Usage

The package comes with some example _known locations_ tables.

Lets take some metadata we have for air quality monitors in Washington state and
create a _known locations_ table for them.

```{r load_data}
wa <- get(data("wa_airfire_meta", package = "MazamaLocationUtils"))
names(wa)
```

```{r load_data_hidden, eval = TRUE, echo = FALSE}
library(MazamaLocationUtils)
wa_monitors_500 <- 
  get(data("wa_monitors_500", package = "MazamaLocationUtils")) %>%
  dplyr::mutate(elevation = as.numeric(NA))
```

## Creating a Known Locations table

We can create a _known locations_ table for them with a minimum 500 meter
separation between distinct locations. _(NOTE:  This will take some time to
performa all the spatial queries.)_

To speed things up, we call `table_addLocation()` with defaults:
`elevationService = NULL, addressService = NULL`. This avoids these slow
web service requests and results in a table with `NA` for these columns.

```{r create_table, eval = FALSE, echo = TRUE}
library(MazamaLocationUtils)

# Initialize with standard directories
initializeMazamaSpatialUtils()
setLocationDataDir("./data")

wa_monitors_500 <-
  table_initialize() %>%
  table_addLocation(wa$longitude, wa$latitude, distanceThreshold = 500) 
```

At this point, our _known locations_ table contains only automatically generated
spatial metadata. 
 
```{r basic_columns}
dplyr::glimpse(wa_monitors_500, width = 75)
```

## Merging external metadata

Perhaps we would like to import some of the original metadata into our new 
table. This is a very common use case where non-spatial metadata like uniform
identifiers or owner information for a monitor can be added.

Just to make it interesting, let's assume that our _known locations_ table is
already large and we are only providing additional metadata for a subset of the
records.

```{r import_colmns}
# Use a subset of the wa metadata
wa_indices <- seq(5,65,5)
wa_sub <- wa[wa_indices,]

# Use a generic name for the location table
locationTbl <- wa_monitors_500

# Find the location IDs associated with our subset
locationID <- table_getLocationID(
  locationTbl, 
  longitude = wa_sub$longitude, 
  latitude = wa_sub$latitude, 
  distanceThreshold = 500
)

# Now add the "AQSID" column for our subset of locations
locationData <- wa_sub$AQSID
locationTbl <- table_updateColumn(
  locationTbl, 
  columnName = "AQSID", 
  locationID = locationID, 
  locationData = locationData
)

# Lets see how we did
locationTbl_indices <- table_getRecordIndex(locationTbl, locationID)
locationTbl[locationTbl_indices, c("city", "AQSID")]
```

Very nice. We have added `AQSID` to our known locations table for a more 
detailed description of each monitors' location.

## Finding known locations

The whole point of a known locations table is to speed up access to spatial
and other metadata. Here's how we can use it with a set of longitudes and
latitudes that are not currently in our table.

```{r new_locations}
# Create new locations near our known locations
lons <- jitter(wa_sub$longitude) 
lats <- jitter(wa_sub$latitude)

# Any known locations within 50 meters?
table_getNearestLocation(
  wa_monitors_500,
  longitude = lons,
  latitude = lats,
  distanceThreshold = 50
) %>% dplyr::pull(city)

# Any known locations within 250 meters
table_getNearestLocation(
  wa_monitors_500,
  longitude = lons,
  latitude = lats,
  distanceThreshold = 250
) %>% dplyr::pull(city)

# How about 5000 meters?
table_getNearestLocation(
  wa_monitors_500,
  longitude = lons,
  latitude = lats,
  distanceThreshold = 5000
) %>% dplyr::pull(city)

```

# Standard Setup

Before using **MazamaLocationUtils** you must first install 
**MazamaSpatialUtils** and then install core spatial data with:

```{r MSU_setup, echo = TRUE, eval = FALSE}
  library(MazamaSpatialUtils)
  setSpatialDataDir("~/Data/Spatial")
  
  installSpatialData("EEZCountries")
  installSpatialData("OSMTimezones")
  installSpatialData("NaturalEarthAdm1")
  installSpatialData("USCensusCounties")
```

The `initializeMazamaSpatialData()` function by default assumes spatial data are 
installed in the standard location and is just a wrapper for:

```{r standard_setup, echo = TRUE, eval = FALSE}
  MazamaSpatialUtils::setSpatialDataDir("~/Data/Spatial")
  
  MazamaSpatialUtils::loadSpatialData("EEZCountries.rda")
  MazamaSpatialUtils::loadSpatialData("OSMTimezones.rda")
  MazamaSpatialUtils::loadSpatialData("NaturalEarthAdm1.rda")
  MazamaSpatialUtils::loadSpatialData("USCensusCounties.rda")
```

Once the required datasets have been installed, the easiest way to set things 
up each session is with:

```{r easy_setup, echo = TRUE, eval = FALSE}
  library(MazamaLocationUtils)
  initializeMazamaSpatialData()
  setLocationDataDir("~/Data/KnownLocations")
```

Every time you `table_save()` your location table, a backup will be created
so you can experiment without losing your work. File sizes are pretty tiny
so you don't have to worry about filling up your disk.

----

Best wishes for well organized spatial metadata!

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{validateLonsLats}
\alias{validateLonsLats}
\title{Validate longitude and latitude vectors}
\usage{
validateLonsLats(longitude = NULL, latitude = NULL, na.rm = FALSE)
}
\description{
See \code{\link[MazamaCoreUtils:validateLonsLats]{MazamaCoreUtils::validateLonsLats()}} for details.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MazamaLocationUtils-package.R
\docType{data}
\name{LocationDataDir}
\alias{LocationDataDir}
\title{Directory for location data}
\format{
Absolute path string.
}
\description{
This package maintains an internal directory path which
users can set using \code{setLocationDataDir()}. All package functions use
this directory whenever known location tables are accessed.

The default setting when the package is loaded is \code{getwd()}.
}
\seealso{
\code{\link[=getLocationDataDir]{getLocationDataDir()}}

\code{\link[=setLocationDataDir]{setLocationDataDir()}}
}
\keyword{environment}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clusterByDistance.R
\name{clusterByDistance}
\alias{clusterByDistance}
\title{Add distance-clustering information to a dataframe}
\usage{
clusterByDistance(
  tbl,
  clusterDiameter = 1000,
  lonVar = "longitude",
  latVar = "latitude",
  maxClusters = 50
)
}
\arguments{
\item{tbl}{Tibble with geolocation information.}

\item{clusterDiameter}{Diameter in meters used to determine the number of
clusters (see description).}

\item{lonVar}{Name of longitude variable in the incoming tibble.}

\item{latVar}{Name of the latitude variable in the incoming tibble.}

\item{maxClusters}{Maximum number of clusters to try.}
}
\value{
Input tibble with additional columns: \verb{clusterLon, clusterLat, clusterID}.
}
\description{
Distance clustering is used to identify unique deployments of a
sensor in an environmental monitoring field study. GPS-reported locations can
be jittery and result in a sensor self-reporting from a cluster of nearby locations.
Clustering helps resolve this by assigning a single location to the cluster.

Standard \code{kmeans} clustering does not work well when clusters can have
widely differing numbers of members. A much better result is acheived with
the Partitioning Around Medoids method available in \code{cluster::pam()}.

The value of \code{clusterDiameter} is compared with the output of
\code{cluster::pam(...)$clusinfo[,'av_diss']} to determine the number of clusters.
}
\note{
In most applications, the \code{\link[=table_addClustering]{table_addClustering()}} function should be
used as it implements two-stage clustering using \code{clusterbyDistance()}.
}
\examples{
library(MazamaLocationUtils)

# Fremont, Seattle 47.6504, -122.3509
# Magnolia, Seattle 47.6403, -122.3997
# Downtown Seattle 47.6055, -122.3370

fremont_x <- jitter(rep(-122.3509, 10), .0005)
fremont_y <- jitter(rep(47.6504, 10), .0005)

magnolia_x <- jitter(rep(-122.3997, 8), .0005)
magnolia_y <- jitter(rep(47.6403, 8), .0005)

downtown_x <- jitter(rep(-122.3370, 3), .0005)
downtown_y <- jitter(rep(47.6055, 3), .0005)

# Apply clustering
tbl <-
  dplyr::tibble(
    longitude = c(fremont_x, magnolia_x, downtown_x),
    latitude = c(fremont_y, magnolia_y, downtown_y)
  ) \%>\%
  clusterByDistance(
    clusterDiameter = 1000
  )

plot(tbl$longitude, tbl$latitude, pch = tbl$clusterID)
}
\references{
\href{https://working-with-data.mazamascience.com/2021/07/15/when-k-means-clustering-fails/}{When k-means clustering fails}
}
\seealso{
\code{\link[=table_removeRecord]{table_removeRecord()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_updateColumn.R
\name{table_updateColumn}
\alias{table_updateColumn}
\title{Update a column of metadata in a table}
\usage{
table_updateColumn(
  locationTbl = NULL,
  columnName = NULL,
  locationID = NULL,
  locationData = NULL,
  verbose = TRUE
)
}
\arguments{
\item{locationTbl}{Tibble of known locations.}

\item{columnName}{Name of an existing/new column in \code{locationTbl} whose data
will be updated/created.}

\item{locationID}{Vector of \code{locationID} strings.}

\item{locationData}{Vector of data to be inserted at records identified by
\code{locationID}.}

\item{verbose}{Logical controlling the generation of progress messages.}
}
\value{
Updated tibble of known locations.
}
\description{
Updates records in a location table. Records are identified
by \code{locationID} and the data found in \code{locationData} is used to
replace any existing value in the \code{columnName} column.
\code{locationID} and \code{locationData} must be of the same length.
Any \code{NA} values in \code{locationID} will be ignored.

If \code{columnName} is not a named column within \code{locationTbl}, a new
column will be created.
}
\examples{
library(MazamaLocationUtils)

locationTbl <- get(data("wa_monitors_500"))
wa <- get(data("wa_airfire_meta"))

# We will merge some metadata from wa into locationTbl

# Record indices for wa
wa_indices <- seq(5,65,5)
wa_sub <- wa[wa_indices,]

locationID <- 
  table_getLocationID(
    locationTbl, 
    wa_sub$longitude, 
    wa_sub$latitude, 
    distanceThreshold = 500
  )
  
locationData <- wa_sub$AQSID

locationTbl <- 
  table_updateColumn(locationTbl, "AQSID", locationID, locationData)

# Look at the data we attempted to merge
wa$AQSID[wa_indices]

# And two columns from the updated locationTbl
locationTbl_indices <- table_getRecordIndex(locationTbl, locationID)
locationTbl[locationTbl_indices, c("city", "AQSID")]

}
\seealso{
\code{\link[=table_addColumn]{table_addColumn()}}

\code{\link[=table_removeColumn]{table_removeColumn()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/location_getSingleElevation_USGS.R
\name{location_getSingleElevation_USGS}
\alias{location_getSingleElevation_USGS}
\title{Get elevation data from a USGS web service}
\usage{
location_getSingleElevation_USGS(
  longitude = NULL,
  latitude = NULL,
  verbose = TRUE
)
}
\arguments{
\item{longitude}{Single longitude in decimal degrees E.}

\item{latitude}{Single latitude in decimal degrees N.}

\item{verbose}{Logical controlling the generation of progress messages.}
}
\value{
Numeric elevation value.
}
\description{
USGS APIs are used to determine the elevation in meters
associated with the \code{longitude} and \code{latitude}.

\emph{Note: The conversion factor for meters to feet is 3.28084.}
}
\examples{
\donttest{
library(MazamaLocationUtils)

# Fail gracefully if any resources are not available
try({

  # Wenatchee
  longitude <- -120.325278
  latitude <- 47.423333

  location_getSingleElevation_USGS(longitude, latitude)
  
}, silent = FALSE)
}

}
\references{
\url{https://apps.nationalmap.gov/epqs/}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{validateLonLat}
\alias{validateLonLat}
\title{Validate longitude and latitude values}
\usage{
validateLonLat(longitude = NULL, latitude = NULL)
}
\description{
See \code{\link[MazamaCoreUtils:validateLonLat]{MazamaCoreUtils::validateLonLat()}} for details.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_getNearestLocation.R
\name{table_getNearestLocation}
\alias{table_getNearestLocation}
\title{Return known locations}
\usage{
table_getNearestLocation(
  locationTbl = NULL,
  longitude = NULL,
  latitude = NULL,
  distanceThreshold = NULL
)
}
\arguments{
\item{locationTbl}{Tibble of known locations.}

\item{longitude}{Vector of longitudes in decimal degrees E.}

\item{latitude}{Vector of latitudes in decimal degrees N.}

\item{distanceThreshold}{Distance in meters.}
}
\value{
Tibble of known locations.
}
\description{
Returns a tibble of the known locations from \code{locationTbl}
that are closest to the vector of target locations specified by \code{longitude}
and \code{latitude}. Only a single known location is returned for each
incoming target location. If no known location is found for a particular
incoming location, that record in the tibble will contain all \code{NA}.
}
\examples{
library(MazamaLocationUtils)

locationTbl <- get(data("wa_monitors_500"))

# Wenatchee
lon <- -120.325278
lat <- 47.423333

# Too small a distanceThreshold will not find a match
table_getNearestLocation(locationTbl, lon, lat, distanceThreshold = 50) \%>\% str()

# Expanding the distanceThreshold will find one
table_getNearestLocation(locationTbl, lon, lat, distanceThreshold = 5000) \%>\% str()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_filterByDistance.R
\name{table_filterByDistance}
\alias{table_filterByDistance}
\title{Return known locations near a target location}
\usage{
table_filterByDistance(
  locationTbl = NULL,
  longitude = NULL,
  latitude = NULL,
  distanceThreshold = NULL,
  measure = c("geodesic", "haversine", "vincenty", "cheap")
)
}
\arguments{
\item{locationTbl}{Tibble of known locations.}

\item{longitude}{Target longitude in decimal degrees E.}

\item{latitude}{Target latitude in decimal degrees N.}

\item{distanceThreshold}{Distance in meters.}

\item{measure}{One of "haversine" "vincenty", "geodesic", or "cheap"
specifying desired method of geodesic distance calculation.}
}
\value{
Tibble of known locations.
}
\description{
Returns a tibble of the known locations from \code{locationTbl}
that are within \code{distanceThreshold} meters of the target location
specified by \code{longitude} and \code{latitude}.
}
\note{
Only a single target location is allowed.
}
\examples{
library(MazamaLocationUtils)

locationTbl <- get(data("wa_monitors_500"))

# Too small a distanceThreshold will not find a match
locationTbl \%>\%
  table_filterByDistance(
    longitude = -117.3647, 
    latitude = 47.6725, 
    distanceThreshold = 10
  ) \%>\% 
  dplyr::glimpse()

# Expanding the distanceThreshold will find several
locationTbl \%>\%
  table_filterByDistance(
    longitude = -117.3647, 
    latitude = 47.6725, 
    distanceThreshold = 10000
  ) \%>\%
  dplyr::glimpse()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_getDistanceFromTarget.R
\name{table_getDistanceFromTarget}
\alias{table_getDistanceFromTarget}
\title{Return distances and directions from a target location to known locations}
\usage{
table_getDistanceFromTarget(
  locationTbl = NULL,
  longitude = NULL,
  latitude = NULL,
  measure = c("geodesic", "haversine", "vincenty", "cheap")
)
}
\arguments{
\item{locationTbl}{Tibble of known locations.}

\item{longitude}{Target longitude in decimal degrees E.}

\item{latitude}{Target latitude in decimal degrees N.}

\item{measure}{One of "geodesic", "haversine", "vincenty" or "cheap"
specifying desired method of geodesic distance calculation.}
}
\value{
Tibble of distances in meters and cardinal directions from a target location.
}
\description{
Returns a tibble with the same number of rows as \code{locationTbl}
containing the distance and direction from the target location specified by
\code{longitude} and \code{latitude} to each known location found in
\code{locationTbl}.
}
\note{
Only a single target location is allowed.
}
\examples{
library(MazamaLocationUtils)

locationTbl <- get(data("wa_monitors_500"))

locationTbl \%>\%
  table_getDistanceFromTarget(
    longitude = -117.3647, 
    latitude = 47.6725
  ) \%>\% 
  dplyr::glimpse()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_getNearestDistance.R
\name{table_getNearestDistance}
\alias{table_getNearestDistance}
\title{Return distances to nearest known locations}
\usage{
table_getNearestDistance(
  locationTbl = NULL,
  longitude = NULL,
  latitude = NULL,
  distanceThreshold = NULL,
  measure = c("geodesic", "haversine", "vincenty", "cheap")
)
}
\arguments{
\item{locationTbl}{Tibble of known locations.}

\item{longitude}{Vector of target longitudes in decimal degrees E.}

\item{latitude}{Vector of target latitudes in decimal degrees N.}

\item{distanceThreshold}{Distance in meters.}

\item{measure}{One of "geodesic", "haversine", "vincenty" or "cheap"
specifying desired method of geodesic distance calculation.}
}
\value{
Vector of closest distances between target locations and known locations.
}
\description{
Returns distances between target locations and the closest
location found in \code{locationTbl} (if any). Target locations are specified
with \code{longitude} and \code{latitude}.

For each target location, only a single distance to the closest known location
is returned. If no known location is found within
\code{distanceThreshold}, the distance associated with that target location
will be \code{NA}. The length and order of resulting distances will match the
order of the incoming target locations.
}
\note{
The measure \code{"cheap"} may be used to speed things up depending on
the spatial scale being considered. Distances calculated with
\code{measure = "cheap"} will vary by a few meters compared with those
calculated using \code{measure = "geodesic"}.

See \code{\link[geodist:geodist]{geodist::geodist()}} for details.
}
\section{Use Case}{

You may have a set of locations of interest for which you want to assess whether
any monitoring locations are nearby. In this case, the locations of interest
will provide \code{longitude} and \code{latitude} while \code{locationTbl}
will be the known location table associated with the monitoring locations.

The resulting vector of distances will tell you the distance, for each
target location, to the nearst monitoring location.
}

\examples{
library(MazamaLocationUtils)

locationTbl <- get(data("wa_monitors_500"))

# Wenatchee
lon <- -120.325278
lat <- 47.423333

# Too small a distanceThreshold will not find a match
table_getNearestDistance(locationTbl, lon, lat, distanceThreshold = 50)

# Expanding the distanceThreshold will find one
table_getNearestDistance(locationTbl, lon, lat, distanceThreshold = 5000)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/location_getSingleAddress_Photon.R
\name{location_getSingleAddress_Photon}
\alias{location_getSingleAddress_Photon}
\title{Get address data from the Photon API to OpenStreetMap}
\usage{
location_getSingleAddress_Photon(
  longitude = NULL,
  latitude = NULL,
  baseUrl = "https://photon.komoot.io/reverse",
  verbose = TRUE
)
}
\arguments{
\item{longitude}{Single longitude in decimal degrees E.}

\item{latitude}{Single latitude in decimal degrees N.}

\item{baseUrl}{Base URL for data queries.}

\item{verbose}{Logical controlling the generation of progress messages.}
}
\value{
List of address components.
}
\description{
The Photon API is used get address data associated with
the \code{longitude} and \code{latitude}. The following list of data
is returned:

\itemize{
\item{\code{houseNumber}}
\item{\code{street}}
\item{\code{city}}
\item{\code{stateCode}}
\item{\code{stateName}}
\item{\code{postalCode}}
\item{\code{countryCode}}
\item{\code{countryName}}
}

The function makes an effort to convert both \code{state} and \code{country}
\code{Name} into \code{Code} with codes defaulting to \code{NA}. Both
\code{Name} and \code{Code} are returned so that improvements can be made in
the conversion algorithm.
}
\examples{
\donttest{
library(MazamaLocationUtils)

# Fail gracefully if any resources are not available
try({

  # Set up standard directories and spatial data
  spatialDataDir <- tempdir() # typically "~/Data/Spatial"
  initializeMazamaSpatialUtils(spatialDataDir)

  # Wenatchee
  lon <- -120.325278
  lat <- 47.423333

  addressList <- location_getSingleAddress_Photon(lon, lat)
  str(addressList)
  
}, silent = FALSE)
}

}
\references{
\url{https://photon.komoot.io}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{wa_airfire_meta}
\alias{wa_airfire_meta}
\title{Washington monitor metadata dataset}
\format{
A tibble with 92 rows and 29 columns of data.
}
\usage{
wa_airfire_meta
}
\description{
The \code{wa_pwfsl_meta} dataset provides a set of Washington
state air quality monitor metadata used by the USFS AirFire group.
This dataset was generated on 2023-10-24 by running:

\preformatted{
library(AirMonitor)

wa_airfire_meta <-
  airnow_loadLatest() \%>\%
  monitor_filter(stateCode == "WA") \%>\%
  monitor_getMeta() \%>\%
  # On 2023-10-24, this metdata still uses zip instead of postalCode
  dplyr::rename(postalCode = zip) \%>\%
  # Remove internal fields
  dplyr::select(-dplyr::starts_with("airnow_"))

save(wa_airfire_meta, file = "data/wa_airfire_meta.rda")
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/location_getSingleAddress_TexasAM.R
\name{location_getSingleAddress_TexasAM}
\alias{location_getSingleAddress_TexasAM}
\title{Get an address from the Texas A&M reverse geocoding service}
\usage{
location_getSingleAddress_TexasAM(
  longitude = NULL,
  latitude = NULL,
  apiKey = NULL,
  verbose = TRUE
)
}
\arguments{
\item{longitude}{Single longitude in decimal degrees E.}

\item{latitude}{Single latitude in decimal degrees N.}

\item{apiKey}{Texas A&M Geocoding requires an API key. The first 2500 requests
are free.}

\item{verbose}{Logical controlling the generation of progress messages.}
}
\value{
Numeric elevation value.
}
\description{
Texas A&M APIs are used to determine the address associated with
the \code{longitude} and \code{latitude}.
}
\examples{
\dontrun{
library(MazamaLocationUtils)

# Fail gracefully if any resources are not available
try({

  # Wenatchee
  longitude <- -122.47
  latitude <- 47.47
  apiKey <- YOUR_PERSONAL_API_KEY

  location_getSingleAddress_TexasAM(longitude, latitude, apiKey)
  
}, silent = FALSE)
}

}
\references{
\url{https://geoservices.tamu.edu/Services/ReverseGeocoding/WebService/v04_01/HTTP.aspx}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/location_getCensusBlock.R
\name{location_getCensusBlock}
\alias{location_getCensusBlock}
\title{Get census block data from the FCC API}
\usage{
location_getCensusBlock(
  longitude = NULL,
  latitude = NULL,
  censusYear = 2010,
  verbose = TRUE
)
}
\arguments{
\item{longitude}{Single longitude in decimal degrees E.}

\item{latitude}{Single latitude in decimal degrees N.}

\item{censusYear}{Year the census was taken.}

\item{verbose}{Logical controlling the generation of progress messages.}
}
\value{
List of census block/county/state data.
}
\description{
The FCC Block API is used get census block, county, and state FIPS associated with
the \code{longitude} and \code{latitude}. The following list of data
is returned:
\itemize{
\item{\code{stateCode}}
\item{\code{countyName}}
\item{\code{censusBlock}}
}
The data from this function should be considered to be the gold standard for state and county.
i.e. this information could and should be used to override information we get elsewhere.
}
\examples{
\donttest{
library(MazamaLocationUtils)

# Fail gracefully if any resources are not available
try({

  # Wenatchee
  lon <- -120.325278
  lat <- 47.423333

  censusList <- location_getCensusBlock(lon, lat)
  str(censusList)
  
}, silent = FALSE)
}
}
\references{
\url{https://geo.fcc.gov/api/census/}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_addLocation.R
\name{table_addLocation}
\alias{table_addLocation}
\title{Add new known location records to a table}
\usage{
table_addLocation(
  locationTbl = NULL,
  longitude = NULL,
  latitude = NULL,
  distanceThreshold = NULL,
  stateDataset = "NaturalEarthAdm1",
  elevationService = NULL,
  addressService = NULL,
  verbose = TRUE
)
}
\arguments{
\item{locationTbl}{Tibble of known locations.}

\item{longitude}{Vector of longitudes in decimal degrees E.}

\item{latitude}{Vector of latitudes in decimal degrees N.}

\item{distanceThreshold}{Distance in meters.}

\item{stateDataset}{Name of spatial dataset to use for determining state
codes, Default: 'NaturalEarthAdm1'}

\item{elevationService}{Name of the elevation service to use for determining
the elevation. Default: NULL skips this step. Accepted values: "usgs".}

\item{addressService}{Name of the address service to use for determining
the street address. Default: NULL skips this step. Accepted values: "photon".}

\item{verbose}{Logical controlling the generation of progress messages.}
}
\value{
Updated tibble of known locations.
}
\description{
Incoming \code{longitude} and \code{latitude} values are compared
against the incoming \code{locationTbl} to see if they are already within
\code{distanceThreshold} meters of an existing entry. A new record is created for
each location that is not already found in \code{locationTbl}.
}
\note{
This function is a vectorized version of \code{table_addSingleLocation()}.
}
\examples{
\donttest{
library(MazamaLocationUtils)

# Fail gracefully if any resources are not available
try({

  # Set up standard directories and spatial data
  spatialDataDir <- tempdir() # typically "~/Data/Spatial"
  initializeMazamaSpatialUtils(spatialDataDir)

  locationTbl <- get(data("wa_monitors_500"))

  # Coulee City, WA
  lon <- -119.290904
  lat <- 47.611942

  locationTbl <- 
    locationTbl \%>\%
    table_addLocation(lon, lat, distanceThreshold = 500)
  
  dplyr::glimpse(locationTbl)
  
}, silent = FALSE)
}
}
\seealso{
\code{\link[=table_addSingleLocation]{table_addSingleLocation()}}

\code{\link[=table_removeRecord]{table_removeRecord()}}

\code{\link[=table_updateSingleRecord]{table_updateSingleRecord()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{validateLocationTbl}
\alias{validateLocationTbl}
\title{Validate a location table}
\usage{
validateLocationTbl(locationTbl = NULL, locationOnly = TRUE)
}
\arguments{
\item{locationTbl}{Tibble of known locations.}

\item{locationOnly}{Logical specifying whether to check for all standard
columns.}
}
\value{
Invisibly returns \code{TRUE} if no error message has been generated.
}
\description{
Ensures that the incoming table has numeric \code{longitude} and
\code{latitude} columns.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-APIKey.R
\docType{data}
\name{APIKeys}
\alias{APIKeys}
\alias{getAPIKey}
\alias{setAPIKey}
\alias{showAPIKeys}
\title{API keys for data services.}
\format{
Character strings.
}
\description{
This package maintains an internal set of API keys which
users can set using \code{setAPIKey()}. These keys will be remembered for
the duration of an R session. This functionality provides an abstraction
layer in dependent packages so that data access functions can test for and
access specific API keys with generic code.
}
\details{
The following functions help with the management of API keys:

\code{getAPIKey()} -- Returns the API key associated with a web service.
If \code{provider == NULL} a list is returned containing all recognized
API keys.

\code{setAPIKey()} -- Sets the API key associated with a web service.
Silently returns previous value of the API key.

\code{showAPIKeys()} -- Returns a list of all currently set API keys.
}
\keyword{environment}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_leaflet.R
\name{table_leaflet}
\alias{table_leaflet}
\title{Leaflet interactive map for known locations}
\usage{
table_leaflet(
  locationTbl = NULL,
  maptype = c("terrain", "roadmap", "satellite", "toner"),
  extraVars = NULL,
  jitter = 5e-04,
  ...
)
}
\arguments{
\item{locationTbl}{Tibble of known locations.}

\item{maptype}{Optional name of leaflet ProviderTiles to use, e.g. \code{terrain}.}

\item{extraVars}{Character vector of addition \code{locationTbl} column names
to be shown in leaflet popups.}

\item{jitter}{Amount to use to slightly adjust locations so that multiple
monitors at the same location can be seen. Use zero or \code{NA} to see
precise locations.}

\item{...}{Additional arguments passed to \code{leaflet::addCircleMarkers()}.}
}
\value{
A leaflet "plot" object which, if not assigned, is rendered in
Rstudio's 'Viewer' tab.
}
\description{
This function creates interactive maps that will be displayed in
RStudio's 'Viewer' tab. The default setting of \code{jitter} will move locations
randomly within an ~50 meter radius so that overlapping locations can be
identified. Set \code{jitter = 0} to see precise locations.
}
\details{
The \code{maptype} argument is mapped onto leaflet "ProviderTile"
names. Current mappings include:
\itemize{
\item{"roadmap" => "OpenStreetMap"}
\item{"satellite" => "Esri.WorldImagery"}
\item{"terrain" => "Esri.WorldTopoMap"}
\item{"toner" => "Stamen.Toner"}
}

If a character string not listed above is provided, it will be used as the
underlying map tile if available. See
\url{https://leaflet-extras.github.io/leaflet-providers/} for a list of
"provider tiles" to use as the background map.
}
\examples{
\donttest{
if (interactive()) {
  # A table with all core metadata
  table_leaflet(wa_monitors_500)
  
  # A table missing some core metadata
  table_leaflet(
    wa_airfire_meta
  )

  # Customizing the map
  table_leaflet(
    wa_airfire_meta,
    extraVars = c("AQSID"),
    radius = 6,
    color = "black",
    weight = 2,
    fillColor = "red",
    fillOpacity = 0.3
  )
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_addClustering.R
\name{table_addClustering}
\alias{table_addClustering}
\title{Add clustering information to a dataframe}
\usage{
table_addClustering(
  tbl,
  clusterDiameter = 1000,
  lonVar = "longitude",
  latVar = "latitude",
  maxClusters = 50
)
}
\arguments{
\item{tbl}{Tibble with geolocation information (\emph{e.g.}.}

\item{clusterDiameter}{Diameter in meters used to determine the number of
clusters (see description).}

\item{lonVar}{Name of longitude variable in the incoming tibble.}

\item{latVar}{Name of the latitude variable in the incoming tibble.}

\item{maxClusters}{Maximum number of clusters to try.}
}
\value{
Input tibble with additional columns: \verb{clusterLon, clusterLat}.
}
\description{
Clustering is used to identify unique deployments of a
sensor in an environmental monitoring field study.

Sensors will be moved around from time to time, sometimes across the country
and sometimes across the street.  We would like to assign unique identifiers
to each new "deployment" but not when the sensor is moved a short distance.

We use clustering to find an appropriate number of unique "deployments".
The sensitivity of this algorithm can be adjused with the clusterDiameter argument.

Standard \code{kmeans} clustering does not work well when clusters can have widely
differing numbers of members. A much better result is acheived with
the Partitioning Around Medoids method available in \code{cluster::pam()}.

The value of \code{clusterRadius} is compared with the output of
\code{cluster::pam(...)$clusinfo[,'av_diss']} to determine the number of clusters.
}
\note{
The \code{table_addClustering()} function implements two-stage clustering
using \code{\link[=clusterByDistance]{clusterByDistance()}}. If the first attempt at clustering produces
clustered locations that are still too close to eachother, another round
of clustering is performed using the results of the previous attempt. This
two-stage approach seems to work well in. practice.
}
\examples{
library(MazamaLocationUtils)

# Fremont, Seattle 47.6504, -122.3509
# Magnolia, Seattle 47.6403, -122.3997
# Downtown Seattle 47.6055, -122.3370

fremont_x <- jitter(rep(-122.3509, 10), .0005)
fremont_y <- jitter(rep(47.6504, 10), .0005)

magnolia_x <- jitter(rep(-122.3997, 8), .0005)
magnolia_y <- jitter(rep(47.6403, 8), .0005)

downtown_x <- jitter(rep(-122.3370, 3), .0005)
downtown_y <- jitter(rep(47.6055, 3), .0005)

# Apply clustering
tbl <-
  dplyr::tibble(
    longitude = c(fremont_x, magnolia_x, downtown_x),
    latitude = c(fremont_y, magnolia_y, downtown_y)
  ) \%>\%
  table_addClustering(
    clusterDiameter = 1000
  )

plot(tbl$longitude, tbl$latitude, pch = tbl$clusterID)
}
\references{
\href{https://working-with-data.mazamascience.com/2021/07/15/when-k-means-clustering-fails/}{When k-means clustering fails}
}
\seealso{
\code{\link[=clusterByDistance]{clusterByDistance()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_load.R
\name{table_load}
\alias{table_load}
\title{Load a known location table}
\usage{
table_load(collectionName = NULL)
}
\arguments{
\item{collectionName}{Character identifier for this table.}
}
\value{
Tibble of known locations.
}
\description{
Load a tibble of known locations from the preferred directory.

The known location table must be named either \verb{<collectionName>.rda}
or \verb{<collectionName>.csv}. If both are found, only
\verb{<collectionName>.rda} will be loaded to ensure that columns will have
the proper type assigned.
}
\examples{
library(MazamaLocationUtils)

# Set the directory for saving location tables
setLocationDataDir(tempdir())

# Load an example table and check the dimensions
locationTbl <- get(data("wa_monitors_500"))
dim(locationTbl)

# Save it as "table_load_example"
table_save(locationTbl, "table_load_example")

# Load it and check the dimensions
my_table <- table_load("table_load_example")
dim(my_table)

# Check the locationDataDir
list.files(getLocationDataDir(), pattern = "table_load_example")
}
\seealso{
\code{\link[=setLocationDataDir]{setLocationDataDir()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_addCoreMetadata.R
\name{table_addCoreMetadata}
\alias{table_addCoreMetadata}
\title{Add missing core metadata columns to a known location table}
\usage{
table_addCoreMetadata(locationTbl = NULL, precision = 10)
}
\arguments{
\item{locationTbl}{Tibble of known locations. This input tibble need not be a
standardized "known location" with all required columns. They will be added.}

\item{precision}{\code{precision} argument passed on to \code{\link[=location_createID]{location_createID()}}.}
}
\value{
Tibble with the metadata columns required in a "known locations" table.
}
\description{
An existing table will be amended to guarantee that it includes
the following core metadata columns.

\itemize{
\item{locationID}
\item{locationName}
\item{longitude}
\item{latitude}
\item{elevation}
\item{countryCode}
\item{stateCode}
\item{countyName}
\item{timezone}
\item{houseNumber}
\item{street}
\item{city}
\item{postalCode}
}

The \code{longitude} and \code{latitude} columns are required to exist in the
incoming tibble but all others are optional.

If any of these core metadata columns are found, they will be retained.

The \code{locationID} will be generated (anew if already found) from existing
longitude and latitude data.

Other core metadata columns will be filled with \code{NA} values of the
proper type.

The result is a tibble with all of the core metadata columns. Theses columns
must then be filled in to create a usable "known locations" table.
}
\note{
No check is performed for overlapping locations. The returned tibble
has the structure of a "known locations" table and is a good starting place
for investigation. But further work is required to produce a valid table of
"known locations" associated with a specific spatial scale.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_getLocationID.R
\name{table_getLocationID}
\alias{table_getLocationID}
\title{Return IDs of known locations}
\usage{
table_getLocationID(
  locationTbl = NULL,
  longitude = NULL,
  latitude = NULL,
  distanceThreshold = NULL,
  measure = c("geodesic", "haversine", "vincenty", "cheap")
)
}
\arguments{
\item{locationTbl}{Tibble of known locations.}

\item{longitude}{Vector of longitudes in decimal degrees E.}

\item{latitude}{Vector of latitudes in decimal degrees N.}

\item{distanceThreshold}{Distance in meters.}

\item{measure}{One of "geodesic", "haversine", "vincenty" or "cheap"
specifying desired method of geodesic distance calculation.
See \code{\link[geodist:geodist]{geodist::geodist()}}.}
}
\value{
Vector of known \code{locationID}s.
}
\description{
Returns a vector of \code{locationID}s for the known locations
that each incoming location will be assigned to within the given. If more
than one known location exists within the given \code{distanceThreshold}, the closest will be
assigned.  \code{NA} will be returned for each incoming that cannot be
assigned to a known location in \code{locationTbl}.
}
\note{
The measure \code{"cheap"} may be used to speed things up depending on
the spatial scale being considered. Distances calculated with
\code{measure = "cheap"} will vary by a few meters compared with those
calculated using \code{measure = "geodesic"}.
}
\examples{
locationTbl <- get(data("wa_monitors_500"))

# Wenatchee
lon <- -120.325278
lat <- 47.423333

# Too small a distanceThreshold will not find a match
table_getLocationID(locationTbl, lon, lat, distanceThreshold = 50)

# Expanding the distanceThreshold will find one
table_getLocationID(locationTbl, lon, lat, distanceThreshold = 5000)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_leafletAdd.R
\name{table_leafletAdd}
\alias{table_leafletAdd}
\title{Add to a leaflet interactive map for known locations}
\usage{
table_leafletAdd(
  map = NULL,
  locationTbl = NULL,
  extraVars = NULL,
  jitter = 5e-04,
  ...
)
}
\arguments{
\item{map}{Leaflet map.}

\item{locationTbl}{Tibble of known locations.}

\item{extraVars}{Character vector of addition \code{locationTbl} column names
to be shown in leaflet popups.}

\item{jitter}{Amount to use to slightly adjust locations so that multiple
monitors at the same location can be seen. Use zero or \code{NA} to see
precise locations.}

\item{...}{Additional arguments passed to \code{leaflet::addCircleMarkers()}.}
}
\value{
A leaflet "plot" object which, if not assigned, is rendered in
Rstudio's 'Viewer' tab.
}
\description{
This function adds a layer to an interactive map displayed in
RStudio's 'Viewer' tab. The default setting of \code{jitter} will move locations
randomly within an ~50 meter radius so that overlapping locations can be
identified. Set \code{jitter = 0} to see precise locations.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_save.R
\name{table_save}
\alias{table_save}
\title{Save a known location table}
\usage{
table_save(
  locationTbl = NULL,
  collectionName = NULL,
  backup = TRUE,
  outputType = "rda"
)
}
\arguments{
\item{locationTbl}{Tibble of known locations.}

\item{collectionName}{Character identifier for this table.}

\item{backup}{Logical specifying whether to save a backup version of any
existing tables sharing \code{collectionName}.}

\item{outputType}{Vecctor of output formats. (Currently only "rda" or "csv" are supported.)}
}
\value{
File path of saved file.
}
\description{
Save a tibble of known locations to the preferred directory. If
\code{outputType} is a vector, the known locations table will be saved to the
preferred directory in multiple formats.
}
\details{
Backup files are saved with "YYYY-mm-ddTHH:MM:SS"
}
\examples{
library(MazamaLocationUtils)

# Set the directory for saving location tables
setLocationDataDir(tempdir())

# Load an example table and check the dimensions
locationTbl <- get(data("wa_monitors_500"))
dim(locationTbl)

# Save it as "table_save_example"
table_save(locationTbl, "table_save_example")

# Add a column and save again 
locationTbl \%>\% 
  table_addColumn("my_column") \%>\% 
  table_save("table_save_example")
  
# Check the locationDataDir
list.files(getLocationDataDir(), pattern = "table_save_example")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MazamaLocationUtils-package.R
\docType{package}
\name{MazamaLocationUtils}
\alias{MazamaLocationUtils-package}
\alias{MazamaLocationUtils}
\title{Manage Spatial Metadata for Known Locations}
\description{
A suite of utility functions for discovering and managing
metadata associated with sets of spatially unique "known locations".

This package is intended to be used in support of data management activities
associated with fixed locations in space. The motivating fields include both
air and water quality monitoring where fixed sensors report at regular time
intervals.
}
\details{
When working with environmental monitoring time series, one of the first things
you have to do is create unique identifiers for each individual time series. In
an ideal world, each environmental time series would have both a
\code{locationID} and a \code{deviceID} that uniquely identify the specific instrument
making measurements and the physical location where measurements are made. A
unique \code{timeseriesID} could
be produced as \code{locationID_deviceID}. Metadata associated with each
\code{timeseriesID} would contain basic information needed for downstream analysis
including at least:

\preformatted{
  timeseriesID, locationID, deviceID, longitude, latitude, ...
}

\itemize{
\item{An extended time series for an occasionally re-positioned sensor would group by \code{deviceID}.}
\item{Multiple sensors placed at a single location could be be grouped by \code{locationID}.}
\item{Maps would be created using \verb{longitude, latitude}.}
\item{Time series would be accessed from a secondary \code{data} table with \code{timeseriesID}.}
}

Unfortunately, we are rarely supplied with a truly unique and truly spatial
\code{locationID}. Instead we often use \code{deviceID} or an associated non-spatial
identifier as a stand-in for \code{locationID}.

Complications we have seen include:

\itemize{
\item{GPS-reported longitude and latitude can have \emph{jitter} in the fourth or fifth
decimal place making it challenging to use them to create a unique \code{locationID}.}
\item{Sensors are sometimes \emph{re-positioned} in what the scientist considers the "same
location".}
\item{Data for a single sensor goes through different processing pipelines using
different identifiers and is later brought together as two separate time series.}
\item{The spatial scale of what constitutes a "single location" depends on the
instrumentation and scientific question being asked.}
\item{Deriving location-based metadata from spatial datasets is computationally
intensive unless saved and identified with a unique \code{locationID}.}
\item{Automated searches for spatial metadata occasionally produce incorrect results
because of the non-infinite resolution of spatial datasets.}
}

This package attempts to address all of these issues by maintaining a table
of known locations for which CPU intensive spatial data calculations have
already been performed. While requests to add new locations to the table may
take some time, searches for spatial metadata associated with existing
locations are simple lookups.

Working in this manner will solve the problems initially mentioned but also
provides further useful functionality.

\itemize{
\item{Administrators can correct entries in the \code{collectionName} table.
(\emph{e.g.} locations in river bends that even high resolution spatial datasets mis-assign)}
\item{Additional, non-automatable metadata can be added to \code{collectionName}.
(\emph{e.g.} commonly used location names within a community of practice)}
\item{Different field campaigns can have separate \code{collectionName} tables.}
\item{\code{.csv} or \code{.rda} versions of well populated tables can be downloaded from a
URL and used locally, giving scientists working with known locations instant
access to spatial data that otherwise requires special skills, large datasets
and lots of compute cycles.}
}
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/MazamaScience/MazamaLocationUtils}
  \item Report bugs at \url{https://github.com/MazamaScience/MazamaLocationUtils/issues}
}

}
\author{
\strong{Maintainer}: Jonathan Callahan \email{jonathan.s.callahan@gmail.com}

Other contributors:
\itemize{
  \item Eli Grosman \email{eligrosman1@gmail.com} [contributor]
  \item Oliver Fogelin \email{oliverfogelin@gmail.com} [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{id_monitors_500}
\alias{id_monitors_500}
\title{Idaho monitor locations dataset}
\format{
A tibble with 30 rows and 13 columns of data.
}
\usage{
id_monitors_500
}
\description{
The \code{id_monitor_500} dataset provides a set of known
locations associated with Idaho state air quality monitors.
This dataset was generated on 2023-10-24 by running:

\preformatted{
library(AirMonitor)
library(MazamaLocationUtils)

initializeMazamaSpatialUtils()
setLocationDataDir("./data")

monitor <- monitor_loadLatest() \%>\% monitor_filter(stateCode == "ID")
lons <- monitor$meta$longitude
lats <- monitor$meta$latitude

table_initialize() \%>\%
  table_addLocation(
    lons, lats,
    distanceThreshold = 500,
    elevationService = "usgs",
    addressService = "photon"
  ) \%>\%
  table_save("id_monitors_500")
}
}
\seealso{
\code{\link[=or_monitors_500]{or_monitors_500()}}

\code{\link[=wa_monitors_500]{wa_monitors_500()}}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_removeColumn.R
\name{table_removeColumn}
\alias{table_removeColumn}
\title{Remove a column of metadata in a table}
\usage{
table_removeColumn(locationTbl = NULL, columnName = NULL, verbose = TRUE)
}
\arguments{
\item{locationTbl}{Tibble of known locations.}

\item{columnName}{Name of the colun to be removed.}

\item{verbose}{Logical controlling the generation of progress messages.}
}
\value{
Updated tibble of known locations.
}
\description{
Remove the column matching \code{columnName}. This function
can be used in pipelines.
}
\examples{
library(MazamaLocationUtils)

# Starting table
locationTbl <- get(data("wa_monitors_500")) 
names(locationTbl)

# Add a new column
locationTbl <-
  locationTbl \%>\%
  table_addColumn("AQSID")
  
names(locationTbl)

# Now remove it
locationTbl <-
  locationTbl \%>\%
  table_removeColumn("AQSID")

names(locationTbl)

try({
  # Cannot remove "core" metadata
  locationTbl <-
    locationTbl \%>\%
    table_removeColumn("longitude")
}, silent = FALSE)

}
\seealso{
\code{\link[=table_addColumn]{table_addColumn()}}

\code{\link[=table_removeColumn]{table_removeColumn()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_findAdjacentDistances.R
\name{table_findAdjacentDistances}
\alias{table_findAdjacentDistances}
\title{Find distances between adjacent locations in a known locations table}
\usage{
table_findAdjacentDistances(
  locationTbl = NULL,
  distanceThreshold = NULL,
  measure = c("geodesic", "haversine", "vincenty", "cheap")
)
}
\arguments{
\item{locationTbl}{Tibble of known locations.}

\item{distanceThreshold}{Distance in meters.}

\item{measure}{One of "haversine" "vincenty", "geodesic", or "cheap"
specifying desired method of geodesic distance calculation.

See \code{\link[geodist:geodist]{geodist::geodist()}} for details.}
}
\value{
Tibble of row indices and distances for those locations separated by
less than \code{distanceThreshold} meters.
}
\description{
Calculate distances between all locations within a known
locations table and return a tibble with the row indices and separation
distances of those records separated by less than \code{distanceThreshold}
meters. Records are returned in order of distance.

It is useful when working with new metadata tables to identify adjacent
locations early on so that decisions can be made about the appropriateness
of the specified \code{distanceThreshold}.
}
\note{
The measure \code{"cheap"} may be used to speed things up depending on
the spatial scale being considered. Distances calculated with
\code{measure = "cheap"} will vary by a few meters compared with those
calculated using \code{measure = "geodesic"}.
}
\examples{
library(MazamaLocationUtils)

meta <- wa_airfire_meta

# Any locations closer than 2 km?
table_findAdjacentDistances(meta, distanceThreshold = 2000)

# How about 4 km?
table_findAdjacentDistances(meta, distanceThreshold = 4000)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MazamaLocationUtils-package.R
\docType{data}
\name{coreMetadataNames}
\alias{coreMetadataNames}
\title{Names of standard spatial metadata columns}
\format{
A character vector
}
\usage{
coreMetadataNames
}
\description{
Character string identifiers of the minimum set of fields
required for a table to be considered a valid "known locations" table.

\preformatted{
coreMetadataNames <- c(
  "locationID",           # from MazamaLocationUtils::location_createID()
  "locationName",         # from MazamaLocationUtils::location_initialize()
  "longitude",            # user supplied
  "latitude",             # user supplied
  "elevation",            # from MazamaLocationUtils::getSingleElevation_USGS()
  "countryCode",          # from MazamaSpatialUtils::getCountryCode()
  "stateCode",            # from MazamaSpatialUtils::getStateCode()
  "countyName",           # from MazamaSpatialUtils::getUSCounty()
  "timezone",             # from MazamaSpatialUtils::getTimezone()
  "houseNumber",
  "street",
  "city",
  "postalCode"
)
}
}
\details{
coreMetadataNames
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MazamaLocationUtils-package.R
\name{setLocationDataDir}
\alias{setLocationDataDir}
\title{Set location data directory}
\usage{
setLocationDataDir(dataDir)
}
\arguments{
\item{dataDir}{Directory where location tables are stored.}
}
\value{
Silently returns previous value of the data directory.
}
\description{
Sets the data directory where known location data tables are located.
If the directory does not exist, it will be created.
}
\seealso{
\code{\link[=LocationDataDir]{LocationDataDir()}}

\code{\link[=getLocationDataDir]{getLocationDataDir()}}
}
\keyword{environment}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/location_initialize.R
\name{location_initialize}
\alias{location_initialize}
\title{Create known location record with core metadata}
\usage{
location_initialize(
  longitude = NULL,
  latitude = NULL,
  stateDataset = "NaturalEarthAdm1",
  elevationService = NULL,
  addressService = NULL,
  precision = 10,
  verbose = TRUE
)
}
\arguments{
\item{longitude}{Single longitude in decimal degrees E.}

\item{latitude}{Single latitude in decimal degrees N.}

\item{stateDataset}{Name of spatial dataset to use for determining state}

\item{elevationService}{Name of the elevation service to use for determining
the elevation. Default: NULL skips this step. Accepted values: "usgs".}

\item{addressService}{Name of the address service to use for determining
the street address. Default: NULL skips this step. Accepted values: "photon".}

\item{precision}{\code{precision} argument passed on to \code{\link[=location_createID]{location_createID()}}.}

\item{verbose}{Logical controlling the generation of progress messages.}
}
\value{
Tibble with a single new known location.
}
\description{
Creates a known location record with the following columns
of core metadata:
\itemize{
\item{locationID}
\item{locationName}
\item{longitude}
\item{latitude}
\item{elevation}
\item{countryCode}
\item{stateCode}
\item{countyName}
\item{timezone}
\item{houseNumber}
\item{street}
\item{city}
\item{postalCode}
}
}
\examples{
\donttest{
library(MazamaLocationUtils)

# Fail gracefully if any resources are not available
try({

  # Set up standard directories and spatial data
  spatialDataDir <- tempdir() # typically "~/Data/Spatial"
  initializeMazamaSpatialUtils(spatialDataDir)

  # Wenatchee
  lon <- -120.325278
  lat <- 47.423333
  
  locationRecord <- location_initialize(lon, lat)
  str(locationRecord)
  
}, silent = FALSE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/initializeMazamaSpatialUtils.R
\name{initializeMazamaSpatialUtils}
\alias{initializeMazamaSpatialUtils}
\title{Initialize MazamaSpatialUtils package}
\usage{
initializeMazamaSpatialUtils(spatialDataDir = "~/Data/Spatial")
}
\arguments{
\item{spatialDataDir}{Directory where \pkg{MazamaSpatialUtils} datasets are found.}
}
\description{
Convenience function that wraps:

\preformatted{
  MazamaSpatialUtils::setSpatialDataDir("~/Data/Spatial")
  MazamaSpatialUtils::loadSpatialData("EEZCountries.rda")
.  MazamaSpatialUtils::loadSpatialData("OSMTimezones.rda")
  MazamaSpatialUtils::loadSpatialData("NaturalEarthAdm1.rda")
  MazamaSpatialUtils::loadSpatialData("USCensusCounties.rda")
}

If spatial data has not yet been installed, an error is returned with an
extended message detailing how to install the appropriate data.
}
\examples{
\donttest{
library(MazamaLocationUtils)

# Fail gracefully if any resources are not available
try({

  # Set up directory for spatial data
  spatialDataDir <- tempdir() # typically "~/Data/Spatial"
  MazamaSpatialUtils::setSpatialDataDir(spatialDataDir)

  exists("NaturalEarthAdm1")
  initializeMazamaSpatialUtils(spatialDataDir)
  exists("NaturalEarthAdm1")
  class(NaturalEarthAdm1)
  
}, silent = FALSE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_findAdjacentLocations.R
\name{table_findAdjacentLocations}
\alias{table_findAdjacentLocations}
\title{Finds adjacent locations in a known locations table.}
\usage{
table_findAdjacentLocations(
  locationTbl = NULL,
  distanceThreshold = NULL,
  measure = c("geodesic", "haversine", "vincenty", "cheap")
)
}
\arguments{
\item{locationTbl}{Tibble of known locations.}

\item{distanceThreshold}{Distance in meters.}

\item{measure}{One of "haversine" "vincenty", "geodesic", or "cheap"
specifying desired method of geodesic distance calculation.

See \code{\link[geodist:geodist]{geodist::geodist()}} for details.}
}
\value{
Tibble of known locations separated by less than
\code{distanceThreshold} meters.
}
\description{
Calculate distances between all locations within a known
locations table and return a tibble containing all records that have an
adjacent location separated by less than \code{distanceThreshold} meters.
The return tibble is ordered by separation distance.

It is useful when working with new metadata tables to identify adjacent
locations early on so that decisions can be made about the appropriateness
of the specified \code{distanceThreshold}.
}
\note{
The measure \code{"cheap"} may be used to speed things up depending on
the spatial scale being considered. Distances calculated with
\code{measure = "cheap"} will vary by a few meters compared with those
calculated using \code{measure = "geodesic"}.
}
\examples{
library(MazamaLocationUtils)

meta <- wa_airfire_meta

# Any locations closer than 2 km?
meta \%>\%
  table_findAdjacentLocations(distanceThreshold = 2000) \%>\%
  dplyr::select(AQSID, timezone)

# How about 4 km?
meta \%>\%
  table_findAdjacentLocations(distanceThreshold = 4000) \%>\%
  dplyr::select(AQSID, timezone)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{validateMazamaSpatialUtils}
\alias{validateMazamaSpatialUtils}
\title{Validate proper setup of MazamaSpatialUtils}
\usage{
validateMazamaSpatialUtils()
}
\value{
Invisibly returns \code{TRUE} if no error message has been generated.
}
\description{
The \pkg{MazamaSpatialUtils} package mus be properly installed
and initialized before using functions from the \pkg{MazamaLocationUtils}
package. This function tests for this.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_getRecordIndex.R
\name{table_getRecordIndex}
\alias{table_getRecordIndex}
\title{Return indexes of known location records}
\usage{
table_getRecordIndex(locationTbl = NULL, locationID = NULL, verbose = TRUE)
}
\arguments{
\item{locationTbl}{Tibble of known locations.}

\item{locationID}{Vector of \code{locationID} strings.}

\item{verbose}{Logical controlling the generation of progress messages.}
}
\value{
Vector of \code{locationTbl} row indexes.
}
\description{
Returns a vector of \code{locationTbl} row indexes for the
locations associated with each \code{locationID}.
}
\examples{
library(MazamaLocationUtils)

locationTbl <- get(data("wa_monitors_500"))

# Wenatchee
lon <- -120.325278
lat <- 47.423333

# Get the locationID first
locationID <- table_getLocationID(locationTbl, lon, lat, distanceThreshold = 5000)

# Now find the row associated with this ID
recordIndex <- table_getRecordIndex(locationTbl, locationID)

str(locationTbl[recordIndex,])

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_initializeExisting.R
\name{table_initializeExisting}
\alias{table_initializeExisting}
\title{Converts an existing table into a known location table}
\usage{
table_initializeExisting(
  locationTbl = NULL,
  stateDataset = "NaturalEarthAdm1",
  countryCodes = NULL,
  distanceThreshold = NULL,
  measure = c("geodesic", "haversine", "vincenty", "cheap"),
  precision = 10,
  verbose = TRUE
)
}
\arguments{
\item{locationTbl}{Tibble of known locations. This input tibble need not be a
standardized "known location" table with all required columns. Missing
columns will be added.}

\item{stateDataset}{Name of spatial dataset to use for determining state
codes, Default: 'NaturalEarthAdm1'}

\item{countryCodes}{Vector of country codes used to optimize spatial
searching. (See ?MazamaSpatialUtils::getStateCode())}

\item{distanceThreshold}{Distance in meters.}

\item{measure}{One of "haversine" "vincenty", "geodesic", or "cheap"
specifying desired method of geodesic distance calculation. See
\code{?geodist::geodist}.}

\item{precision}{\code{precision} argument passed on to \code{\link[=location_createID]{location_createID()}}.}

\item{verbose}{Logical controlling the generation of progress messages.}
}
\value{
Known location tibble with the specified metadata columns. Any
locations whose circles (as defined by \code{distanceThreshold}) overlap will
generate warning messages.

It is incumbent upon the user to address overlapping locations by one of:

\enumerate{
\item{reduce the distanceThreshold until no overlaps occur}
\item{assign one of the overlapping locations to the other location}
}
}
\description{
An existing table may have much of the data that is needed
for a known location table. This function accepts an incoming table and
searches for required columns:

\itemize{
\item{locationID}
\item{locationName}
\item{longitude}
\item{latitude}
\item{elevation}
\item{countryCode}
\item{stateCode}
\item{countyName}
\item{timezone}
\item{houseNumber}
\item{street}
\item{city}
\item{postalCode}
}

The \code{longitude} and \code{latitude} columns are required but all others
are optional.

If any of these optional columns are found, they will be used and the often
slow and sometimes slightly inaccurate steps to generate that information
will be skipped for locations that have non-missing data. Any additional
columns of information that are
not part of the required core metadata will be retained.

This method skips the assignment of columns like \code{elevation} and all
address related fields that require web service requests.

Compared to initializing a brand new table and populating it one record at a
time, this is a much faster way of creating a known location table from a
pre-existing table of metadata.
}
\note{
The measure \code{"cheap"} may be used to speed things up depending on
the spatial scale being considered. Distances calculated with
\code{measure = "cheap"} will vary by a few meters compared with those
calculated using \code{measure = "geodesic"}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_addSingleLocation.R
\name{table_addSingleLocation}
\alias{table_addSingleLocation}
\title{Add a single new known location record to a table}
\usage{
table_addSingleLocation(
  locationTbl = NULL,
  longitude = NULL,
  latitude = NULL,
  distanceThreshold = NULL,
  stateDataset = "NaturalEarthAdm1",
  elevationService = NULL,
  addressService = NULL,
  verbose = TRUE
)
}
\arguments{
\item{locationTbl}{Tibble of known locations.}

\item{longitude}{Single longitude in decimal degrees E.}

\item{latitude}{Single latitude in decimal degrees N.}

\item{distanceThreshold}{Distance in meters.}

\item{stateDataset}{Name of spatial dataset to use for determining state
codes, Default: "NaturalEarthAdm1".}

\item{elevationService}{Name of the elevation service to use for determining
the elevation. Default: NULL. Accepted values: "usgs".}

\item{addressService}{Name of the address service to use for determining
the street address. Default: NULL. Accepted values: "photon".}

\item{verbose}{Logical controlling the generation of progress messages.}
}
\value{
Updated tibble of known locations.
}
\description{
Incoming \code{longitude} and \code{latitude} values are compared
against the incoming \code{locationTbl} to see if they are already within
\code{distanceThreshold} meters of an existing entry.  A new record is created for
if the location is not already found in \code{locationTbl}.
}
\examples{
\donttest{
library(MazamaLocationUtils)

# Fail gracefully if any resources are not available
try({

  # Set up standard directories and spatial data
  spatialDataDir <- tempdir() # typically "~/Data/Spatial"
  initializeMazamaSpatialUtils(spatialDataDir)

  locationTbl <- get(data("wa_monitors_500"))
  
  nrow(locationTbl)

  # Coulee City, WA
  lon <- -119.290904
  lat <- 47.611942

  locationTbl <- 
    locationTbl \%>\%
    table_addSingleLocation(lon, lat, distanceThreshold = 500)
    
  nrow(locationTbl)
  
}, silent = FALSE)
}

}
\seealso{
\code{\link[=table_addLocation]{table_addLocation()}}

\code{\link[=table_removeRecord]{table_removeRecord()}}

\code{\link[=table_updateSingleRecord]{table_updateSingleRecord()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/location_getOpenCageInfo.R
\name{location_getOpenCageInfo}
\alias{location_getOpenCageInfo}
\title{Get location information from OpenCage}
\usage{
location_getOpenCageInfo(longitude = NULL, latitude = NULL, verbose = FALSE)
}
\arguments{
\item{longitude}{Single longitude in decimal degrees E.}

\item{latitude}{Single latitude in decimal degrees N.}

\item{verbose}{Logical controlling the generation of progress messages.}
}
\value{
Single-row tibble with OpenCage information.
}
\description{
The OpenCage reverse geocoding service is used to obtain all
available information for a specific location.

The data from OpenCage should be considered to be the gold standard for address
information could and should be used to override information we get elsewhere.
}
\note{
The OpenCage service requires an API key which can be obtained from
their web site. This API key must be set as an environment variable with:

\preformatted{
Sys.setenv("OPENCAGE_KEY" = "YOUR_PERSONAL_API_KEY")
}

The OpenCage "free trial" level allows for 1 request/sec and a maximum of
2500 requests per day.
}
\examples{
\donttest{
library(MazamaLocationUtils)

# Fail gracefully if any resources are not available
try({

  # Wenatchee
  lon <- -120.325278
  lat <- 47.423333

  Sys.setenv("OPENCAGE_KEY" = "YOUR_PERSONAL_API_KEY")

  openCageTbl <- location_getOpenCageInfo(lon, lat)
  dplyr::glimpse(openCageTbl)
  
}, silent = FALSE)
}

}
\references{
\url{https://opencagedata.com}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_removeRecord.R
\name{table_removeRecord}
\alias{table_removeRecord}
\title{Remove location records from a table}
\usage{
table_removeRecord(locationTbl = NULL, locationID = NULL, verbose = TRUE)
}
\arguments{
\item{locationTbl}{Tibble of known locations.}

\item{locationID}{Vector of \code{locationID} strings.}

\item{verbose}{Logical controlling the generation of progress messages.}
}
\value{
Updated tibble of known locations.
}
\description{
Incoming \code{locationID} values are compared
against the incoming \code{locationTbl} and any matches are removed.
}
\examples{
library(MazamaLocationUtils)

locationTbl <- get(data("wa_monitors_500"))
dim(locationTbl)

# Wenatchee
lon <- -120.325278
lat <- 47.423333

# Get the locationID first
locationID <- table_getLocationID(locationTbl, lon, lat, distanceThreshold = 500)

# Remove it
locationTbl <- table_removeRecord(locationTbl, locationID)
dim(locationTbl)

# Test
table_getLocationID(locationTbl, lon, lat, distanceThreshold = 500)

}
\seealso{
\code{\link[=table_addLocation]{table_addLocation()}}

\code{\link[=table_addSingleLocation]{table_addSingleLocation()}}

\code{\link[=table_updateSingleRecord]{table_updateSingleRecord()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_addOpenCageInfo.R
\name{table_addOpenCageInfo}
\alias{table_addOpenCageInfo}
\title{Add address fields to a known location table}
\usage{
table_addOpenCageInfo(
  locationTbl = NULL,
  replaceExisting = FALSE,
  retainOpenCage = FALSE,
  verbose = FALSE
)
}
\arguments{
\item{locationTbl}{Tibble of known locations.}

\item{replaceExisting}{Logical specifying whether to replace existing data
with data obtained from OpenCage.}

\item{retainOpenCage}{Logical specifying whether to retain all fields obtained
from OpenCage, each named with a prefix of \code{opencage_}.}

\item{verbose}{Logical controlling the generation of progress messages.}
}
\value{
Tibble of "known locations" enhanced with information from the
OpenCage reverse geocoding service.
}
\description{
The OpenCage reverse geocoding service is used to update an
existing table. Updated columns include:

\itemize{
\item{countryCode}
\item{stateCode}
\item{countyName}
\item{timezone}
\item{houseNumber}
\item{street}
\item{city}
\item{postalCode}
\item{address}
}

When \code{replaceExisting = TRUE}, all existing address fields are discarded
in favor of the OpenCage versions. To only fill in missing values in
\code{locationTbl}, use \code{replaceExisting = FALSE}.

The OpenCage service returns a large number of fields, some of which may be
useful. To add all OpenCage fields to a location table, use
\code{retainOpenCage = TRUE}. This will append 78+ fields of information,
each each named with a prefix of \code{"opencage_"}.
}
\note{
The OpenCage service requires an API key which can be obtained from
their web site. This API key must be set as an environment variable with:

\preformatted{
Sys.setenv("OPENCAGE_KEY" = "<your api key>")
}

Parameters are set for use at the OpenCage "free trial" level which allows
for 1 request/sec and a maximum of 2500 requests per day.

Because of the 1 request/sec default, it is recommended that
\code{table_addOpenCageInfo()} only be used in an interactive session when
updating a table with a large number of records.
}
\examples{
\donttest{
library(MazamaLocationUtils)

# Fail gracefully if any resources are not available
try({

  myTbl <- id_monitors_500[1:3,]
  myTbl$countryCode[1] <- NA
  myTbl$countryCode[2] <- "WRONG"
  myTbl$countyName[3] <- "WRONG"
  myTbl$timezone <- NA

  dplyr::glimpse(myTbl)

  Sys.setenv("OPENCAGE_KEY" = "<YOUR_KEY>")

  table_addOpenCageInfo(myTbl) \%>\% 
    dplyr::glimpse()

  table_addOpenCageInfo(myTbl, replaceExisting = TRUE) \%>\% 
    dplyr::glimpse()

  table_addOpenCageInfo(myTbl, replaceExisting = TRUE, retainOpenCage = TRUE) \%>\% 
    dplyr::glimpse()
  
}, silent = FALSE)
}
}
\references{
\url{https://opencagedata.com}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_updateSingleRecord.R
\name{table_updateSingleRecord}
\alias{table_updateSingleRecord}
\title{Update a single known location record in a table}
\usage{
table_updateSingleRecord(
  locationTbl = NULL,
  locationList = NULL,
  verbose = TRUE
)
}
\arguments{
\item{locationTbl}{Tibble of known locations.}

\item{locationList}{List containing \code{locationID} and one or more named
columns whose values are to be replaced.}

\item{verbose}{Logical controlling the generation of progress messages.}
}
\value{
Updated tibble of known locations.
}
\description{
Information in the \code{locationList} is used to replace
existing information found in \code{locationTbl}. This function can be used
for small tweaks to an existing \code{locationTbl}. Wholesale replacement of
records should be performed with \code{table_removeRecord()} followed by
\code{table_addLocation()}.
}
\examples{
library(MazamaLocationUtils)

locationTbl <- get(data("wa_monitors_500"))

# Wenatchee
wenatcheeRecord <- 
  locationTbl \%>\% 
  dplyr::filter(city == "Wenatchee")

str(wenatcheeRecord)

wenatcheeID <- wenatcheeRecord$locationID

locationTbl <- table_updateSingleRecord(
  locationTbl,
  locationList = list(
    locationID = wenatcheeID,
    locationName = "Wenatchee-Fifth St"
  )
)

# Look at the new record
locationTbl \%>\% 
  dplyr::filter(city == "Wenatchee") \%>\%
  str()

}
\seealso{
\code{\link[=table_addLocation]{table_addLocation()}}

\code{\link[=table_addSingleLocation]{table_addSingleLocation()}}

\code{\link[=table_removeRecord]{table_removeRecord()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_addColumn.R
\name{table_addColumn}
\alias{table_addColumn}
\title{Add a new column of metadata to a table}
\usage{
table_addColumn(
  locationTbl = NULL,
  columnName = NULL,
  locationID = NULL,
  locationData = NULL,
  verbose = TRUE
)
}
\arguments{
\item{locationTbl}{Tibble of known locations.}

\item{columnName}{Name to use for the new column.}

\item{locationID}{Vector of \code{locationID} strings.}

\item{locationData}{Vector of data to used at matching records.}

\item{verbose}{Logical controlling the generation of progress messages.}
}
\value{
Updated tibble of known locations.
}
\description{
A new metadata column is added to the \code{locationTbl}. For
matching \code{locationID} records, the associated \code{locationData} is
inserted. Otherwise, the new column will be initialized with \code{NA}.
}
\examples{
library(MazamaLocationUtils)

# Starting table
locationTbl <- get(data("wa_monitors_500")) 
names(locationTbl)

# Add an empty column
locationTbl <-
  locationTbl \%>\%
  table_addColumn("AQSID")
  
names(locationTbl)

}
\seealso{
\code{\link[=table_removeColumn]{table_removeColumn()}}

\code{\link[=table_updateColumn]{table_updateColumn()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/location_createID.R
\name{location_createID}
\alias{location_createID}
\title{Create one or more unique locationIDs}
\usage{
location_createID(
  longitude = NULL,
  latitude = NULL,
  algorithm = c("geohash", "digest"),
  precision = 10
)
}
\arguments{
\item{longitude}{Vector of longitudes in decimal degrees E.}

\item{latitude}{Vector of latitudes in decimal degrees N.}

\item{algorithm}{Algorithm to use -- either \code{"geohash"} or \code{"digest"}.}

\item{precision}{\code{precision} argument used when encoding with \code{"geohash"}.}
}
\value{
Vector of character locationIDs.
}
\description{
A unique locationID is created for each incoming
\code{longitude} and \code{latitude}.

See \code{\link[MazamaCoreUtils:createLocationID]{MazamaCoreUtils::createLocationID()}} for details.

At \code{precision = 10}, this results in a maximum error of 0.6 meters which
is more than precise enough for environmental monitoring studies making use
of this package.

An excellent way to become familiar with geohash is through the
\href{https://geohash.softeng.co}{GeoHash Explorer}.
}
\note{
The \code{"digest"} algorithm is deprecated but provided for backwards
compatibility with databases that were built using locationIDs generated
with this algorithm.
}
\examples{
library(MazamaLocationUtils)

# Wenatchee
lon <- -120.325278
lat <- 47.423333
locationID <- location_createID(lon, lat)
print(locationID)

location_createID(lon, lat, algorithm = "geohash")
location_createID(lon, lat, algorithm = "geohash", precision = 7)

}
\references{
\url{https://en.wikipedia.org/wiki/Decimal_degrees}

\url{https://www.johndcook.com/blog/2017/01/10/probability-of-secure-hash-collisions/}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_initialize.R
\name{table_initialize}
\alias{table_initialize}
\title{Create an empty known location table}
\usage{
table_initialize()
}
\value{
Empty known location tibble with the specified metadata columns.
}
\description{
Creates an empty known location tibble with the following
columns of core metadata:

\itemize{
\item{locationID}
\item{locationName}
\item{longitude}
\item{latitude}
\item{elevation}
\item{countryCode}
\item{stateCode}
\item{countyName}
\item{timezone}
\item{houseNumber}
\item{street}
\item{city}
\item{postalCode}
}
}
\examples{
library(MazamaLocationUtils)

# Create an empty Tbl
emptyTbl <- table_initialize()
dplyr::glimpse(emptyTbl)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MazamaLocationUtils-package.R
\name{getLocationDataDir}
\alias{getLocationDataDir}
\title{Get location data directory}
\usage{
getLocationDataDir()
}
\value{
Absolute path string.
}
\description{
Returns the directory path where known location data tables are located.
}
\seealso{
\code{\link[=LocationDataDir]{LocationDataDir()}}

\code{\link[=setLocationDataDir]{setLocationDataDir()}}
}
\keyword{environment}
